using System;
using System.Drawing;
using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.DirectInput;
using Microsoft.DirectX.Direct3D;
using Direct3D = Microsoft.DirectX.Direct3D;

namespace entity.Renderer
{
    public class BSPCollisionViewer : Form
    {
        // Our global variables for this project
 
        Direct3D.Device device = null; // Our rendering device
 
        PresentParameters presentParams = new PresentParameters();
        bool pause = false;
        BSP.BSPModel.BSPCollision coll;
        Matrix WorldTransform = Matrix.Identity;
        entity.Renderer.Camera2 cam;
        int mapnumber;
        //Materials
         Material BlueMaterial;
        Material BlackMaterial;
        Material GreenMaterial;
        Material DefaultMaterial;
        MetaContainers.DirectXBSPCollision dxcoll;


        public BSPCollisionViewer(BSP.BSPModel.BSPCollision tempcoll, int mapnumberx)
            
        {
            mapnumber = mapnumberx;
            // Set the initial size of our form
            this.ClientSize = new System.Drawing.Size(800, 600);
            // And its caption
            this.Text = "BSP Viewer";
            this.MouseDown += new MouseEventHandler(BSPCollisionViewer_MouseDown);
            this.MouseMove += new System.Windows.Forms.MouseEventHandler(this.ModelViewer_MouseDown);
            coll = tempcoll;
          
              

     

            Main();
        }

        void BSPCollisionViewer_MouseDown(object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Right)
            {
                cam.oldx = e.X;
                cam.oldy = e.Y;
            }
        }

        

        public bool InitializeGraphics()
        {
            
            //try
            //{
            this.Show();
            this.Focus();
            Application.DoEvents();
            presentParams.Windowed = true; // We don't want to run fullscreen
            presentParams.PresentationInterval = Direct3D.PresentInterval.Default;
            presentParams.FullScreenRefreshRateInHz =
        Direct3D.PresentParameters.DefaultPresentRate;

            presentParams.SwapEffect = SwapEffect.Discard; // Discard the frames 
            presentParams.EnableAutoDepthStencil = true; // Turn on a Depth stencil
            presentParams.AutoDepthStencilFormat = DepthFormat.D16; // And the stencil format
            device = new Direct3D.Device(0, Direct3D.DeviceType.Hardware, this, CreateFlags.SoftwareVertexProcessing, presentParams); //Create a device
            device.DeviceReset += new System.EventHandler(this.OnResetDevice);
            this.OnResetDevice(device, null);
            pause = false;
       
   
       
             DefaultMaterial = new Material();
            DefaultMaterial.Diffuse = Color.White;
            DefaultMaterial.Ambient = Color.White;
            BlackMaterial  = new Material();
            BlackMaterial.Diffuse = Color.Black;
            BlackMaterial.Ambient = Color.Black;
            BlueMaterial = new Material();
            BlueMaterial.Diffuse = Color.Blue;
            BlueMaterial.Ambient = Color.Blue;
            GreenMaterial = new Material();
            GreenMaterial.Diffuse = Color.GreenYellow;
            GreenMaterial.Ambient = Color.GreenYellow;

            LoadMesh();
            return true;

        }

        public void LoadMesh()
        {

            dxcoll = new entity.MetaContainers.DirectXBSPCollision(ref device,ref coll);
           
             
           
        }

        public void OnResetDevice(object sender, EventArgs e)
        {
            Direct3D.Device dev = (Direct3D.Device)sender;
            // Turn off culling, so we see the front and back of the triangle
            dev.RenderState.CullMode = Cull.None ;
            // Turn off D3D lighting
            dev.RenderState.Lighting = true;
            // Turn on the ZBuffer
            dev.RenderState.ZBufferEnable = true;
            dev.RenderState.FillMode = FillMode.Solid;
            dev.RenderState.SourceBlend = Blend.SourceAlpha;
            dev.RenderState.DestinationBlend = Blend.InvSourceAlpha;
            dev.RenderState.Ambient = Color.White;
            device.SamplerState[0].MinFilter = TextureFilter.Linear;
            device.SamplerState[0].MagFilter = TextureFilter.Linear;
            device.SamplerState[0].MipFilter = TextureFilter.Linear;
            device.SamplerState[1].MinFilter = TextureFilter.Linear;
            device.SamplerState[1].MagFilter = TextureFilter.Linear;
            device.SamplerState[1].MipFilter = TextureFilter.Linear;
        }


        private void SetupMatrices()
        {
            device.Transform.World =  Matrix.Identity;// Matrix.RotationAxis(new Vector3((float)Math.Cos(Environment.TickCount / 250.0f), 1, (float)Math.Sin(Environment.TickCount / 250.0f)), Environment.TickCount / 1000.0f);
            device.Transform.View = Matrix.LookAtRH(cam.Position, cam.LookAt, cam.UpVector);
            device.Transform.Projection = Matrix.PerspectiveFovRH((float)Math.PI / 4.0f, 1.0f, 0.1f, 10000.0f);
        }

        private void Render()
        {
            if (pause)
                return;

            
            device.Clear(ClearFlags.Target | ClearFlags.ZBuffer, System.Drawing.Color.Black, 1.0f, 0);
            device.BeginScene();
            cam.move();
         
            SetupMatrices();



         
            device.RenderState.FillMode = FillMode.WireFrame;
            device.Material = DefaultMaterial;
  
            device.Transform.World=Matrix.Identity;
            //Raw.ParsedModel.DisplayedInfo.Draw(ref device, ref bsp.SkyBox);
            dxcoll.Draw(ref device);

       
            device.EndScene();
            // Update the screen
            device.Present();
        }

        protected override void OnPaint(System.Windows.Forms.PaintEventArgs e)
        {
            this.Render(); // Render on painting
        }
        protected override void OnKeyPress(System.Windows.Forms.KeyPressEventArgs e)
        {
            if ((int)(byte)e.KeyChar == (int)System.Windows.Forms.Keys.Escape)
                this.Dispose(); // Esc was pressed
        }
        protected override void OnResize(System.EventArgs e)
        {
       //     this.Render();//
            pause = ((this.WindowState == FormWindowState.Minimized) || !this.Visible);
        }



        /// <summary>
        /// The main entry point for the application.
        /// </summary>
        public void Main()
        {

            using (BSPCollisionViewer frm = (BSPCollisionViewer)this)
            {
                if (!frm.InitializeGraphics()) // Initialize Direct3D
                {
                    MessageBox.Show("Could not initialize Direct3D.  This tutorial will exit.");
                    return;
                }
                frm.Show();
                frm.Focus();
                cam = new entity.Renderer.Camera2(this);// Camera(device);
                // While the form is still valid, render and process messages
                while (frm.Created)
                {
                    frm.Render();
                    Application.DoEvents();
                }
            }
        }

        private void InitializeComponent()
        {
            this.SuspendLayout();
            // 
            // BSPViewer
            // 
            this.ClientSize = new System.Drawing.Size(292, 266);
            this.Name = "BSPViewer";
            this.ResumeLayout(false);

        }

        private void ModelViewer_MouseDown(object sender, MouseEventArgs e)
        {

         if (e.Button== MouseButtons.Right)
         {
                cam.change(e.X, e.Y);
         }
           // if (cam.mousenavigation == Camera.MouseNavigationTypes.None)
           // {
            //    cam.StartMouseNavigation(e);
          //  }
          //  cam.NavigateByMouseTrackball(e);
        
        }


        private void ModelViewer_MouseUp(object sender, MouseEventArgs e)
        {

          //  cam.StopMouseNavigation(e);

        }


}
}
