using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using System.Windows.Forms;
using System.Collections;
using entity.Renderer.BSP_Renderer;
using entity.Raw;
namespace entity.BSP
{
    public struct HaloBSPVertex
    {
        public Vector3 Position;
        public Vector3 Normal;

        public int diffuse;
       // public int specular;
        public float Tu0, Tv0;
        public float Tu1, Tv1;
        public float Tu2, Tv2;
        public float Tu3, Tv3;
        public static readonly VertexFormats FVF = VertexFormats.PositionNormal | VertexFormats.Diffuse  | VertexFormats.Texture4;
    }

    public class BSPModel : IDisposable
    {
        #region Fields
        public int BSPRawDataMetaChunksOffset;
        public BSPRawDataMetaChunk[] BSPRawDataMetaChunks;
        public int BSPPermutationRawDataMetaChunksOffset;
        public BSPPermutationRawDataMetaChunk[] BSPPermutationRawDataMetaChunks;
        public PermutationPlacement[] PermutationInfo;
        public BSPShaderContainer Shaders;
        public BSPDisplayedInfo Display;
        public MetaContainers.Sky sky;
        public ParsedModel SkyBox;
        public SpawnInfo Spawns;
        public string Name;
        public UnknownChunk[] UnknownChunks;
        public BSP.ClusterPVS[] ClusterInfo;
        public Raw.ParsedBitmap LightMapParsedBitmap;
        public Texture[] LightMapTexture;
        public System.Drawing.Bitmap[] LightMapBitmap;
        public Texture[] SceneryLightMapTexture;
        public System.Drawing.Bitmap[] SceneryLightMapBitmap;
        public  int BspNumber;
     
        #endregion
        #region Constructor
        public BSPModel()
        {
        }

        public BSPModel(ref Meta meta, int mapnumber)
        {
            string[] temps = meta.name.Split('\\');
            Name = temps[temps.Length - 1];
            BspNumber = Maps.map[mapnumber].BSP.FindBSPNumberByBSPIdent(meta.ident);

            switch (Maps.map[mapnumber].HaloVersion)
            {
                case Map.HaloVersionEnum.Halo2:
            LoadModelStructure(ref meta, mapnumber);
            LoadPermutations(ref meta, mapnumber);
            LoadUnknowns(ref meta, mapnumber);
            Shaders = new BSPShaderContainer(this,ref meta, mapnumber);
            Display = new BSPDisplayedInfo();
            LoadLightmaps(mapnumber);
            sky = new entity.MetaContainers.Sky();
            LoadSky(ref meta, mapnumber);
            Spawns = new SpawnInfo(mapnumber);
            ClusterInfo = BSP.ClustersPVS.GetClustersPVSInfo(ref meta, mapnumber);
            break;
        case Map.HaloVersionEnum.Halo1:
      case Map.HaloVersionEnum.HaloCE:
          Shaders = new BSPShaderContainer();
          this.BSPPermutationRawDataMetaChunks = new BSPPermutationRawDataMetaChunk[0];
          this.LightMapBitmap = new System.Drawing.Bitmap[0];
          this.SceneryLightMapBitmap = new System.Drawing.Bitmap[0];
          this.PermutationInfo = new PermutationPlacement[0];
            LoadModelStructure(ref meta, mapnumber);

          
            Display = new BSPDisplayedInfo();

            Spawns = new SpawnInfo(mapnumber);
            sky = new entity.MetaContainers.Sky();
           
            break;
        }


        }
        #endregion
      
        #region CleanUp
        public void Dispose()
        {
            this.BSPPermutationRawDataMetaChunks = null;
            this.PermutationInfo = null;
            this.BSPRawDataMetaChunks = null;
            this.Display = null;

            this.Shaders = null;
            this.SkyBox = null;
            this.Spawns = null;
            GC.SuppressFinalize(this);
        }
        #endregion
        #region Vertices
        public static void AddUncompressedVertice(ref BinaryReader BR, ref BSPRawDataMetaChunk chunk)
        {
            Vector3 vertice = new Vector3(BR.ReadSingle(), BR.ReadSingle(), BR.ReadSingle());
            chunk.Vertices.Add(vertice);
            Vector3 normal = new Vector3(BR.ReadSingle(), BR.ReadSingle(), BR.ReadSingle());
            chunk.Normals.Add(normal);
            Vector3 binormal = new Vector3(BR.ReadSingle(), BR.ReadSingle(), BR.ReadSingle());
            chunk.Binormals.Add(binormal);
            Vector3 tangent = new Vector3(BR.ReadSingle(), BR.ReadSingle(), BR.ReadSingle());
            chunk.Tangents.Add(tangent);
            Vector2 uv = new Vector2(BR.ReadSingle(), BR.ReadSingle());

            chunk.UVs.Add(uv);
        }
        public static void AddCompressedVertice(ref BinaryReader BR,ref BSPRawDataMetaChunk chunk)
        {
                         //vertexes
                        Vector3 vertice = new Vector3(BR.ReadSingle(), BR.ReadSingle(), BR.ReadSingle());
                        chunk.Vertices.Add(vertice);
                        //normals
                        int test = BR.ReadInt32();
                        Vector3 normal2 = ParsedModel.DecompressNormal(test);
                        chunk.Normals.Add(normal2);
                        test = BR.ReadInt32();
                        Vector3 binormal2 = ParsedModel.DecompressNormal(test);
                        chunk.Binormals.Add(binormal2);
                        test = BR.ReadInt32();
                        Vector3 tangent2 = ParsedModel.DecompressNormal(test);
                        chunk.Tangents.Add(tangent2);
                        //basemap uvs
                   
      
                       // short testx = BR.ReadInt16();

                        float u = BR.ReadSingle();// chunk.DecompressVertice(Convert.ToSingle(testx), -1, 1);
                        //testx = BR.ReadInt16();
                        float v = BR.ReadSingle(); //chunk.DecompressVertice(Convert.ToSingle(testx), -1, 1);// % 1;
                        Vector2 uv2 = new Vector2(u, v);
                        chunk.UVs.Add(uv2);
           
                        //lightmap uvs
                      // testx = BR.ReadInt16();
                       // float ux = chunk.DecompressVertice(Convert.ToSingle(testx), -1, 1);// %1 ;
                       // testx = BR.ReadInt16();
                       // float vx =chunk.DecompressVertice(Convert.ToSingle(testx), -1, 1);// % 1;
                       // Vector2 uv2x = new Vector2(ux, vx);
                       // chunk.LightMapUVs.Add(uv2);
        }
        #endregion
        #region Faces
        public class Face
        {
            public int vertex0Index;
            public int vertex1Index;
            public int vertex2Index;
        }
        #endregion
        #region Sky
        public void LoadSky(ref Meta meta, int mapnumber)
        {
            Maps.map[mapnumber].OpenMap(MapTypes.Internal);


            Maps.map[mapnumber].BR.BaseStream.Position = Maps.map[mapnumber].MetaInfo.Offset[3] + 8;
            int tempc = Maps.map[mapnumber].BR.ReadInt32();
            int tempr = Maps.map[mapnumber].BR.ReadInt32() - Maps.map[mapnumber].SecondaryMagic;
            if (tempc == 0) { return; }
            Maps.map[mapnumber].BR.BaseStream.Position = tempr + 4;

            int tempident = Maps.map[mapnumber].Functions.Meta.FindMetaByID(Maps.map[mapnumber].BR.ReadInt32(), mapnumber);

            if (tempident != -1)
            {
                sky = new entity.MetaContainers.Sky(tempident, mapnumber);


                Maps.map[mapnumber].BR.BaseStream.Position = Maps.map[mapnumber].MetaInfo.Offset[tempident] + 4;
                tempident = Maps.map[mapnumber].Functions.Meta.FindMetaByID(Maps.map[mapnumber].BR.ReadInt32(), mapnumber);
                if (tempident != -1)
                {
                    Meta tempmeta = new Meta();
                    tempmeta.ReadMetaFromMap(tempident, mapnumber, false);
                    SkyBox = new Raw.ParsedModel(ref tempmeta, mapnumber);

                }
            }




            Maps.map[mapnumber].CloseMap();
        }
        #endregion
        #region Visual Mesh
        public class BSPRawDataMetaChunk : ParsedModel.RawDataMetaChunk
        {
            public int rawdatainfooffset;
            public int shadertagnumber;
            public List<Vector2> LightMapUVs = new List<Vector2>();
            #region Constructor
            public BSPRawDataMetaChunk():base()
            {
            }
            public BSPRawDataMetaChunk(int offset, int chunknumber, ref Meta meta, Face[] faces)
                : base()
            {
                switch (Maps.map[meta.mapnumber].HaloVersion)
                {

                    case Map.HaloVersionEnum.Halo2:
                        H2BSPRawDataMetaChunk(offset, chunknumber, ref meta);
                        break;
                    case Map.HaloVersionEnum.HaloCE:
                        CEBSPRawDataMetaChunk(offset, chunknumber, ref meta, faces);
                        break;
                    case Map.HaloVersionEnum.Halo1:
                        H1BSPRawDataMetaChunk(offset, chunknumber, ref meta,faces);
                        break;
                }
            }
            public void CEBSPRawDataMetaChunk(int offset, int chunknumber, ref Meta meta, Face[] faces)
            {
                BinaryReader BR = new BinaryReader(meta.MS);
                BR.BaseStream.Position = offset + 12;
                this.shadertagnumber = Maps.map[meta.mapnumber].Functions.Meta.FindMetaByID(BR.ReadInt32(), meta.mapnumber);
                this.RawDataChunkInfo = new ParsedModel.RawDataMetaChunk.RawDataOffsetChunk[1];
                BR.BaseStream.Position = offset + 20;
                int surfacestart = BR.ReadInt32();
                this.FaceCount = BR.ReadInt32();
                this.SubMeshInfo = new ModelSubMeshInfo[1];
                this.SubMeshInfo[0] = new ModelSubMeshInfo();
                this.SubMeshInfo[0].IndiceCount = this.FaceCount * 3;
                this.SubMeshInfo[0].ShaderNumber = chunknumber;
                this.IndiceCount = this.SubMeshInfo[0].IndiceCount;
                this.Indices = new short[this.IndiceCount];
                for (int x = 0; x < this.FaceCount; x++)
                {
                    this.Indices[(x * 3) + 0] = (short)faces[surfacestart + x].vertex0Index;
                    this.Indices[(x * 3) + 1] = (short)faces[surfacestart + x].vertex1Index;
                    this.Indices[(x * 3) + 2] = (short)faces[surfacestart + x].vertex2Index;
                }
                BR.BaseStream.Position = offset + 180;
                this.VerticeCount = BR.ReadInt32();
                BR.BaseStream.Position = offset + 228;
                int vertstranslation = BR.ReadInt32() - meta.magic - meta.offset;
                BR.BaseStream.Position = vertstranslation;
                BSPRawDataMetaChunk tempchunk = new BSPRawDataMetaChunk();
                for (int x = 0; x < this.VerticeCount; x++)
                {

                    AddUncompressedVertice(ref BR, ref tempchunk);
                }
                this.Vertices = tempchunk.Vertices;
                this.Normals = tempchunk.Normals;
                this.Binormals = tempchunk.Binormals;
                this.Tangents = tempchunk.Tangents;
                this.UVs = tempchunk.UVs;
                this.LightMapUVs = tempchunk.LightMapUVs;
            }
            public void H1BSPRawDataMetaChunk(int offset, int chunknumber, ref Meta meta,Face[] faces)
            {
                BinaryReader BR = new BinaryReader(meta.MS);
                BR.BaseStream.Position = offset + 12;
                this.shadertagnumber = Maps.map[meta.mapnumber].Functions.Meta.FindMetaByID(BR.ReadInt32(), meta.mapnumber);
                this.RawDataChunkInfo = new ParsedModel.RawDataMetaChunk.RawDataOffsetChunk[1];
                BR.BaseStream.Position = offset + 20;
                int surfacestart = BR.ReadInt32();
                this.FaceCount = BR.ReadInt32();
                this.SubMeshInfo = new ModelSubMeshInfo[1];
                this.SubMeshInfo[0] = new ModelSubMeshInfo();
                this.SubMeshInfo[0].IndiceCount = this.FaceCount * 3;
                this.SubMeshInfo[0].ShaderNumber = chunknumber;
                this.IndiceCount = this.SubMeshInfo[0].IndiceCount;
                this.Indices = new short[this.IndiceCount];
                for (int x = 0; x < this.FaceCount; x++)
                {
                    this.Indices[(x * 3) + 0] = (short)faces[surfacestart+x].vertex0Index;
                    this.Indices[(x * 3) + 1] = (short)faces[surfacestart+x].vertex1Index;
                    this.Indices[(x * 3) + 2] = (short)faces[surfacestart+x].vertex2Index;
                }
                BR.BaseStream.Position = offset + 180;
                this.VerticeCount = BR.ReadInt32();
                BR.BaseStream.Position = offset + 248;
                int vertstranslation=BR.ReadInt32()-meta.magic-meta.offset;
                BR.BaseStream.Position=vertstranslation;
                BSPRawDataMetaChunk tempchunk = new BSPRawDataMetaChunk();
                for (int x = 0; x < this.VerticeCount; x++)
                {
              
                    AddCompressedVertice(ref BR, ref tempchunk);
                }
               this.Vertices=tempchunk.Vertices;
               this.Normals = tempchunk.Normals;
               this.Binormals = tempchunk.Binormals;
               this.Tangents = tempchunk.Tangents;
               this.UVs = tempchunk.UVs;
               this.LightMapUVs = tempchunk.LightMapUVs;
            }

            public void H2BSPRawDataMetaChunk(int offset, int chunknumber, ref Meta meta)
            {
                BinaryReader BR = new BinaryReader(meta.MS);
                BR.BaseStream.Position = offset;
                this.VerticeCount = BR.ReadUInt16();
                this.FaceCount = BR.ReadUInt16();
                BR.BaseStream.Position = offset + 48;
                this.HeaderSize = BR.ReadInt32() + 8;
                BR.ReadInt32();
                int tempc = BR.ReadInt32();
                int tempr = BR.ReadInt32() - meta.magic - meta.offset;
                this.RawDataChunkInfo = new RawDataOffsetChunk[tempc];
                rawdatainfooffset = tempr;
                for (int x = 0; x < tempc; x++)
                {
                    this.RawDataChunkInfo[x] = new RawDataOffsetChunk();
                    BR.BaseStream.Position = tempr + (x * 16) + 6;
                    this.RawDataChunkInfo[x].ChunkSize = BR.ReadUInt16();
                    this.RawDataChunkInfo[x].Size = BR.ReadInt32();
                    if (this.RawDataChunkInfo[x].ChunkSize == 0) { this.RawDataChunkInfo[x].ChunkSize = this.RawDataChunkInfo[x].Size; }
                    this.RawDataChunkInfo[x].Offset = BR.ReadInt32();
                    this.RawDataChunkInfo[x].ChunkCount = this.RawDataChunkInfo[x].Size / this.RawDataChunkInfo[x].ChunkSize;
                }
                if (this.RawDataChunkInfo.Length == 0) { return; }
                BR = new BinaryReader(((RawDataChunk)meta.raw.rawChunks[chunknumber]).MS);
                this.SubMeshInfo = new ModelSubMeshInfo[this.RawDataChunkInfo[0].ChunkCount];
                for (int x = 0; x < this.RawDataChunkInfo[0].ChunkCount; x++)
                {
                    this.SubMeshInfo[x] = new ModelSubMeshInfo();
                    BR.BaseStream.Position = this.HeaderSize + this.RawDataChunkInfo[0].Offset + (x * 72) + 4;
                    this.SubMeshInfo[x].ShaderNumber = BR.ReadUInt16();
                    this.SubMeshInfo[x].IndiceStart = BR.ReadUInt16();
                    this.SubMeshInfo[x].IndiceCount = BR.ReadUInt16();

                }

                BR.BaseStream.Position = 40;
                this.IndiceCount = BR.ReadUInt16();
                int indicechunk = 0;
                int verticechunk = 0;
                int uvchunk = 0;
                for (int x = 0; x < RawDataChunkInfo.Length; x++)
                {
                    if (RawDataChunkInfo[x].ChunkSize == 2)
                    {
                        indicechunk = x;
                        break;
                    }
                }
                int normalchunk = 0;
                for (int x = indicechunk; x < RawDataChunkInfo.Length; x++)
                {
                    if (RawDataChunkInfo[x].ChunkCount == 1)
                    {
                        verticechunk = x;
                        uvchunk = x + 1;
                        normalchunk = x + 2;
                        break;
                    }
                }
                this.Indices = new short[this.RawDataChunkInfo[indicechunk].ChunkCount];
                BR.BaseStream.Position = this.HeaderSize + this.RawDataChunkInfo[indicechunk].Offset;
                for (int x = 0; x < this.IndiceCount; x++)
                {
                    this.Indices[x] = (short)BR.ReadUInt16();

                }

               
                this.RawDataChunkInfo[verticechunk].ChunkSize = this.RawDataChunkInfo[verticechunk].Size / VerticeCount;
                for (int x = 0; x < this.VerticeCount; x++)
                {
                    Vector3 vec = new Vector3();
                    BR.BaseStream.Position = this.HeaderSize + this.RawDataChunkInfo[verticechunk].Offset + (this.RawDataChunkInfo[verticechunk].ChunkSize * x);
                    vec.X = BR.ReadSingle();
                    vec.Y = BR.ReadSingle();
                    vec.Z = BR.ReadSingle();
                    Vertices.Add(vec);
                }

               
                this.RawDataChunkInfo[uvchunk].ChunkSize = 8;
                for (int x = 0; x < this.VerticeCount; x++)
                {
                    Vector2 tempuv = new Vector2();
                    BR.BaseStream.Position = this.HeaderSize + this.RawDataChunkInfo[uvchunk].Offset + (this.RawDataChunkInfo[uvchunk].ChunkSize * x);
                    tempuv.X = BR.ReadSingle();
                    tempuv.Y = BR.ReadSingle();
                    this.UVs.Add(tempuv);
                }

                this.RawDataChunkInfo[normalchunk].ChunkSize = 12;
                for (int x = 0; x < this.VerticeCount; x++)
                {
                    Vector2 tempuv = new Vector2();
                    BR.BaseStream.Position = this.HeaderSize + this.RawDataChunkInfo[normalchunk].Offset + (this.RawDataChunkInfo[normalchunk].ChunkSize * x);
                    Vector3 normal = Raw.ParsedModel.DecompressNormal(BR.ReadInt32());
                    this.Normals.Add(normal);
                    Vector3 binormal = Raw.ParsedModel.DecompressNormal(BR.ReadInt32());
                    this.Binormals.Add(binormal);
                    Vector3 tangent = Raw.ParsedModel.DecompressNormal(BR.ReadInt32());
                    this.Tangents.Add(tangent);
                }

                int lightmapuvchunk = -1;
                for (int x = normalchunk + 1; x < RawDataChunkInfo.Length; x++)
                {
                    if (RawDataChunkInfo[x].ChunkSize == 3)
                    {
                        lightmapuvchunk = x;
                        break;
                    }
                }

                if (lightmapuvchunk == -1) return;
                RawDataChunkInfo[lightmapuvchunk].ChunkSize = 4;
                for (int x = 0; x < this.VerticeCount; x++)
                {
                    Vector2 tempuv = new Vector2();
                    BR.BaseStream.Position = this.HeaderSize + this.RawDataChunkInfo[lightmapuvchunk].Offset + (this.RawDataChunkInfo[lightmapuvchunk].ChunkSize * x);
                    short testx = BR.ReadInt16();
                    float u = DecompressVertice(Convert.ToSingle(testx), -1, 1);
                    testx = BR.ReadInt16();
                    float v = DecompressVertice(Convert.ToSingle(testx), -1, 1);
                     Vector2 uv2 = new Vector2(u, v);
                    this.LightMapUVs.Add(uv2);
                }

            }
            #endregion
        }
        public void LoadModelStructure(ref Meta meta, int mapnumber)
        {
            Maps.map[mapnumber].OpenMap(MapTypes.Internal);
            BinaryReader BR = new BinaryReader(meta.MS);
            switch (Maps.map[mapnumber].HaloVersion)
            {
                case Map.HaloVersionEnum.Halo2:
                    #region H2
                    BR.BaseStream.Position = 172;
                    int tempc = BR.ReadInt32();
                    int tempr = BR.ReadInt32() - meta.magic - meta.offset;
                    BSPRawDataMetaChunks = new BSPRawDataMetaChunk[tempc];
                    BSPRawDataMetaChunksOffset = tempr;
                    for (int x = 0; x < tempc; x++)
                    {
                        BSPRawDataMetaChunks[x] = new BSPRawDataMetaChunk(tempr + (x * 176), x, ref meta,null);
                    }
                    #endregion
                    break;
                case Map.HaloVersionEnum.HaloCE:
                case Map.HaloVersionEnum.Halo1:
                    List<BSPRawDataMetaChunk> tempchunks = new List<BSPRawDataMetaChunk>();


                 
       

                    BR.BaseStream.Position = Maps.map[mapnumber].BSP.sbsp[BspNumber].lightmapoffset + 248;
                    int facescount = BR.ReadInt32();
                    int facestranslation =  BR.ReadInt32() - meta.magic - meta.offset;
                    Face[] faces = new Face[facescount];


                    BR.BaseStream.Position = facestranslation;
                    for (int x = 0; x < facescount; x++)
                    {
                        faces[x] = new Face();
                        faces[x].vertex0Index = BR.ReadInt16();
                        faces[x].vertex1Index = BR.ReadInt16();
                        faces[x].vertex2Index = BR.ReadInt16();
                    }


                    BR.BaseStream.Position = Maps.map[mapnumber].BSP.sbsp[BspNumber].lightmapoffset+260;
                   int tempc2 = BR.ReadInt32();
                   int tempr2 = BR.ReadInt32() - meta.magic - meta.offset;
                  
                   List<MetaContainers.ShaderInfo> tempshad=new List<entity.MetaContainers.ShaderInfo>();
                   BSPRawDataMetaChunksOffset = tempr2;
                    for (int x = 0; x < tempc2; x++)
                    {
                        BR.BaseStream.Position = tempr2 + (32 * x)+20;
                        int tempc22 = BR.ReadInt32();
                        int tempr22 = BR.ReadInt32() - meta.magic - meta.offset;
                        for (int xx = 0; xx < tempc22; xx++)
                        {
          
                            BSPRawDataMetaChunk tempraw= new BSPRawDataMetaChunk(tempr22 + (256 * xx),tempchunks.Count,ref meta,faces);
                            tempchunks.Add(tempraw);
                            tempshad.Add(new entity.MetaContainers.ShaderInfo(tempraw.shadertagnumber, meta.mapnumber));
                        }


                    }

                    this.Shaders.Shader = tempshad.ToArray();
                    this.BSPRawDataMetaChunks = tempchunks.ToArray();

                    
                   
                    break;
            }

            Maps.map[mapnumber].CloseMap();
        }
         #endregion
        #region Unknown
        public class UnknownChunk
        {
            #region Fields
            public ParsedModel.RawDataMetaChunk.RawDataOffsetChunk[] RawDataChunkInfo;
            public int VerticeCount;
            public int HeaderSize;
            public int IndiceCount;
            public int facecount;
            public short[] Indices;
            public List<Vector3> Vertices = new List<Vector3>();
            #endregion
            #region Constructor
            public UnknownChunk(int offset, int chunknumber, ref Meta meta)
            {
                #region Meta
                BinaryReader BR = new BinaryReader(meta.MS);
                BR.BaseStream.Position = offset;


                BR.BaseStream.Position = offset + 8;
                this.HeaderSize = BR.ReadInt32() + 8;
                #region LoadRawDataChunkInfo
                BR.BaseStream.Position = offset + 16;
                int tempc = BR.ReadInt32();
                int tempr = BR.ReadInt32() - meta.magic - meta.offset;
                this.RawDataChunkInfo = new ParsedModel.RawDataMetaChunk.RawDataOffsetChunk[tempc];
                for (int x = 0; x < tempc; x++)
                {
                    this.RawDataChunkInfo[x] = new ParsedModel.RawDataMetaChunk.RawDataOffsetChunk();
                    BR.BaseStream.Position = tempr + (x * 16) + 6;
                    this.RawDataChunkInfo[x].ChunkSize = BR.ReadUInt16();
                    this.RawDataChunkInfo[x].Size = BR.ReadInt32();
                    if (this.RawDataChunkInfo[x].ChunkSize == 0) { this.RawDataChunkInfo[x].ChunkSize = this.RawDataChunkInfo[x].Size; }
                    this.RawDataChunkInfo[x].Offset = BR.ReadInt32();
                    this.RawDataChunkInfo[x].ChunkCount = this.RawDataChunkInfo[x].Size / this.RawDataChunkInfo[x].ChunkSize;
                }
                if (this.RawDataChunkInfo.Length == 0) { return; }
                #endregion
                #endregion
                #region RawData
                int firstbsp2chunk = 0;
                #region FindFirstRawData
                for (int x = 0; x < meta.raw.rawChunks.Count; x++)
                {
                    if (((RawDataChunk)meta.raw.rawChunks[x]).rawDataType == RawDataType.bsp3)
                    {
                        firstbsp2chunk = x;
                        break;
                    }
                }
                #endregion
                BR = new BinaryReader(((RawDataChunk)meta.raw.rawChunks[firstbsp2chunk + chunknumber]).MS);

                #region DetermineChunkNumers
                int indicechunk = 0;
                int verticechunk = 0;

                for (int x = 0; x < RawDataChunkInfo.Length; x++)
                {
                    if (RawDataChunkInfo[x].ChunkSize == 32)
                    {

                        verticechunk = x;
                        indicechunk = x + 1;
                        break;
                    }
                }

                #endregion
                #region Faces
                //    BR.BaseStream.Position = 88;
                this.IndiceCount = this.RawDataChunkInfo[indicechunk].ChunkCount;
                this.Indices = new short[this.IndiceCount];
                BR.BaseStream.Position = this.HeaderSize + this.RawDataChunkInfo[indicechunk].Offset;
                for (int x = 0; x < this.IndiceCount; x++)
                {
                    this.Indices[x] = (short)BR.ReadUInt16();

                }
                #endregion


                #region Vertices
                this.VerticeCount = this.RawDataChunkInfo[verticechunk].Size / 32;

                for (int x = 0; x < this.VerticeCount; x++)
                {
                    Vector3 vec = new Vector3();
                    BR.BaseStream.Position = this.HeaderSize + this.RawDataChunkInfo[verticechunk].Offset + (this.RawDataChunkInfo[verticechunk].ChunkSize * x);
                    vec.X = BR.ReadSingle();
                    vec.Y = BR.ReadSingle();
                    vec.Z = BR.ReadSingle();
                    Vertices.Add(vec);
                }
                #endregion

            }
                #endregion
            #endregion
            #region Classes

            #endregion
        }
        public void LoadUnknowns(ref Meta meta, int mapnumber)
        {
            Maps.map[mapnumber].OpenMap(MapTypes.Internal);
            BinaryReader BR = new BinaryReader(meta.MS);
            BR.BaseStream.Position = 580;
            int tempc = BR.ReadInt32();
            int tempr = BR.ReadInt32() - meta.magic - meta.offset;
            BR.BaseStream.Position = tempr + 16;
            tempc = BR.ReadInt32();
            tempr = BR.ReadInt32() - meta.magic - meta.offset;
            UnknownChunks = new UnknownChunk[tempc];
            for (int x = 0; x < tempc; x++)
            {
                UnknownChunks[x] = new UnknownChunk(tempr + (x * 44), x, ref meta);
            }
            Maps.map[mapnumber].CloseMap();
        }
        #endregion
        #region Permutations
         public class BSPPermutationRawDataMetaChunk : ParsedModel.RawDataMetaChunk
        {
            #region Fields
            public BSPPermutationBoundingBoxContainer BoundingBox;
            public List<Vector2> LightMapUVs = new List<Vector2>();
            #endregion
            #region Constructor
            public BSPPermutationRawDataMetaChunk(int offset, int chunknumber, ref Meta meta)
            {
                #region Meta
                BinaryReader BR = new BinaryReader(meta.MS);
                BR.BaseStream.Position = offset;
                this.VerticeCount = BR.ReadUInt16();
                this.FaceCount = BR.ReadUInt16();
                #region LoadBoundingBox
                BR.BaseStream.Position = offset + 24;
                int tempc = BR.ReadInt32();
                int tempr = BR.ReadInt32() - meta.magic - meta.offset;
                BR.BaseStream.Position = tempr;
                //BoundingBox = new BSPPermutationBoundingBoxContainer(ref BR);
                #endregion
                BR.BaseStream.Position = offset + 48;
                this.HeaderSize = BR.ReadInt32() + 8;
                #region LoadRawDataChunkInfo
                BR.BaseStream.Position = offset + 56;
                tempc = BR.ReadInt32();
                tempr = BR.ReadInt32() - meta.magic - meta.offset;
                this.RawDataChunkInfo = new RawDataOffsetChunk[tempc];
                for (int x = 0; x < tempc; x++)
                {
                    this.RawDataChunkInfo[x] = new RawDataOffsetChunk();
                    BR.BaseStream.Position = tempr + (x * 16) + 6;
                    this.RawDataChunkInfo[x].ChunkSize = BR.ReadUInt16();
                    this.RawDataChunkInfo[x].Size = BR.ReadInt32();
                    if (this.RawDataChunkInfo[x].ChunkSize == 0) { this.RawDataChunkInfo[x].ChunkSize = this.RawDataChunkInfo[x].Size; }
                    this.RawDataChunkInfo[x].Offset = BR.ReadInt32();
                    this.RawDataChunkInfo[x].ChunkCount = this.RawDataChunkInfo[x].Size / this.RawDataChunkInfo[x].ChunkSize;
                }
                if (this.RawDataChunkInfo.Length == 0) { return; }
                #endregion
                #endregion
                #region RawData
                int firstbsp2chunk = 0;
                #region FindFirstPermutationRawData
                for (int x = 0; x < meta.raw.rawChunks.Count; x++)
                {
                    if (((RawDataChunk)meta.raw.rawChunks[x]).rawDataType == RawDataType.bsp2)
                    {
                        firstbsp2chunk = x;
                        break;
                    }
                }
                #endregion
                #region LoadSubMeshInfo
                BR = new BinaryReader(((RawDataChunk)meta.raw.rawChunks[firstbsp2chunk + chunknumber]).MS);
                this.SubMeshInfo = new ModelSubMeshInfo[this.RawDataChunkInfo[0].ChunkCount];
                for (int x = 0; x < this.RawDataChunkInfo[0].ChunkCount; x++)
                {
                    this.SubMeshInfo[x] = new ModelSubMeshInfo();
                    BR.BaseStream.Position = this.HeaderSize + this.RawDataChunkInfo[0].Offset + (x * 72) + 4;
                    this.SubMeshInfo[x].ShaderNumber = BR.ReadUInt16();
                    this.SubMeshInfo[x].IndiceStart = BR.ReadUInt16();
                    this.SubMeshInfo[x].IndiceCount = BR.ReadUInt16();

                }
                #endregion
                #region DetermineChunkNumers
                int indicechunk = 0;
                int verticechunk = 0;
                int uvchunk = 0;
                int normalchunk = 0;
                for (int x = 0; x < RawDataChunkInfo.Length; x++)
                {
                    if (RawDataChunkInfo[x].ChunkSize == 2)
                    {
                        indicechunk = x;
                        break;
                    }
                }
                for (int x = indicechunk; x < RawDataChunkInfo.Length; x++)
                {
                    if (RawDataChunkInfo[x].ChunkCount == 1)
                    {
                        verticechunk = x;
                        uvchunk = x + 1;
                        normalchunk = x + 2;
                        break;
                    }
                }
                #endregion
                #region Faces
                BR.BaseStream.Position = 40;
                this.IndiceCount = BR.ReadUInt16();
                this.Indices = new short[this.IndiceCount];
                BR.BaseStream.Position = this.HeaderSize + this.RawDataChunkInfo[indicechunk].Offset;
                for (int x = 0; x < this.IndiceCount; x++)
                {
                    this.Indices[x] = (short)BR.ReadUInt16();

                }
                #endregion
                #region Vertices
                this.RawDataChunkInfo[verticechunk].ChunkSize = this.RawDataChunkInfo[verticechunk].Size / this.VerticeCount;

                for (int x = 0; x < this.VerticeCount; x++)
                {
                    Vector3 vec = new Vector3();
                    BR.BaseStream.Position = this.HeaderSize + this.RawDataChunkInfo[verticechunk].Offset + (this.RawDataChunkInfo[verticechunk].ChunkSize * x);
                    vec.X = BR.ReadSingle();
                    vec.Y = BR.ReadSingle();
                    vec.Z = BR.ReadSingle();
                    Vertices.Add(vec);
                }
                #endregion
                #region UV's

                this.RawDataChunkInfo[uvchunk].ChunkSize = 8;
                for (int x = 0; x < this.VerticeCount; x++)
                {
                    Vector2 tempuv = new Vector2();
                    BR.BaseStream.Position = this.HeaderSize + this.RawDataChunkInfo[uvchunk].Offset + (this.RawDataChunkInfo[uvchunk].ChunkSize * x);
                    tempuv.X = BR.ReadSingle();
                    //if (this.UVs[x].X > 1) { this.UVs[x].X = this.UVs[x].X - 1; }
                    // if (this.UVs[x].X < 0) { this.UVs[x].X = this.UVs[x].X + 1; }
                    tempuv.Y = BR.ReadSingle();
                    //  if (this.UVs[x].Y > 1) { this.UVs[x].Y = this.UVs[x].Y - 1; }
                    //   if (this.UVs[x].Y < 0) { this.UVs[x].Y = this.UVs[x].Y + 1; }
                    this.UVs.Add(tempuv);
                }
                #endregion

                #region Normals
                this.RawDataChunkInfo[normalchunk].ChunkSize = 12;
                for (int x = 0; x < this.VerticeCount; x++)
                {
                    Vector2 tempuv = new Vector2();
                    BR.BaseStream.Position = this.HeaderSize + this.RawDataChunkInfo[normalchunk].Offset + (this.RawDataChunkInfo[normalchunk].ChunkSize * x);
                    Vector3 normal = Raw.ParsedModel.DecompressNormal(BR.ReadInt32());
                    this.Normals.Add(normal);
                    Vector3 binormal = Raw.ParsedModel.DecompressNormal(BR.ReadInt32());
                    this.Binormals.Add(binormal);
                    Vector3 tangent = Raw.ParsedModel.DecompressNormal(BR.ReadInt32());
                    this.Tangents.Add(tangent);
                }
                #endregion
                int lightmapuvchunk = -1;
                for (int x = normalchunk + 1; x < RawDataChunkInfo.Length; x++)
                {
                    if (RawDataChunkInfo[x].ChunkSize == 3)
                    {
                        lightmapuvchunk = x;
                        break;
                    }
                }

                if (lightmapuvchunk == -1) return;
                RawDataChunkInfo[lightmapuvchunk].ChunkSize = 4;
                for (int x = 0; x < this.VerticeCount; x++)
                {
                    Vector2 tempuv = new Vector2();
                    BR.BaseStream.Position = this.HeaderSize + this.RawDataChunkInfo[lightmapuvchunk].Offset + (this.RawDataChunkInfo[lightmapuvchunk].ChunkSize * x);
                    short testx = BR.ReadInt16();
                    float u = DecompressVertice(Convert.ToSingle(testx), -1, 1);
                    testx = BR.ReadInt16();
                    float v = DecompressVertice(Convert.ToSingle(testx), -1, 1);
                    //   if (u < 0) u += 1;
                    //   if (v < 0) v += 1;
                    Vector2 uv2 = new Vector2(u, v);
                    this.LightMapUVs.Add(uv2);
                }
            }
                #endregion
            #endregion
            #region Classes
            public class BSPPermutationBoundingBoxContainer
            {
                #region Fields
                public float MinX = 0f;
                public float MaxX = 0f;
                public float MinY = 0f;
                public float MaxY = 0f;
                public float MinZ = 0f;
                public float MaxZ = 0f;
                public float MinU = 0f;
                public float MaxU = 0f;
                public float MinV = 0f;
                public float MaxV = 0f;
                #endregion;
                #region Constructor
                public BSPPermutationBoundingBoxContainer(ref BinaryReader BR)
                {

                    MinX = BR.ReadSingle();
                    MaxX = BR.ReadSingle();
                    MinY = BR.ReadSingle();
                    MaxY = BR.ReadSingle();
                    MinZ = BR.ReadSingle();
                    MaxZ = BR.ReadSingle();
                    MinU = BR.ReadSingle();
                    MaxU = BR.ReadSingle();
                    MinV = BR.ReadSingle();
                    MaxV = BR.ReadSingle();
                }
                #endregion;
            }
            #endregion
        }
        public class PermutationPlacement
        {
            public short sceneryIndex;
            public Matrix mat = Matrix.Identity;
            public float scale;

            public PermutationPlacement(ref BinaryReader BR)
            {
                byte[] tmpBuff = BR.ReadBytes(0x58);
                scale = BitConverter.ToSingle(tmpBuff, 0x0);

                mat = Matrix.Identity;
                mat.M11 = BitConverter.ToSingle(tmpBuff, 0x4);
                mat.M12 = BitConverter.ToSingle(tmpBuff, 0x8);
                mat.M13 = BitConverter.ToSingle(tmpBuff, 0xC);
                mat.M14 = 0;

                mat.M21 = BitConverter.ToSingle(tmpBuff, 0x10);
                mat.M22 = BitConverter.ToSingle(tmpBuff, 0x14);
                mat.M23 = BitConverter.ToSingle(tmpBuff, 0x18);
                mat.M24 = 0;

                mat.M31 = BitConverter.ToSingle(tmpBuff, 0x1C);
                mat.M32 = BitConverter.ToSingle(tmpBuff, 0x20);
                mat.M33 = BitConverter.ToSingle(tmpBuff, 0x24);
                mat.M34 = 0;

                mat.M41 = BitConverter.ToSingle(tmpBuff, 0x28);
                mat.M42 = BitConverter.ToSingle(tmpBuff, 0x2C);
                mat.M43 = BitConverter.ToSingle(tmpBuff, 0x30);
                mat.M44 = 1;

                sceneryIndex = BitConverter.ToInt16(tmpBuff, 0x34);
                tmpBuff = null;
            }
        }
        public void LoadPermutations(ref Meta meta, int mapnumber)
        {
            Maps.map[mapnumber].OpenMap(MapTypes.Internal);
            BinaryReader BR = new BinaryReader(meta.MS);
            BR.BaseStream.Position = 328;
            int tempc = BR.ReadInt32();
            int tempr = BR.ReadInt32() - meta.magic - meta.offset;
            BSPPermutationRawDataMetaChunks = new BSPPermutationRawDataMetaChunk[tempc];
            for (int x = 0; x < tempc; x++)
            {
                BSPPermutationRawDataMetaChunks[x] = new BSPPermutationRawDataMetaChunk(tempr + (x * 200), x, ref meta);
            }

            BR.BaseStream.Position = 336;
            tempc = BR.ReadInt32();
            tempr = BR.ReadInt32() - meta.magic - meta.offset;
            PermutationInfo = new PermutationPlacement[tempc];
            for (int x = 0; x < tempc; x++)
            {
                BR.BaseStream.Position = tempr + (x * 88);
                PermutationInfo[x] = new PermutationPlacement(ref BR);

            }







            Maps.map[mapnumber].CloseMap();
            // MessageBox.Show("Test");
        }
    
        #endregion
        #region LightMaps
        public void LoadLightmaps(int mapnumber)
        {
            if (Maps.map[mapnumber].BSP.sbsp[this.BspNumber].LightMap_TagNumber == -1) {
                LightMapBitmap = new System.Drawing.Bitmap[this.BSPRawDataMetaChunks.Length];
                SceneryLightMapBitmap = new System.Drawing.Bitmap[this.PermutationInfo.Length ];
                return; 
            }
            Maps.map[mapnumber].OpenMap(MapTypes.Internal);
            Meta m = new Meta();
            m.ReadMetaFromMap(Maps.map[mapnumber].BSP.sbsp[this.BspNumber].LightMap_TagNumber, mapnumber, false);

            ParsedBitmap pb = new ParsedBitmap(ref m, mapnumber);
            LightMapBitmap = new System.Drawing.Bitmap[this.BSPRawDataMetaChunks.Length];
            for (int x = 0; x < this.BSPRawDataMetaChunks.Length; x++)
            {
                int visualchunk = Maps.map[mapnumber].BSP.sbsp[this.BspNumber].VisualChunk_Bitmap_Index[x];
                if (visualchunk == -1) { LightMapBitmap[x] = null; continue; }
                LightMapBitmap[x] = pb.FindChunkAndDecode(visualchunk, 0, ref m, mapnumber, x,this.BspNumber);
            }


            SceneryLightMapBitmap = new System.Drawing.Bitmap[this.PermutationInfo.Length];
            for (int x = 0; x < this.PermutationInfo.Length; x++)
            {
                int visualchunk = Maps.map[mapnumber].BSP.sbsp[this.BspNumber].SceneryChunk_Bitmap_Index[x];
                if (visualchunk == -1) { SceneryLightMapBitmap[x] = null; continue; }
                SceneryLightMapBitmap[x] = pb.FindChunkAndDecode(visualchunk, 0, ref m, mapnumber, -x - 1, this.BspNumber);
            }
            Maps.map[mapnumber].CloseMap();
        }
        #endregion
        #region IO
        #region OBJ
        public void ExtractModel(string path)
        {
            if (path[path.Length - 1] != '\\') { path += "\\"; }
            #region ExportTextures
            string texturepath = path + "Textures\\";
            Directory.CreateDirectory(texturepath);
            List<string> names = new List<string>();
            string mtllib = this.Name + ".mtl";
            FileStream FS = new FileStream(path + mtllib, FileMode.Create);
            StreamWriter SW = new StreamWriter(FS);

            Panel p = new Panel();
            Renderer.Renderer r = new entity.Renderer.Renderer();
            r.CreateDevice(p);

            for (int x = 0; x < this.Shaders.Shader.Length; x++)
            {
                string[] namesplit = this.Shaders.Shader[x].shaderName.Split('\\');
                string temps = namesplit[namesplit.Length - 1];
                temps = temps.Replace(' ', '_');
                names.Add(temps);
                this.Shaders.Shader[x].MakeTextures(ref r.device);


                try
                {
                    Microsoft.DirectX.Direct3D.TextureLoader.Save(texturepath + temps + ".dds", Microsoft.DirectX.Direct3D.ImageFileFormat.Dds, this.Shaders.Shader[x].MainTexture);
                }
                catch
                {
                }
                if (x > 0)
                {

                    //space

                    SW.WriteLine("");

                }
                SW.WriteLine("newmtl " + temps);
                SW.WriteLine("Ka 1.000000 1.000000 1.000000");
                SW.WriteLine("Kd 1.000000 1.000000 1.000000");
                SW.WriteLine("Ks 1.000000 1.000000 1.000000");
                SW.WriteLine("Ns 0.000000");
                SW.WriteLine("map_Kd .\\Textures\\" + temps + ".dds");
            }
            SW.Close();
            FS.Close();
            #endregion
            #region ExportBSPMeshes
            for (int x = 0; x < this.BSPRawDataMetaChunks.Length; x++)
            {

                if (this.BSPRawDataMetaChunks[x].VerticeCount == 0) { continue; }

                FS = new FileStream(path + this.Name + "[" + x.ToString() + "].obj", FileMode.Create);
                SW = new StreamWriter(FS);
                SW.WriteLine("# ------------------------------------");
                SW.WriteLine("# Halo 2 BSP Mesh - Extracted with Entity");
                SW.WriteLine("# ------------------------------------");
                SW.WriteLine("mtllib " + mtllib);
                for (int y = 0; y < this.BSPRawDataMetaChunks[x].VerticeCount; y++)
                {
                    string temps = "v " + this.BSPRawDataMetaChunks[x].Vertices[y].X.ToString()
                                 + " " + this.BSPRawDataMetaChunks[x].Vertices[y].Y.ToString()
                                 + " " + this.BSPRawDataMetaChunks[x].Vertices[y].Z.ToString();
                    SW.WriteLine(temps);
                }
                SW.WriteLine("# " + this.BSPRawDataMetaChunks[x].Vertices.Count.ToString() + " vertices");
                for (int y = 0; y < this.BSPRawDataMetaChunks[x].VerticeCount; y++)
                {
                    string temps = "vt " + this.BSPRawDataMetaChunks[x].UVs[y].X.ToString()
                                 + " " + this.BSPRawDataMetaChunks[x].UVs[y].Y.ToString();
                    SW.WriteLine(temps);
                }
                SW.WriteLine("# " + this.BSPRawDataMetaChunks[x].Vertices.Count.ToString() + " texture vertices");
                for (int y = 0; y < this.BSPRawDataMetaChunks[x].VerticeCount; y++)
                {
                    string temps = "vn " + this.BSPRawDataMetaChunks[x].Normals[y].X
                                 + " " + this.BSPRawDataMetaChunks[x].Normals[y].Y
                                 + " " + this.BSPRawDataMetaChunks[x].Normals[y].Z;
                    SW.WriteLine(temps);
                }
                SW.WriteLine("# " + this.BSPRawDataMetaChunks[x].Vertices.Count.ToString() + " normals");
                for (int y = 0; y < this.BSPRawDataMetaChunks[x].SubMeshInfo.Length; y++)
                {

                    SW.WriteLine("g 0." + y.ToString());
                    //    SW.WriteLine("s 0." + y.ToString());
                    SW.WriteLine("usemtl  " + names[this.BSPRawDataMetaChunks[x].SubMeshInfo[y].ShaderNumber]);

                    short[] shite = new short[100000];
                    int s = 0;
                    //if (this.BSPRawDataMetaChunks[x].SubMeshInfo[y].IndiceCount == (this.BSPRawDataMetaChunks[x].SubMeshInfo[y].IndiceCount/3)*3)
                    if (this.BSPRawDataMetaChunks[x].FaceCount * 3 != this.BSPRawDataMetaChunks[x].IndiceCount)
                    {

                        int m = this.BSPRawDataMetaChunks[x].SubMeshInfo[y].IndiceStart;




                        bool dir = false;
                        short tempx;
                        short tempy;
                        short tempz;

                        do
                        {
                            //if (mode.EndOfIndices[x][j]>m+2){break;}

                            tempx = this.BSPRawDataMetaChunks[x].Indices[m];
                            tempy = this.BSPRawDataMetaChunks[x].Indices[m + 1];
                            tempz = this.BSPRawDataMetaChunks[x].Indices[m + 2];


                            if (tempx != tempy && tempx != tempz && tempy != tempz)
                            {
                                if (dir == false)
                                {
                                    shite[s] = tempx;
                                    shite[s + 1] = tempy;
                                    shite[s + 2] = tempz;
                                    s += 3;

                                    dir = true;
                                }
                                else
                                {
                                    shite[s] = tempx;
                                    shite[s + 1] = tempz;
                                    shite[s + 2] = tempy;
                                    s += 3;
                                    dir = false;
                                }
                                m += 1;
                            }
                            else
                            {
                                if (dir == true) { dir = false; }
                                else { dir = true; }

                                m += 1;
                            }
                        }
                        while (m < this.BSPRawDataMetaChunks[x].SubMeshInfo[y].IndiceStart + this.BSPRawDataMetaChunks[x].SubMeshInfo[y].IndiceCount - 2);



                    }
                    else
                    {

                        Array.Copy(this.BSPRawDataMetaChunks[x].Indices, this.BSPRawDataMetaChunks[x].SubMeshInfo[y].IndiceStart, shite, 0, this.BSPRawDataMetaChunks[x].SubMeshInfo[y].IndiceCount);
                        s = this.BSPRawDataMetaChunks[x].SubMeshInfo[y].IndiceCount;
                    }


                    for (int xx = 0; xx < s; xx += 3)
                    {
                        string temps = "f " + (shite[xx] + 1) + "/" + (shite[xx] + 1) + "/" + (shite[xx] + 1)
                                    + " " + (shite[xx + 1] + 1) + "/" + (shite[xx + 1] + 1) + "/" + (shite[xx + 1] + 1)
                                    + " " + (shite[xx + 2] + 1) + "/" + (shite[xx + 2] + 1) + "/" + (shite[xx + 2] + 1);
                        SW.WriteLine(temps);
                    }
                    SW.WriteLine("# " + (s / 3) + " elements");

                }
                SW.Close();
                FS.Close();

            }

            #endregion
            #region ExportBSPPermutationMeshes
            for (int x = 0; x < this.BSPPermutationRawDataMetaChunks.Length; x++)
            {

                if (this.BSPPermutationRawDataMetaChunks[x].SubMeshInfo == null) { continue; }

                FS = new FileStream(path + this.Name + "-Permutation[" + x.ToString() + "].obj", FileMode.Create);
                SW = new StreamWriter(FS);
                SW.WriteLine("# ------------------------------------");
                SW.WriteLine("# Halo 2 BSP Permutation Mesh - Extracted with Entity");
                SW.WriteLine("# ------------------------------------");
                SW.WriteLine("mtllib " + mtllib);
                for (int y = 0; y < this.BSPPermutationRawDataMetaChunks[x].VerticeCount; y++)
                {
                    string temps = "v " + this.BSPPermutationRawDataMetaChunks[x].Vertices[y].X.ToString()
                                 + " " + this.BSPPermutationRawDataMetaChunks[x].Vertices[y].Y.ToString()
                                 + " " + this.BSPPermutationRawDataMetaChunks[x].Vertices[y].Z.ToString();
                    SW.WriteLine(temps);
                }
                SW.WriteLine("# " + this.BSPPermutationRawDataMetaChunks[x].Vertices.Count.ToString() + " vertices");
                for (int y = 0; y < this.BSPPermutationRawDataMetaChunks[x].VerticeCount; y++)
                {
                    string temps = "vt " + this.BSPPermutationRawDataMetaChunks[x].UVs[y].X.ToString()
                                 + " " + this.BSPPermutationRawDataMetaChunks[x].UVs[y].Y.ToString();
                    SW.WriteLine(temps);
                }
                SW.WriteLine("# " + this.BSPPermutationRawDataMetaChunks[x].Vertices.Count.ToString() + " texture vertices");
                for (int y = 0; y < this.BSPPermutationRawDataMetaChunks[x].VerticeCount; y++)
                {
                    string temps = "vn " + this.BSPPermutationRawDataMetaChunks[x].Normals[y].X
                                 + " " + this.BSPPermutationRawDataMetaChunks[x].Normals[y].Y
                                 + " " + this.BSPPermutationRawDataMetaChunks[x].Normals[y].Z;
                    SW.WriteLine(temps);
                }
                SW.WriteLine("# " + this.BSPPermutationRawDataMetaChunks[x].Vertices.Count.ToString() + " normals");
                for (int y = 0; y < this.BSPPermutationRawDataMetaChunks[x].SubMeshInfo.Length; y++)
                {

                    SW.WriteLine("g 0." + y.ToString());
                    //  SW.WriteLine("s 0." + y.ToString());
                    SW.WriteLine("usemtl  " + names[this.BSPPermutationRawDataMetaChunks[x].SubMeshInfo[y].ShaderNumber]);

                    short[] shite = new short[100000];
                    int s = 0;
                    if (this.BSPPermutationRawDataMetaChunks[x].FaceCount * 3 != this.BSPPermutationRawDataMetaChunks[x].IndiceCount)
                    {

                        int m = this.BSPPermutationRawDataMetaChunks[x].SubMeshInfo[y].IndiceStart;




                        bool dir = false;
                        short tempx;
                        short tempy;
                        short tempz;

                        do
                        {
                            //if (mode.EndOfIndices[x][j]>m+2){break;}

                            tempx = this.BSPPermutationRawDataMetaChunks[x].Indices[m];
                            tempy = this.BSPPermutationRawDataMetaChunks[x].Indices[m + 1];
                            tempz = this.BSPPermutationRawDataMetaChunks[x].Indices[m + 2];


                            if (tempx != tempy && tempx != tempz && tempy != tempz)
                            {
                                if (dir == false)
                                {
                                    shite[s] = tempx;
                                    shite[s + 1] = tempy;
                                    shite[s + 2] = tempz;
                                    s += 3;

                                    dir = true;
                                }
                                else
                                {
                                    shite[s] = tempx;
                                    shite[s + 1] = tempz;
                                    shite[s + 2] = tempy;
                                    s += 3;
                                    dir = false;
                                }
                                m += 1;
                            }
                            else
                            {
                                if (dir == true) { dir = false; }
                                else { dir = true; }

                                m += 1;
                            }
                        }
                        while (m < this.BSPPermutationRawDataMetaChunks[x].SubMeshInfo[y].IndiceStart + this.BSPPermutationRawDataMetaChunks[x].SubMeshInfo[y].IndiceCount - 2);



                    }
                    else
                    {

                        Array.Copy(this.BSPPermutationRawDataMetaChunks[x].Indices, this.BSPPermutationRawDataMetaChunks[x].SubMeshInfo[y].IndiceStart, shite, 0, this.BSPPermutationRawDataMetaChunks[x].SubMeshInfo[y].IndiceCount);
                        s = this.BSPPermutationRawDataMetaChunks[x].SubMeshInfo[y].IndiceCount;
                    }


                    for (int xx = 0; xx < s; xx += 3)
                    {
                        string temps = "f " + (shite[xx] + 1) + "/" + (shite[xx] + 1) + "/" + (shite[xx] + 1)
                                    + " " + (shite[xx + 1] + 1) + "/" + (shite[xx + 1] + 1) + "/" + (shite[xx + 1] + 1)
                                    + " " + (shite[xx + 2] + 1) + "/" + (shite[xx + 2] + 1) + "/" + (shite[xx + 2] + 1);
                        SW.WriteLine(temps);
                    }
                    SW.WriteLine("# " + (s / 3) + " elements");

                }
                SW.Close();
                FS.Close();

            }

            #endregion


            if (this.UnknownChunks == null) { return; }

            #region ExportUnknown
            for (int x = 0; x < this.UnknownChunks.Length; x++)
            {
                if (this.UnknownChunks[x].IndiceCount == 0)
                { continue; }


                FS = new FileStream(path + this.Name + "-Unknown[" + x.ToString() + "].obj", FileMode.Create);
                SW = new StreamWriter(FS);
                SW.WriteLine("# ------------------------------------");
                SW.WriteLine("# Halo 2 BSP Unknown Mesh - Extracted with Entity");
                SW.WriteLine("# ------------------------------------");

                for (int y = 0; y < this.UnknownChunks[x].VerticeCount; y++)
                {
                    string temps = "v " + this.UnknownChunks[x].Vertices[y].X.ToString()
                                 + " " + this.UnknownChunks[x].Vertices[y].Y.ToString()
                                 + " " + this.UnknownChunks[x].Vertices[y].Z.ToString();
                    SW.WriteLine(temps);
                }
                SW.WriteLine("# " + this.UnknownChunks[x].Vertices.Count.ToString() + " vertices");




                SW.WriteLine("g 0.0");


                short[] shite = new short[100000];
                int s = 0;


                Array.Copy(this.UnknownChunks[x].Indices, 0, shite, 0, UnknownChunks[x].IndiceCount);
                s = this.UnknownChunks[x].IndiceCount;




                for (int xx = 0; xx < s; xx += 3)
                {
                    string temps = "f " + (shite[xx] + 1) + "/" + (shite[xx] + 1) + "/" + (shite[xx] + 1)
                                + " " + (shite[xx + 1] + 1) + "/" + (shite[xx + 1] + 1) + "/" + (shite[xx + 1] + 1)
                                + " " + (shite[xx + 2] + 1) + "/" + (shite[xx + 2] + 1) + "/" + (shite[xx + 2] + 1);
                    SW.WriteLine(temps);
                }
                SW.WriteLine("# " + (s / 3) + " elements");

                SW.Close();
                FS.Close();
            }




            #endregion



        }




        public void ExtractModelAsSingleMesh(string path)
        {
            int ndex = path.LastIndexOf('\\');
            string fuck = path.Remove(ndex + 1);
            #region ExportTextures

            string texturepath = fuck + "Textures\\";
            Directory.CreateDirectory(texturepath);
            List<string> names = new List<string>();
            string mtllib = this.Name + ".mtl";
            FileStream FS = new FileStream(fuck + mtllib, FileMode.Create);
            StreamWriter SW = new StreamWriter(FS);
            Panel p = new Panel();
            Renderer.Renderer r = new entity.Renderer.Renderer();
            r.CreateDevice(p);
            for (int x = 0; x < this.Shaders.Shader.Length; x++)
            {
                string[] namesplit = this.Shaders.Shader[x].shaderName.Split('\\');
                string temps = namesplit[namesplit.Length - 1];
                names.Add(temps);
                this.Shaders.Shader[x].MakeTextures(ref r.device);


                Microsoft.DirectX.Direct3D.TextureLoader.Save(texturepath + temps + ".dds", Microsoft.DirectX.Direct3D.ImageFileFormat.Dds, this.Shaders.Shader[x].MainTexture);
                if (x > 0)
                {
                    SW.WriteLine("");

                }
                SW.WriteLine("newmtl " + temps);
                SW.WriteLine("Ka 1.000000 1.000000 1.000000");
                SW.WriteLine("Kd 1.000000 1.000000 1.000000");
                SW.WriteLine("Ks 1.000000 1.000000 1.000000");
                SW.WriteLine("Ns 0.000000");
                SW.WriteLine("map_Kd .\\Textures\\" + temps + ".dds");
            }
            SW.Close();
            FS.Close();
            #endregion
            #region ExportBSPMeshes
            FS = new FileStream(path, FileMode.Create);
            SW = new StreamWriter(FS);
            SW.WriteLine("# ------------------------------------");
            SW.WriteLine("# Halo 2 BSP Mesh - Extracted with Entity");
            SW.WriteLine("# ------------------------------------");
            SW.WriteLine("mtllib " + mtllib);
            int vertcount = 0;

            for (int x = 0; x < this.BSPRawDataMetaChunks.Length; x++)
            {




                for (int y = 0; y < this.BSPRawDataMetaChunks[x].VerticeCount; y++)
                {
                    string temps = "v " + this.BSPRawDataMetaChunks[x].Vertices[y].X.ToString()
                                 + " " + this.BSPRawDataMetaChunks[x].Vertices[y].Y.ToString()
                                 + " " + this.BSPRawDataMetaChunks[x].Vertices[y].Z.ToString();
                    SW.WriteLine(temps);
                }
                SW.WriteLine("# " + this.BSPRawDataMetaChunks[x].Vertices.Count.ToString() + " vertices");
                for (int y = 0; y < this.BSPRawDataMetaChunks[x].VerticeCount; y++)
                {
                    string temps = "vt " + this.BSPRawDataMetaChunks[x].UVs[y].X.ToString()
                                 + " " + this.BSPRawDataMetaChunks[x].UVs[y].Y.ToString();
                    SW.WriteLine(temps);
                }
                SW.WriteLine("# " + this.BSPRawDataMetaChunks[x].Vertices.Count.ToString() + " texture vertices");
                for (int y = 0; y < this.BSPRawDataMetaChunks[x].VerticeCount; y++)
                {
                    string temps = "vn " + this.BSPRawDataMetaChunks[x].Normals[y].X
                                 + " " + this.BSPRawDataMetaChunks[x].Normals[y].Y
                                 + " " + this.BSPRawDataMetaChunks[x].Normals[y].Z;
                    SW.WriteLine(temps);
                }

                SW.WriteLine("# " + this.BSPRawDataMetaChunks[x].Vertices.Count.ToString() + " normals");
                for (int y = 0; y < this.BSPRawDataMetaChunks[x].SubMeshInfo.Length; y++)
                {

                    SW.WriteLine("g " + x.ToString() + "." + y.ToString());
                    //  SW.WriteLine("s "+x.ToString()+"." + y.ToString());
                    SW.WriteLine("usemtl  " + names[this.BSPRawDataMetaChunks[x].SubMeshInfo[y].ShaderNumber]);

                    short[] shite = new short[100000];
                    int s = 0;
                    if (this.BSPRawDataMetaChunks[x].FaceCount * 3 != this.BSPRawDataMetaChunks[x].IndiceCount)
                    {

                        int m = this.BSPRawDataMetaChunks[x].SubMeshInfo[y].IndiceStart;




                        bool dir = false;
                        short tempx;
                        short tempy;
                        short tempz;

                        do
                        {
                            //if (mode.EndOfIndices[x][j]>m+2){break;}

                            tempx = this.BSPRawDataMetaChunks[x].Indices[m];
                            tempy = this.BSPRawDataMetaChunks[x].Indices[m + 1];
                            tempz = this.BSPRawDataMetaChunks[x].Indices[m + 2];


                            if (tempx != tempy && tempx != tempz && tempy != tempz)
                            {
                                if (dir == false)
                                {
                                    shite[s] = (short)(vertcount + tempx);
                                    shite[s + 1] = (short)(vertcount + tempy);
                                    shite[s + 2] = (short)(vertcount + tempz);
                                    s += 3;

                                    dir = true;
                                }
                                else
                                {
                                    shite[s] = (short)(vertcount + tempx);
                                    shite[s + 1] = (short)(vertcount + tempz);
                                    shite[s + 2] = (short)(vertcount + tempy);
                                    s += 3;
                                    dir = false;
                                }
                                m += 1;
                            }
                            else
                            {
                                if (dir == true) { dir = false; }
                                else { dir = true; }

                                m += 1;
                            }
                        }
                        while (m < this.BSPRawDataMetaChunks[x].SubMeshInfo[y].IndiceStart + this.BSPRawDataMetaChunks[x].SubMeshInfo[y].IndiceCount - 2);



                    }
                    else
                    {
                        for (int u = 0; u < this.BSPRawDataMetaChunks[x].SubMeshInfo[y].IndiceCount; u++)
                        {
                            shite[u] = (short)(vertcount + this.BSPRawDataMetaChunks[x].Indices[this.BSPRawDataMetaChunks[x].SubMeshInfo[y].IndiceStart + u]);
                        }

                        s = this.BSPRawDataMetaChunks[x].SubMeshInfo[y].IndiceCount;
                    }


                    for (int xx = 0; xx < s; xx += 3)
                    {
                        string temps = "f " + (shite[xx] + 1) + "/" + (shite[xx] + 1) + "/" + (shite[xx] + 1)
                                    + " " + (shite[xx + 1] + 1) + "/" + (shite[xx + 1] + 1) + "/" + (shite[xx + 1] + 1)
                                    + " " + (shite[xx + 2] + 1) + "/" + (shite[xx + 2] + 1) + "/" + (shite[xx + 2] + 1);
                        SW.WriteLine(temps);
                    }
                    SW.WriteLine("# " + (s / 3) + " elements");

                }

                vertcount += this.BSPRawDataMetaChunks[x].VerticeCount;
            }

            #endregion
            SW.Close();
            FS.Close();
            return;
            #region ExportBSPPermutationMeshes
            SW.WriteLine("# ------------------------------------");
            SW.WriteLine("# Halo 2 BSP Permutation Mesh");
            SW.WriteLine("# ------------------------------------");
            for (int x = 0; x < this.BSPPermutationRawDataMetaChunks.Length; x++)
            {






                for (int y = 0; y < this.BSPPermutationRawDataMetaChunks[x].VerticeCount; y++)
                {
                    string temps = "v " + this.BSPPermutationRawDataMetaChunks[x].Vertices[y].X.ToString()
                                 + " " + this.BSPPermutationRawDataMetaChunks[x].Vertices[y].Y.ToString()
                                 + " " + this.BSPPermutationRawDataMetaChunks[x].Vertices[y].Z.ToString();
                    SW.WriteLine(temps);
                }
                SW.WriteLine("# " + this.BSPPermutationRawDataMetaChunks[x].Vertices.Count.ToString() + " vertices");
                for (int y = 0; y < this.BSPPermutationRawDataMetaChunks[x].VerticeCount; y++)
                {
                    string temps = "vt " + this.BSPPermutationRawDataMetaChunks[x].UVs[y].X.ToString()
                                 + " " + this.BSPPermutationRawDataMetaChunks[x].UVs[y].Y.ToString();
                    SW.WriteLine(temps);
                }
                SW.WriteLine("# " + this.BSPPermutationRawDataMetaChunks[x].Vertices.Count.ToString() + " texture vertices");
                for (int y = 0; y < this.BSPPermutationRawDataMetaChunks[x].VerticeCount; y++)
                {
                    string temps = "vn " + this.BSPPermutationRawDataMetaChunks[x].Normals[y].X
                                                    + " " + this.BSPPermutationRawDataMetaChunks[x].Normals[y].Y
                                                    + " " + this.BSPPermutationRawDataMetaChunks[x].Normals[y].Z;
                    SW.WriteLine(temps);
                }
                SW.WriteLine("# " + this.BSPPermutationRawDataMetaChunks[x].Vertices.Count.ToString() + " normals");
                for (int y = 0; y < this.BSPPermutationRawDataMetaChunks[x].SubMeshInfo.Length; y++)
                {

                    SW.WriteLine("g 0." + y.ToString());
                    //   SW.WriteLine("s 0." + y.ToString());
                    SW.WriteLine("usemtl  " + names[this.BSPPermutationRawDataMetaChunks[x].SubMeshInfo[y].ShaderNumber]);

                    short[] shite = new short[100000];
                    int s = 0;
                    if (this.BSPPermutationRawDataMetaChunks[x].FaceCount * 3 != this.BSPPermutationRawDataMetaChunks[x].IndiceCount)
                    {

                        int m = this.BSPPermutationRawDataMetaChunks[x].SubMeshInfo[y].IndiceStart;




                        bool dir = false;
                        short tempx;
                        short tempy;
                        short tempz;

                        do
                        {
                            //if (mode.EndOfIndices[x][j]>m+2){break;}

                            tempx = this.BSPPermutationRawDataMetaChunks[x].Indices[m];
                            tempy = this.BSPPermutationRawDataMetaChunks[x].Indices[m + 1];
                            tempz = this.BSPPermutationRawDataMetaChunks[x].Indices[m + 2];


                            if (tempx != tempy && tempx != tempz && tempy != tempz)
                            {
                                if (dir == false)
                                {
                                    shite[s] = (short)(vertcount + tempx);
                                    shite[s + 1] = (short)(vertcount + tempy);
                                    shite[s + 2] = (short)(vertcount + tempz);
                                    s += 3;

                                    dir = true;
                                }
                                else
                                {
                                    shite[s] = (short)(vertcount + tempx);
                                    shite[s + 1] = (short)(vertcount + tempz);
                                    shite[s + 2] = (short)(vertcount + tempy);
                                    s += 3;
                                    dir = false;
                                }
                                m += 1;
                            }
                            else
                            {
                                if (dir == true) { dir = false; }
                                else { dir = true; }

                                m += 1;
                            }
                        }
                        while (m < this.BSPPermutationRawDataMetaChunks[x].SubMeshInfo[y].IndiceStart + this.BSPPermutationRawDataMetaChunks[x].SubMeshInfo[y].IndiceCount - 2);



                    }
                    else
                    {

                        for (int u = 0; u < this.BSPPermutationRawDataMetaChunks[x].SubMeshInfo[y].IndiceCount; u++)
                        {
                            shite[u] = (short)(vertcount + this.BSPPermutationRawDataMetaChunks[x].Indices[this.BSPPermutationRawDataMetaChunks[x].SubMeshInfo[y].IndiceStart + u]);
                        }

                        s = this.BSPPermutationRawDataMetaChunks[x].SubMeshInfo[y].IndiceCount;
                    }


                    for (int xx = 0; xx < s; xx += 3)
                    {
                        string temps = "f " + (shite[xx] + 1) + "/" + (shite[xx] + 1) + "/" + (shite[xx] + 1)
                                    + " " + (shite[xx + 1] + 1) + "/" + (shite[xx + 1] + 1) + "/" + (shite[xx + 1] + 1)
                                    + " " + (shite[xx + 2] + 1) + "/" + (shite[xx + 2] + 1) + "/" + (shite[xx + 2] + 1);
                        SW.WriteLine(temps);
                    }
                    SW.WriteLine("# " + (s / 3) + " elements");
                    vertcount += this.BSPPermutationRawDataMetaChunks[x].VerticeCount;
                }


            }
            SW.Close();
            FS.Close();
            #endregion







        }




        public Meta InjectModel(string FilePath, Meta meta, int mapnumber)
        {

            LoadFromOBJ(FilePath, mapnumber);

            char[] crsr = new char[4] { 'c', 'r', 's', 'r' };
            char[] fklb = new char[4] { 'f', 'k', 'l', 'b' };
      
            for (int x = 0; x < this.BSPRawDataMetaChunks.Length; x++)
            {
                #region Raw Data
                MemoryStream raw = new MemoryStream();
                BinaryWriter newraw = new BinaryWriter(raw);
                BinaryReader oldraw = new BinaryReader(meta.raw.rawChunks[x].MS);

                int newrawsize = 0;
                int rawchunkid = 0;

                #region Write Header
                oldraw.BaseStream.Position = 0;
                newraw.BaseStream.Position = 0;
                newraw.Write(oldraw.ReadBytes(this.BSPRawDataMetaChunks[x].HeaderSize));
                newrawsize += this.BSPRawDataMetaChunks[x].HeaderSize;
                #endregion



                #region Write Submesh Info
                newraw.BaseStream.Position = newrawsize;
                newraw.Write(crsr);
                newrawsize += 4;
                for (int y = 0; y < this.BSPRawDataMetaChunks[x].SubMeshInfo.Length; y++)
                {
                    oldraw.BaseStream.Position = this.BSPRawDataMetaChunks[x].HeaderSize + this.BSPRawDataMetaChunks[x].RawDataChunkInfo[0].Offset;
                    newraw.BaseStream.Position = newrawsize + (y * 72);
                    newraw.Write(oldraw.ReadBytes(72));
                    newraw.BaseStream.Position = newrawsize + (y * 72) + 4;
                    newraw.Write((short)this.BSPRawDataMetaChunks[x].SubMeshInfo[y].ShaderNumber);
                    newraw.Write((short)this.BSPRawDataMetaChunks[x].SubMeshInfo[y].IndiceStart);
                    newraw.Write((short)this.BSPRawDataMetaChunks[x].SubMeshInfo[y].IndiceCount);

                }
                this.BSPRawDataMetaChunks[x].RawDataChunkInfo[0].ChunkCount = this.BSPRawDataMetaChunks[x].SubMeshInfo.Length;
                this.BSPRawDataMetaChunks[x].RawDataChunkInfo[0].Size = this.BSPRawDataMetaChunks[x].SubMeshInfo.Length * 72;
                newrawsize += this.BSPRawDataMetaChunks[x].SubMeshInfo.Length * 72;
                //write count
                newraw.BaseStream.Position = 8;
                newraw.Write(this.BSPRawDataMetaChunks[x].SubMeshInfo.Length);
                #endregion

                #region Write Unknown
                rawchunkid = 1;

                while (newrawsize > 0)
                {
                    if (this.BSPRawDataMetaChunks[x].RawDataChunkInfo[rawchunkid].ChunkSize == 2) { break; }
                    oldraw.BaseStream.Position = this.BSPRawDataMetaChunks[x].HeaderSize + this.BSPRawDataMetaChunks[x].RawDataChunkInfo[rawchunkid].Offset;
                    newraw.BaseStream.Position = newrawsize;
                    newraw.Write(crsr);
                    newrawsize += 4;
                    newraw.Write(oldraw.ReadBytes(this.BSPRawDataMetaChunks[x].RawDataChunkInfo[rawchunkid].Size));
                    this.BSPRawDataMetaChunks[x].RawDataChunkInfo[rawchunkid].Offset = newrawsize - this.BSPRawDataMetaChunks[x].HeaderSize;
                    newrawsize += this.BSPRawDataMetaChunks[x].RawDataChunkInfo[rawchunkid].Size;
                    rawchunkid++;
                }
                #endregion

                #region Write Indices
                int indicechunkid = rawchunkid;
                newraw.BaseStream.Position = newrawsize;
                newraw.Write(crsr);
                newrawsize += 4;
                for (int y = 0; y < this.BSPRawDataMetaChunks[x].Indices.Length; y++)
                {
                    newraw.Write(this.BSPRawDataMetaChunks[x].Indices[y]);

                }
                this.BSPRawDataMetaChunks[x].RawDataChunkInfo[rawchunkid].Offset = newrawsize - this.BSPRawDataMetaChunks[x].HeaderSize;
                this.BSPRawDataMetaChunks[x].RawDataChunkInfo[rawchunkid].ChunkCount = this.BSPRawDataMetaChunks[x].Indices.Length;
                this.BSPRawDataMetaChunks[x].RawDataChunkInfo[rawchunkid].Size = this.BSPRawDataMetaChunks[x].Indices.Length * 2;
                newrawsize += this.BSPRawDataMetaChunks[x].Indices.Length * 2;
                rawchunkid++;
                //indice count
                newraw.BaseStream.Position = 40;
                newraw.Write((short)this.BSPRawDataMetaChunks[x].Indices.Length);
                #endregion



                #region Write Unknown
                int verticechunkid = 0;
                while (newrawsize > 0)
                {
                    if (this.BSPRawDataMetaChunks[x].RawDataChunkInfo[rawchunkid].ChunkCount == 1) { verticechunkid = rawchunkid; break; }

                    oldraw.BaseStream.Position = this.BSPRawDataMetaChunks[x].HeaderSize + this.BSPRawDataMetaChunks[x].RawDataChunkInfo[rawchunkid].Offset;
                    newraw.BaseStream.Position = newrawsize;
                    newraw.Write(crsr);
                    newrawsize += 4;
                    newraw.Write(oldraw.ReadBytes(this.BSPRawDataMetaChunks[x].RawDataChunkInfo[rawchunkid].Size));
                    this.BSPRawDataMetaChunks[x].RawDataChunkInfo[rawchunkid].Offset = newrawsize - this.BSPRawDataMetaChunks[x].HeaderSize;
                    newrawsize += this.BSPRawDataMetaChunks[x].RawDataChunkInfo[rawchunkid].Size;

                    rawchunkid++;
                }
                #endregion

                #region Write Vertices

                newraw.BaseStream.Position = newrawsize;
                newraw.Write(crsr);
                newrawsize += 4;
                this.BSPRawDataMetaChunks[x].RawDataChunkInfo[verticechunkid].ChunkSize = 12;
                newraw.Write(new byte[this.BSPRawDataMetaChunks[x].RawDataChunkInfo[verticechunkid].ChunkSize * this.BSPRawDataMetaChunks[x].VerticeCount]);


                for (int y = 0; y < this.BSPRawDataMetaChunks[x].VerticeCount; y++)
                {
                    newraw.BaseStream.Position = newrawsize + (y * this.BSPRawDataMetaChunks[x].RawDataChunkInfo[verticechunkid].ChunkSize);
                    float vx = this.BSPRawDataMetaChunks[x].Vertices[y].X;
                    float vy = this.BSPRawDataMetaChunks[x].Vertices[y].Y;
                    float vz = this.BSPRawDataMetaChunks[x].Vertices[y].Z;
                    newraw.Write(vx);
                    newraw.Write(vy);
                    newraw.Write(vz);

                }
                this.BSPRawDataMetaChunks[x].RawDataChunkInfo[rawchunkid].Offset = newrawsize - this.BSPRawDataMetaChunks[x].HeaderSize;
                newrawsize += this.BSPRawDataMetaChunks[x].RawDataChunkInfo[verticechunkid].ChunkSize * this.BSPRawDataMetaChunks[x].VerticeCount;
                this.BSPRawDataMetaChunks[x].RawDataChunkInfo[rawchunkid].Size = this.BSPRawDataMetaChunks[x].RawDataChunkInfo[verticechunkid].ChunkSize * this.BSPRawDataMetaChunks[x].VerticeCount;
                this.BSPRawDataMetaChunks[x].RawDataChunkInfo[rawchunkid].ChunkSize = 0;// this.BSPRawDataMetaChunks[x].RawDataChunkInfo[verticechunkid].ChunkSize * this.BSPRawDataMetaChunks[x].VerticeCount; ;
                // newraw.BaseStream.Position = 100;
                // newraw.Write(this.BSPRawDataMetaChunks[x].VerticeCount);

                rawchunkid++;
                #endregion

                #region Write UVs
                int uvchunkid = verticechunkid + 1;
                newraw.BaseStream.Position = newrawsize;
                newraw.Write(crsr);
                newrawsize += 4;
                this.BSPRawDataMetaChunks[x].RawDataChunkInfo[rawchunkid].ChunkSize = 8;
                for (int y = 0; y < this.BSPRawDataMetaChunks[x].VerticeCount; y++)
                {

                    float u = this.BSPRawDataMetaChunks[x].UVs[y].X;
                    float v = this.BSPRawDataMetaChunks[x].UVs[y].Y;
                    newraw.Write(u);
                    newraw.Write(v);

                }
                this.BSPRawDataMetaChunks[x].RawDataChunkInfo[rawchunkid].Offset = newrawsize - this.BSPRawDataMetaChunks[x].HeaderSize;
                this.BSPRawDataMetaChunks[x].RawDataChunkInfo[rawchunkid].ChunkSize = 1;
                this.BSPRawDataMetaChunks[x].RawDataChunkInfo[rawchunkid].Size = 8 * this.BSPRawDataMetaChunks[x].VerticeCount;
                newrawsize += 8 * this.BSPRawDataMetaChunks[x].VerticeCount;
                rawchunkid++;
                #endregion

                #region Write Normals

                newraw.BaseStream.Position = newrawsize;
                newraw.Write(crsr);
                newrawsize += 4;
                this.BSPRawDataMetaChunks[x].RawDataChunkInfo[rawchunkid].ChunkSize = 12;
                for (int y = 0; y < this.BSPRawDataMetaChunks[x].VerticeCount; y++)
                {

                    float u = 1;
                    float v = 1;
                    newraw.Write(u);
                    newraw.Write(v);
                    newraw.Write(v);
                }

                this.BSPRawDataMetaChunks[x].RawDataChunkInfo[rawchunkid].Offset = newrawsize - this.BSPRawDataMetaChunks[x].HeaderSize;
                this.BSPRawDataMetaChunks[x].RawDataChunkInfo[rawchunkid].ChunkSize = 2;
                this.BSPRawDataMetaChunks[x].RawDataChunkInfo[rawchunkid].Size = 12 * this.BSPRawDataMetaChunks[x].VerticeCount;
                newrawsize += 12 * this.BSPRawDataMetaChunks[x].VerticeCount;
                rawchunkid++;
                #endregion

                #region Write Other Stuff Not Yet Implemented


                while (rawchunkid < this.BSPRawDataMetaChunks[x].RawDataChunkInfo.Length)
                {
                    oldraw.BaseStream.Position = this.BSPRawDataMetaChunks[x].HeaderSize + this.BSPRawDataMetaChunks[x].RawDataChunkInfo[rawchunkid].Offset;
                    newraw.BaseStream.Position = newrawsize;
                    newraw.Write(crsr);
                    newrawsize += 4;
                    newraw.Write(oldraw.ReadBytes(this.BSPRawDataMetaChunks[x].RawDataChunkInfo[rawchunkid].Size));
                    this.BSPRawDataMetaChunks[x].RawDataChunkInfo[rawchunkid].Offset = newrawsize - this.BSPRawDataMetaChunks[x].HeaderSize;
                    newrawsize += this.BSPRawDataMetaChunks[x].RawDataChunkInfo[rawchunkid].Size;
                    rawchunkid++;
                }
                #endregion

                //footer
                newraw.Write(fklb);
                newrawsize += 4;

                //raw size
                int rawsize = newrawsize - this.BSPRawDataMetaChunks[x].HeaderSize - 4;
                newraw.BaseStream.Position = 4;
                newraw.Write(rawsize);
                meta.raw.rawChunks[x].MS = raw;
                meta.raw.rawChunks[x].size = newrawsize;
                #endregion

            }
            #region Meta




            #region metachunks



            BinaryWriter BW = new BinaryWriter(meta.MS);

            for (int l = 0; l < this.BSPRawDataMetaChunks.Length; l++)
            {


                BW.BaseStream.Position = this.BSPRawDataMetaChunksOffset + (176 * l);
                short facecount = (short)(this.BSPRawDataMetaChunks[l].FaceCount);
                BW.Write((short)this.BSPRawDataMetaChunks[l].VerticeCount);
                BW.Write(facecount);
                BW.BaseStream.Position = this.BSPRawDataMetaChunksOffset + (176 * l) + 52;
                BW.Write((int)(meta.raw.rawChunks[l].size - this.BSPRawDataMetaChunks[l].HeaderSize - 4));

                for (int h = 0; h < this.BSPRawDataMetaChunks[l].RawDataChunkInfo.Length; h++)
                {


                    BW.BaseStream.Position = this.BSPRawDataMetaChunks[l].rawdatainfooffset + (h * 16) + 6;
                    BW.Write((short)this.BSPRawDataMetaChunks[l].RawDataChunkInfo[h].ChunkSize);
                    BW.Write(this.BSPRawDataMetaChunks[l].RawDataChunkInfo[h].Size);
                    BW.Write(this.BSPRawDataMetaChunks[l].RawDataChunkInfo[h].Offset);


                }



            }

            #endregion

            // fucker.raw = meta.raw;
            #endregion
            return meta;
        }

        public void LoadFromOBJ(string FilePath, int mapnumber)
        {
            if (FilePath[FilePath.Length - 1] != '\\') { FilePath += "\\"; }
            #region Read Material Names
            FileStream FS = new FileStream(FilePath + this.Name + ".mtl", FileMode.Open);
            StreamReader SR = new StreamReader(FS);
            List<string> MaterialNames = new List<string>();
            string temps = "";
            while (temps != null)
            {
                temps = SR.ReadLine();
                if (temps == null) { break; }
                string[] split = temps.Split(' ');
                if (split[0] == "newmtl")
                {
                    if (MaterialNames.IndexOf(split[1]) == -1)
                    {
                        MaterialNames.Add(split[1]);
                    }
                }
            }

            SR.Close();
            FS.Close();
            #endregion

            #region Bounding Box Fields
            float minx = 0;
            float maxx = 0;
            float miny = 0;
            float maxy = 0;
            float minz = 0;
            float maxz = 0;
            float minu = 0;
            float maxu = 0;
            float minv = 0;
            float maxv = 0;
            #endregion

            #region Read OBJ Files
            for (int x = 0; x < this.BSPRawDataMetaChunks.Length; x++)
            {

                #region Fields
                int verticecount = 0;
                int facecount = 0;
                List<Vector3> vertices = new List<Vector3>();
                List<Vector3> normals = new List<Vector3>();
                List<Vector2> uvs = new List<Vector2>();

                List<List<short>> faces = new List<List<short>>();
                List<List<short>> facesuv = new List<List<short>>();
                List<List<short>> facesnormal = new List<List<short>>();
                Hashtable Materials = new Hashtable();
                int groupcount = 0;
                #endregion

                FS = new FileStream(FilePath + this.Name + "[" + x.ToString() + "].obj", FileMode.Open);
                SR = new StreamReader(FS);

                #region ParseFile
                do
                {
                    temps = SR.ReadLine();
                    if (temps == null) { continue; }
                    temps = temps.Replace("  ", " ");
                    string[] tempstrings = temps.Split(',', ' ');
                    switch (tempstrings[0])
                    {
                        #region Vertices
                        case "v":
                            Vector3 tempv = new Vector3();
                            tempv.X = float.Parse(tempstrings[1]);
                            tempv.Y = float.Parse(tempstrings[2]);
                            tempv.Z = float.Parse(tempstrings[3]);
                            if (tempv.X < minx) { minx = tempv.X; }
                            if (tempv.X > maxx) { maxx = tempv.X; }
                            if (tempv.Y < miny) { miny = tempv.Y; }
                            if (tempv.Y > maxy) { maxy = tempv.Y; }
                            if (tempv.Z < minz) { minz = tempv.Z; }
                            if (tempv.Z > maxz) { maxz = tempv.Z; }
                            vertices.Add(tempv);
                            verticecount++;

                            break;
                        #endregion
                        #region Normals
                        case "vn":
                            Vector3 tempvn = new Vector3();
                            tempvn.X = float.Parse(tempstrings[1]);
                            tempvn.Y = float.Parse(tempstrings[2]);
                            tempvn.Z = float.Parse(tempstrings[3]);
                            normals.Add(tempvn);


                            break;
                        #endregion
                        #region UVs
                        case "vt":
                            Vector2 tempv2 = new Vector2();
                            tempv2.X = float.Parse(tempstrings[1]);
                            tempv2.Y = float.Parse(tempstrings[2]);
                            if (tempv2.X < minu) { minu = tempv2.X; }
                            if (tempv2.X > maxu) { maxu = tempv2.X; }
                            if (tempv2.Y < minv) { minv = tempv2.Y; }
                            if (tempv2.Y > maxv) { maxv = tempv2.Y; }
                            uvs.Add(tempv2);
                            verticecount++;
                            break;
                        #endregion
                        #region Group
                        case "g":

                            List<short> templist = new List<short>();
                            List<short> templist2 = new List<short>();
                            List<short> templist3 = new List<short>();
                            faces.Add(templist);
                            facesuv.Add(templist2);
                            facesnormal.Add(templist3);
                            groupcount++;
                            break;
                        #endregion
                        #region Faces
                        case "f":
                            string[] split1 = tempstrings[1].Split('/');
                            string[] split2 = tempstrings[2].Split('/');
                            string[] split3 = tempstrings[3].Split('/');
                            short temp1 = short.Parse(split1[0]);
                            short temp2 = short.Parse(split2[0]);
                            short temp3 = short.Parse(split3[0]);
                            temp1--;
                            temp2--;
                            temp3--;
                            faces[groupcount - 1].Add(temp1);
                            faces[groupcount - 1].Add(temp2);
                            faces[groupcount - 1].Add(temp3);

                            temp1 = short.Parse(split1[1]);
                            temp2 = short.Parse(split2[1]);
                            temp3 = short.Parse(split3[1]);
                            temp1--;
                            temp2--;
                            temp3--;
                            facesuv[groupcount - 1].Add(temp1);
                            facesuv[groupcount - 1].Add(temp2);
                            facesuv[groupcount - 1].Add(temp3);

                            temp1 = short.Parse(split1[2]);
                            temp2 = short.Parse(split2[2]);
                            temp3 = short.Parse(split3[2]);
                            temp1--;
                            temp2--;
                            temp3--;
                            facesnormal[groupcount - 1].Add(temp1);
                            facesnormal[groupcount - 1].Add(temp2);
                            facesnormal[groupcount - 1].Add(temp3);
                            facecount += 3;
                            break;
                        #endregion
                        #region Materials
                        case "usemtl":
                            Materials.Add(groupcount - 1, tempstrings[1]);
                            break;
                        #endregion
                    }
                }
                while (temps != null);
                #endregion


                SR.Close();
                FS.Close();
                int count = 0;
                while (count < faces.Count)
                {
                start:
                    if (faces[count].Count == 0)
                    {
                        faces.RemoveAt(count);
                        facesuv.RemoveAt(count);
                        facesnormal.RemoveAt(count);
                        count = 0;
                        groupcount--;
                        goto start;
                    }
                    count++;
                }




                Renderer.Renderer temprender = new entity.Renderer.Renderer();
                Panel fakepanel = new Panel();
                temprender.CreateDevice(fakepanel);
                List<List<short>> Faces = new List<List<short>>();
                List<List<short>> Facesuv = new List<List<short>>();
                List<List<short>> Facesnormal = new List<List<short>>();
                List<short> newIndices = new List<short>();
                #region Submeshes
                this.BSPRawDataMetaChunks[x].SubMeshInfo = new ParsedModel.RawDataMetaChunk.ModelSubMeshInfo[groupcount];
                int totalindicecount = 0;
                this.BSPRawDataMetaChunks[x].Vertices.Clear();
                this.BSPRawDataMetaChunks[x].UVs.Clear();
                this.BSPRawDataMetaChunks[x].Normals.Clear();
                for (int y = 0; y < groupcount; y++)
                {
                    Application.DoEvents();
                    Faces.Add(new List<short>());
                    for (int h = 0; h < faces[y].Count; h++)
                    {

                        int tempvert = faces[y][h];
                        int tempuv = facesuv[y][h];
                        int tempnorm = facesnormal[y][h];
                        for (int i = 0; i < y + 1; i++)
                        {
                            for (int j = 0; j < faces[i].Count; j++)
                            {
                                if (i == y && j == h) goto gohere1;
                                int tempvert2 = faces[i][j];
                                int tempuv2 = facesuv[i][j];
                                int tempnorm2 = facesnormal[i][j];
                                if (tempvert == tempvert2 && tempuv == tempuv2 && tempnorm == tempnorm2)
                                {
                                    Faces[y].Add(Faces[i][j]);

                                    goto gohere;
                                }

                            }
                        }
                    gohere1:
                        this.BSPRawDataMetaChunks[x].UVs.Add(uvs[facesuv[y][h]]);
                        this.BSPRawDataMetaChunks[x].Vertices.Add(vertices[faces[y][h]]);
                        this.BSPRawDataMetaChunks[x].Normals.Add(normals[facesnormal[y][h]]);

                        Faces[y].Add((short)(this.BSPRawDataMetaChunks[x].Vertices.Count - 1));
                    gohere:
                        int xxxxxx;
                    }
                    #region Make Triangle Strip

                    Mesh m = temprender.MakeMesh(this.BSPRawDataMetaChunks[x].Vertices, Faces[y], this.BSPRawDataMetaChunks[x].UVs);
             
                    int indicecount = 0;
       
                    unsafe
                    {




                        int[] adj = new int[m.NumberFaces * 3];
                        m.GenerateAdjacency(0.0f, adj);
                        int[] test;
                        int[] test2;
                        GraphicsStream oi;
                        Mesh oid = m.Optimize(MeshFlags.OptimizeCompact | MeshFlags.OptimizeIgnoreVerts | MeshFlags.OptimizeAttributeSort, adj, out test, out test2, out oi);



                        IndexBuffer ib = oid.IndexBuffer;


                        ib = Mesh.ConvertMeshSubsetToSingleStrip(oid, 0, MeshFlags.OptimizeCompact | MeshFlags.OptimizeIgnoreVerts | MeshFlags.OptimizeStripeReorder | MeshFlags.IbManaged, out indicecount);
                        GraphicsStream xxxx = ib.Lock(0, 0, LockFlags.None);
                        short* ind = (short*)xxxx.InternalData.ToPointer();

                        for (int z = 0; z < indicecount; z++)
                        {
                            newIndices.Add((short)ind[z]);
                        }


                    }


                    m.Dispose();
                    m = null;

                    #endregion
                    #region SubmeshInfo

                    BSPModel.BSPRawDataMetaChunk.ModelSubMeshInfo submesh = new BSPModel.BSPRawDataMetaChunk.ModelSubMeshInfo();
                    submesh.IndiceStart = totalindicecount;
                    totalindicecount += indicecount;
                    submesh.IndiceCount = indicecount;
                    submesh.ShaderNumber = 0;
                    object tempobject = Materials[y];
                    if (tempobject != null)
                    {
                        int tempint = MaterialNames.IndexOf((string)tempobject);
                        if (tempint != -1)
                        {
                            submesh.ShaderNumber = tempint;
                        }
                    }
                    this.BSPRawDataMetaChunks[x].SubMeshInfo[y] = submesh;
                    #endregion
                }


                #endregion
                this.BSPRawDataMetaChunks[x].FaceCount = facecount / 3;

                //  int tempindcount = newIndices.Count;
                // if (facecount  == tempindcount)
                // {
                //   newIndices.Add(0);
                // }

                this.BSPRawDataMetaChunks[x].Indices = newIndices.ToArray();
                this.BSPRawDataMetaChunks[x].IndiceCount = this.BSPRawDataMetaChunks[x].Indices.Length;


                this.BSPRawDataMetaChunks[x].VerticeCount = this.BSPRawDataMetaChunks[x].Vertices.Count;






            }








            #endregion

        }

        #endregion

        #endregion

        #region Collision
        public class BSPCollision
        {
            public Vector4[] Planes;
            public int PlaneReflexiveCount;
            public int PlaneReflexiveOffset;
            public int PlaneReflexiveTranslation;
            public Vector3[] Vertices;
            public ushort[] Faces;
            public int FaceReflexiveCount;
            public int FaceReflexiveOffset;
            public int FaceReflexiveTranslation;
            public int VerticeReflexiveCount;
            public int VerticeReflexiveOffset;
            public int VerticeReflexiveTranslation;
            public int SurfaceReflexiveCount;
            public int SurfaceReflexiveTranslation;
            #region Constructor
            public BSPCollision(Meta meta)
            {
                BinaryReader BR = new BinaryReader(meta.MS);
                if (Maps.map[meta.mapnumber].HaloVersion == Map.HaloVersionEnum.Halo2)
                {
                    BR.BaseStream.Position = 36;
         
                int tempc = BR.ReadInt32();
                int tempr = BR.ReadInt32() - meta.magic - meta.offset;
                int tempblah = tempr + 44;
                SurfaceReflexiveCount = BR.ReadInt32();
                SurfaceReflexiveTranslation = BR.ReadInt32() - meta.magic - meta.offset;

                //planes
                PlaneReflexiveOffset = tempr + 8;
                BR.BaseStream.Position = PlaneReflexiveOffset;
                PlaneReflexiveCount = BR.ReadInt32();
                PlaneReflexiveTranslation = BR.ReadInt32() - meta.magic - meta.offset;
                Planes = new Vector4[PlaneReflexiveCount];
                for (int x = 0; x < FaceReflexiveCount; x++)
                {
                    Planes[x].X = BR.ReadSingle();
                    Planes[x].Y = BR.ReadSingle();
                    Planes[x].Z = BR.ReadSingle();
                    Planes[x].W = BR.ReadSingle();
                }
                //faces
                FaceReflexiveOffset = tempr + 48;
                BR.BaseStream.Position = FaceReflexiveOffset;
                FaceReflexiveCount = BR.ReadInt32();
                FaceReflexiveTranslation = BR.ReadInt32() - meta.magic - meta.offset;

                Faces = new ushort[FaceReflexiveCount * 3];
                for (int x = 0; x < FaceReflexiveCount; x++)
                {
                    BR.BaseStream.Position = FaceReflexiveTranslation + (x * 12);
                    Faces[(x * 3)] = BR.ReadUInt16();
                    Faces[(x * 3) + 1] = BR.ReadUInt16();
                    Faces[(x * 3) + 2] = Faces[(x * 3) + 1];
                }
                //vertices
                VerticeReflexiveOffset = tempr + 56;
                BR.BaseStream.Position = VerticeReflexiveOffset;
                VerticeReflexiveCount = BR.ReadInt32();
                VerticeReflexiveTranslation = BR.ReadInt32() - meta.magic - meta.offset;
                Vertices = new Vector3[VerticeReflexiveCount];
                for (int x = 0; x < VerticeReflexiveCount; x++)
                {
                    BR.BaseStream.Position = VerticeReflexiveTranslation + (x * 16);
                    Vertices[x].X = BR.ReadSingle();
                    Vertices[x].Y = BR.ReadSingle();
                    Vertices[x].Z = BR.ReadSingle();
                }
            }
            else
            {
                BR.BaseStream.Position = 200;
                int tempc = BR.ReadInt32();
                int tempr = BR.ReadInt32() - meta.magic - meta.offset;
                int tempblah = tempr + 44;
                SurfaceReflexiveCount = BR.ReadInt32();
                SurfaceReflexiveTranslation = BR.ReadInt32() - meta.magic - meta.offset;

                //planes
                PlaneReflexiveOffset = tempr + 12;
                BR.BaseStream.Position = PlaneReflexiveOffset;
                PlaneReflexiveCount = BR.ReadInt32();
                PlaneReflexiveTranslation = BR.ReadInt32() - meta.magic - meta.offset;
                Planes = new Vector4[PlaneReflexiveCount];
                for (int x = 0; x < FaceReflexiveCount; x++)
                {
                    Planes[x].X = BR.ReadSingle();
                    Planes[x].Y = BR.ReadSingle();
                    Planes[x].Z = BR.ReadSingle();
                    Planes[x].W = BR.ReadSingle();
                }
                //faces
                FaceReflexiveOffset = tempr + 72;
                BR.BaseStream.Position = FaceReflexiveOffset;
                FaceReflexiveCount = BR.ReadInt32();
                FaceReflexiveTranslation = BR.ReadInt32() - meta.magic - meta.offset;

                Faces = new ushort[FaceReflexiveCount * 3];
                for (int x = 0; x < FaceReflexiveCount; x++)
                {
                    BR.BaseStream.Position = FaceReflexiveTranslation + (x * 24);
                   
                    Faces[(x * 3)] = (ushort)BR.ReadInt32();
                    Faces[(x * 3) + 1] = (ushort)BR.ReadInt32();
                    Faces[(x * 3) + 2] = Faces[(x * 3) + 1];
                }
                //vertices
                VerticeReflexiveOffset = tempr + 84;
                BR.BaseStream.Position = VerticeReflexiveOffset;
                VerticeReflexiveCount = BR.ReadInt32();
                VerticeReflexiveTranslation = BR.ReadInt32() - meta.magic - meta.offset;
                Vertices = new Vector3[VerticeReflexiveCount];
                for (int x = 0; x < VerticeReflexiveCount; x++)
                {
                    BR.BaseStream.Position = VerticeReflexiveTranslation + (x * 16);
                    Vertices[x].X = BR.ReadSingle();
                    Vertices[x].Y = BR.ReadSingle();
                    Vertices[x].Z = BR.ReadSingle();
                }
            }
            }
            #endregion
            #region Methods
            public void ExtractCollisonMesh(string FilePath)
            {
                FileStream FS = new FileStream(FilePath, FileMode.Create);
                StreamWriter SW = new StreamWriter(FS);
                SW.WriteLine("# ------------------------------------");
                SW.WriteLine("# Halo 2 BSP Collision Mesh - Extracted with Entity");
                SW.WriteLine("# ------------------------------------");
                for (int x = 0; x < Vertices.Length; x++)
                {
                    string temps = "v " + Vertices[x].X.ToString() + " "
                                        + Vertices[x].Y.ToString() + " "
                                        + Vertices[x].Z.ToString();
                    SW.WriteLine(temps);
                }
                for (int x = 0; x < Faces.Length; x += 3)
                {
                    string temps = "f " + (Faces[x] + 1) + " "
                                        + (Faces[x + 1] + 1) + " "
                                        + (Faces[x + 2] + 1);
                    SW.WriteLine(temps);
                }

                SW.Close();
                FS.Close();
            }
            public void InjectCollisonMesh(string FilePath, Meta meta, int mapnumber)
            {
                FileStream FS = new FileStream(FilePath, FileMode.Open);
                StreamReader SR = new StreamReader(FS);
                string temps = "";
                int verticecount = 0;
                int facecount = 0;
                List<Vector3> vertices = new List<Vector3>();
                List<ushort> faces = new List<ushort>();
                do
                {
                    temps = SR.ReadLine();
                    if (temps == null) { continue; }
                    string[] tempstrings = temps.Split(',', ' ');
                    switch (tempstrings[0])
                    {
                        case "v":
                            Vector3 tempv = new Vector3();
                            tempv.X = float.Parse(tempstrings[1]);
                            tempv.Y = float.Parse(tempstrings[2]);
                            tempv.Z = float.Parse(tempstrings[3]);
                            vertices.Add(tempv);
                            verticecount++;

                            break;
                        case "f":
                            string[] split1 = tempstrings[1].Split('/');
                            string[] split2 = tempstrings[2].Split('/');
                            string[] split3 = tempstrings[3].Split('/');
                            ushort temp1 = ushort.Parse(split1[0]);
                            ushort temp2 = ushort.Parse(split2[0]);
                            ushort temp3 = ushort.Parse(split3[0]);
                            temp1--;
                            temp2--;
                            temp3--;
                            faces.Add(temp1);
                            faces.Add(temp2);
                            faces.Add(temp3);
                            facecount += 3;
                            break;



                    }
                }
                while (temps != null);
                SR.Close();
                FS.Close();
                if (this.Vertices.Length >= verticecount && this.Faces.Length >= facecount)
                {

                    Maps.map[mapnumber].OpenMap(MapTypes.Internal);
                    Maps.map[mapnumber].BW.BaseStream.Position = meta.offset + this.VerticeReflexiveOffset;
                    Maps.map[mapnumber].BW.Write(verticecount);
                    for (int x = 0; x < verticecount; x++)
                    {
                        Maps.map[mapnumber].BW.BaseStream.Position = meta.offset + this.VerticeReflexiveTranslation + (x * 16);
                        Maps.map[mapnumber].BW.Write(vertices[x].X);
                        Maps.map[mapnumber].BW.Write(vertices[x].Y);
                        Maps.map[mapnumber].BW.Write(vertices[x].Z);
                    }

                    Maps.map[mapnumber].BW.BaseStream.Position = meta.offset + this.FaceReflexiveOffset;
                    int tempint = (int)(facecount / 3);
                    Maps.map[mapnumber].BW.Write(tempint);
                    for (int x = 0; x < facecount; x += 3)
                    {
                        int fuck = (x / 3);
                        if (Maps.map[meta.mapnumber].HaloVersion == Map.HaloVersionEnum.Halo2)
                        {
                            Maps.map[mapnumber].BW.BaseStream.Position = meta.offset + this.FaceReflexiveTranslation + (fuck * 12);
                            Maps.map[mapnumber].BW.Write(faces[x]);

                            Maps.map[mapnumber].BW.Write(faces[x + 1]);
                        }
                        else
                        {
                            Maps.map[mapnumber].BW.BaseStream.Position = meta.offset + this.FaceReflexiveTranslation + (fuck * 24);
                            Maps.map[mapnumber].BW.Write((int)faces[x]);

                            Maps.map[mapnumber].BW.Write((int)faces[x + 1]);
                        }
                    }
             
                    Maps.map[mapnumber].CloseMap();
                }
            }

            #endregion
        }
        #endregion
        public class BSPShaderContainer : ParsedModel.ShaderContainer
        {
            public BSPShaderContainer()
            {
            }
            public BSPShaderContainer(BSPModel bsp, ref Meta meta, int mapnumber)
                : base()
            {
                switch (Maps.map[meta.mapnumber].HaloVersion)
                {
            
                    case Map.HaloVersionEnum.Halo2:
                        H2BSPShaderContainer(ref meta,mapnumber);
                        break;
                }
            }
           
            public void H2BSPShaderContainer(ref Meta meta, int mapnumber)
            {
                BinaryReader BR = new BinaryReader(meta.MS);
                BR.BaseStream.Position = 180;
                int tempc = BR.ReadInt32();
                int tempr = BR.ReadInt32() - meta.magic - meta.offset;
                this.Shader = new MetaContainers.ShaderInfo[tempc];


                for (int x = 0; x < tempc; x++)
                {
                    BR.BaseStream.Position = tempr + (x * 32) + 12;
                    int tempint = Maps.map[mapnumber].Functions.Meta.FindMetaByID(BR.ReadInt32(), mapnumber);
                    this.Shader[x] = new MetaContainers.ShaderInfo(tempint, mapnumber);
             
                }
            }

        }
        public class BSPDisplayedInfo : ParsedModel.DisplayedInfo
        {
           // private Mesh[] mesh;
            private VertexBuffer[] permvertexBuffer = null;
            private IndexBuffer[] permindexBuffer = null;
            Matrix watermatrix = Matrix.Identity;
            public static void FindDisplayedPieces()
            {

            }

            public static void LoadDirectXTexturesAndBuffers(ref Device device, ref BSPModel bsp)
            {

                BSPDisplayedInfo.CreateVertexBuffers(ref device, ref bsp);
                BSPDisplayedInfo.CreateIndexBuffers(ref device, ref bsp);
                BSPDisplayedInfo.LoadShaderTextures(ref device, ref bsp);
                BSPDisplayedInfo.LoadLightmapTextures(ref device, ref bsp);
            }


            public static void LoadLightmapTextures(ref Device device, ref BSPModel bsp)
            {
                bsp.LightMapTexture = new Texture[bsp.BSPRawDataMetaChunks.Length];
                for (int x = 0; x < bsp.LightMapBitmap.Length; x++)
                    bsp.LightMapTexture[x] = MetaContainers.ShaderInfo.CreateTexture(ref device, bsp.LightMapBitmap[x]);
                bsp.SceneryLightMapTexture = new Texture[bsp.SceneryLightMapBitmap.Length];
                for (int x = 0; x < bsp.SceneryLightMapBitmap.Length; x++)
                    bsp.SceneryLightMapTexture[x] = MetaContainers.ShaderInfo.CreateTexture(ref device, bsp.SceneryLightMapBitmap[x]);
            }
            public static void LoadShaderTextures(ref Device device, ref BSPModel bsp)
            {
                for (int x = 0; x < bsp.Shaders.Shader.Length; x++)
                {

                    bsp.Shaders.Shader[x].MakeTextures(ref device);
                    bsp.Shaders.Shader[x].KillBitmaps();
                }
                //MessageBox.Show("test");
            }
            public static void CreateIndexBuffers(ref Device device, ref BSPModel bsp)
            {
                bsp.Display.indexBuffer = new IndexBuffer[bsp.BSPRawDataMetaChunks.Length];
                for (int x = 0; x < bsp.BSPRawDataMetaChunks.Length; x++)
                {

                    if (bsp.BSPRawDataMetaChunks[x].RawDataChunkInfo.Length == 0)
                    {
                        continue;
                    }
                    bsp.Display.indexBuffer[x] = new IndexBuffer(typeof(short), bsp.BSPRawDataMetaChunks[x].IndiceCount, device, Usage.WriteOnly, Pool.Managed);
                    IndexBuffer ib = (IndexBuffer)bsp.Display.indexBuffer[x];
                    ib.SetData(bsp.BSPRawDataMetaChunks[x].Indices, 0, LockFlags.None);
                    ib.Unlock();
                }
                bsp.Display.permindexBuffer = new IndexBuffer[bsp.BSPPermutationRawDataMetaChunks.Length];
                for (int x = 0; x < bsp.BSPPermutationRawDataMetaChunks.Length; x++)
                {

                    if (bsp.BSPPermutationRawDataMetaChunks[x].RawDataChunkInfo.Length == 0)
                    {
                        continue;
                    }
                    bsp.Display.permindexBuffer[x] = new IndexBuffer(typeof(short), bsp.BSPPermutationRawDataMetaChunks[x].IndiceCount, device, Usage.WriteOnly, Pool.Managed);
                    IndexBuffer ib = (IndexBuffer)bsp.Display.permindexBuffer[x];
                    ib.SetData(bsp.BSPPermutationRawDataMetaChunks[x].Indices, 0, LockFlags.None);
                    ib.Unlock();
                }

            }
            public static void CreateVertexBuffers(ref Device device, ref BSPModel bsp)
            {


                bsp.Display.vertexBuffer = new VertexBuffer[bsp.BSPRawDataMetaChunks.Length];
                for (int x = 0; x < bsp.BSPRawDataMetaChunks.Length; x++)
                {
                    int rawindex = x;
                    if (bsp.BSPRawDataMetaChunks[rawindex].RawDataChunkInfo.Length == 0)
                    {
                        continue;
                    }
                    bsp.Display.vertexBuffer[rawindex] = new VertexBuffer(typeof(HaloBSPVertex), bsp.BSPRawDataMetaChunks[rawindex].VerticeCount, device, Usage.WriteOnly, HaloBSPVertex.FVF, Pool.Managed);
                    HaloBSPVertex[] verts = (HaloBSPVertex[])bsp.Display.vertexBuffer[rawindex].Lock(0, 0); // Lock the buffer (which will return our structs)
                    for (int i = 0; i < bsp.BSPRawDataMetaChunks[rawindex].VerticeCount; i++)
                    {
                        verts[i].Position = new Vector3(bsp.BSPRawDataMetaChunks[rawindex].Vertices[i].X, bsp.BSPRawDataMetaChunks[rawindex].Vertices[i].Y, bsp.BSPRawDataMetaChunks[rawindex].Vertices[i].Z);
                        verts[i].Tu0 = bsp.BSPRawDataMetaChunks[rawindex].UVs[i].X;
                        verts[i].Tv0 = bsp.BSPRawDataMetaChunks[rawindex].UVs[i].Y;
                        verts[i].Normal = bsp.BSPRawDataMetaChunks[rawindex].Normals[i];

                        //verts[i].specular = 1;
                        verts[i].diffuse = 1;

                        verts[i].Tu1 = verts[i].Tu0;
                        verts[i].Tv1 = verts[i].Tv0;
                        if (bsp.BSPRawDataMetaChunks[rawindex].LightMapUVs.Count != 0)
                        {
                            verts[i].Tu2 = bsp.BSPRawDataMetaChunks[rawindex].LightMapUVs[i].X;
                            verts[i].Tv2 = bsp.BSPRawDataMetaChunks[rawindex].LightMapUVs[i].Y;
                        }
                        else
                        {
                            verts[i].Tu2 = verts[i].Tu0;
                            verts[i].Tv2 = verts[i].Tv0;
                        }
                        verts[i].Tu3 = verts[i].Tu0;
                        verts[i].Tv3 = verts[i].Tv0;
                    }
                    bsp.Display.vertexBuffer[rawindex].Unlock();
                }



                bsp.Display.permvertexBuffer = new VertexBuffer[bsp.BSPPermutationRawDataMetaChunks.Length];
                for (int x = 0; x < bsp.BSPPermutationRawDataMetaChunks.Length; x++)
                {
                    int rawindex = x;
                    if (bsp.BSPPermutationRawDataMetaChunks[rawindex].RawDataChunkInfo.Length == 0)
                    {
                        continue;
                    }
                    bsp.Display.permvertexBuffer[rawindex] = new VertexBuffer(typeof(HaloBSPVertex), bsp.BSPPermutationRawDataMetaChunks[rawindex].VerticeCount, device, Usage.WriteOnly, HaloBSPVertex.FVF, Pool.Managed);
                    HaloBSPVertex[] verts = (HaloBSPVertex[])bsp.Display.permvertexBuffer[rawindex].Lock(0, 0); // Lock the buffer (which will return our structs)
                    for (int i = 0; i < bsp.BSPPermutationRawDataMetaChunks[rawindex].VerticeCount; i++)
                    {
                        verts[i].Position = new Vector3(bsp.BSPPermutationRawDataMetaChunks[rawindex].Vertices[i].X, bsp.BSPPermutationRawDataMetaChunks[rawindex].Vertices[i].Y, bsp.BSPPermutationRawDataMetaChunks[rawindex].Vertices[i].Z);
                        verts[i].Tu0 = bsp.BSPPermutationRawDataMetaChunks[rawindex].UVs[i].X;
                        verts[i].Tv0 = bsp.BSPPermutationRawDataMetaChunks[rawindex].UVs[i].Y;
                        verts[i].Normal = bsp.BSPPermutationRawDataMetaChunks[rawindex].Normals[i];

                        //verts[i].specular = 1;
                        verts[i].diffuse = 1;
                        verts[i].Tu1 = verts[i].Tu0;
                        verts[i].Tv1 = verts[i].Tv0;
                        if (bsp.BSPPermutationRawDataMetaChunks[rawindex].LightMapUVs.Count != 0)
                        {
                            verts[i].Tu2 = bsp.BSPPermutationRawDataMetaChunks[rawindex].LightMapUVs[i].X;
                            verts[i].Tv2 = bsp.BSPPermutationRawDataMetaChunks[rawindex].LightMapUVs[i].Y;
                        }
                        else
                        {
                            verts[i].Tu2 = verts[i].Tu0;
                            verts[i].Tv2 = verts[i].Tv0;
                        }
                        verts[i].Tu3 = verts[i].Tu0;
                        verts[i].Tv3 = verts[i].Tv0;

                    }
                    bsp.Display.permvertexBuffer[rawindex].Unlock();
                }
            }


            public static void ResetTextureStates(ref Device device)
            {
                device.SamplerState[0].AddressU = TextureAddress.Wrap;
                device.SamplerState[0].AddressV = TextureAddress.Wrap;
                device.SamplerState[1].AddressU = TextureAddress.Wrap;
                device.SamplerState[1].AddressV = TextureAddress.Wrap;
                device.SamplerState[2].AddressU = TextureAddress.Wrap;
                device.SamplerState[2].AddressV = TextureAddress.Wrap;
                device.SamplerState[3].AddressU = TextureAddress.Wrap;
                device.SamplerState[3].AddressV = TextureAddress.Wrap;

                device.Transform.Texture0 = Matrix.Identity;
                device.TextureState[0].TextureTransform = TextureTransform.Disable;
                device.TextureState[0].TextureCoordinateIndex = 0;
                device.TextureState[0].ColorOperation = TextureOperation.Disable;
                device.SetTexture(0, null);

                device.Transform.Texture1 = Matrix.Identity;
                device.TextureState[1].TextureTransform = TextureTransform.Disable;
                device.TextureState[1].TextureCoordinateIndex = 1;
                device.TextureState[1].ColorOperation = TextureOperation.Disable;
                device.SetTexture(1, null);

                device.Transform.Texture2 = Matrix.Identity;
                device.TextureState[2].TextureTransform = TextureTransform.Disable;
                device.TextureState[2].TextureCoordinateIndex = 2;
                device.TextureState[2].ColorOperation = TextureOperation.Disable;
                device.SetTexture(2, null);

                device.Transform.Texture3 = Matrix.Identity;
                device.TextureState[3].TextureTransform = TextureTransform.Disable;
                device.TextureState[3].TextureCoordinateIndex = 3;
                device.TextureState[3].ColorOperation = TextureOperation.Disable;
                device.SetTexture(3, null);
            }

            public static void Draw(ref Device device, ref BSPModel bsp, bool Textured, ref entity.Renderer.Camera2 cam, DXShader shaderx)
            {
                #region BSPMainMesh
                int count = 0;
                for (int x = 0; x < bsp.BSPRawDataMetaChunks.Length; x++)
                {
                    int rawindex = x;
                    if (bsp.BSPRawDataMetaChunks[rawindex].RawDataChunkInfo.Length == 0)
                    {
                        continue;
                    }



                    device.SetStreamSource(0, bsp.Display.vertexBuffer[rawindex], 0);
                    device.VertexFormat = HaloBSPVertex.FVF;
                    device.Indices = bsp.Display.indexBuffer[rawindex];
                    for (int xx = 0; xx < bsp.BSPRawDataMetaChunks[rawindex].SubMeshInfo.Length; xx++)
                    {
                        ResetTextureStates(ref device);
                        int tempshade = bsp.BSPRawDataMetaChunks[rawindex].SubMeshInfo[xx].ShaderNumber;
                        #region AlphaBlending
                        Renderer.Renderer.SetAlphaBlending(bsp.Shaders.Shader[tempshade].Alpha, ref device);
                        #endregion
                        #region ChooseTexture
                        if (Textured == true)
                        {

                            if (bsp.Shaders.Shader[tempshade].MainName.IndexOf("water!") != -1)
                            {
                                for (int u = 0; u < bsp.Shaders.Shader[tempshade].BitmapTextures.Length; u++)
                                {
                                    if (bsp.Shaders.Shader[tempshade].BitmapNames[u].IndexOf("reflection") != -1 || bsp.Shaders.Shader[tempshade].BitmapNames[u].IndexOf("mask") != -1)
                                    {

                                        device.SetTexture(0, bsp.Shaders.Shader[tempshade].BitmapTextures[u]);


                                        if (bsp.Shaders.Shader[tempshade].MainName.IndexOf("mip") != -1)
                                        {
                                            device.SetTexture(0, bsp.Shaders.Shader[tempshade].BitmapTextures[u]);
                                            device.SetTexture(1, bsp.Shaders.Shader[tempshade].BitmapTextures[u]);
                                            //  

                                        }

                                        else
                                        {
                                            device.SetTexture(1, bsp.Shaders.Shader[tempshade].MainTexture);

                                        }
                                        device.TextureState[0].ColorOperation = TextureOperation.BumpEnvironmentMap;
                                        device.TextureState[0].ColorArgument1 = TextureArgument.TextureColor;
                                        device.TextureState[0].ColorArgument2 = TextureArgument.Current;
                                        float r = 0.07F;
                                        device.TextureState[0].BumpEnvironmentMaterial00 = r * (float)Math.Cos((Environment.TickCount / 4) * 11.0F);
                                        device.TextureState[0].BumpEnvironmentMaterial01 = -r * (float)Math.Sin((Environment.TickCount / 4) * 11.0F);
                                        device.TextureState[0].BumpEnvironmentMaterial10 = r * (float)Math.Sin((Environment.TickCount / 4) * 11.0F);
                                        device.TextureState[0].BumpEnvironmentMaterial11 = r * (float)Math.Cos((Environment.TickCount / 4) * 11.0F);


                                        device.TextureState[1].ColorOperation = TextureOperation.Modulate;
                                        device.TextureState[1].ColorArgument1 = TextureArgument.TextureColor;
                                        device.TextureState[1].ColorArgument2 = TextureArgument.Current;

                                        if (bsp.LightMapTexture[x] != null && bsp.BSPRawDataMetaChunks[x].LightMapUVs.Count != 0)
                                        {
                                            //    device.SetTexture(2, bsp.LightMapTexture[x]);
                                            //   device.TextureState[2].ColorOperation = TextureOperation.Modulate2X;
                                            //   device.TextureState[2].ColorArgument1 = TextureArgument.TextureColor;
                                            //   device.TextureState[2].ColorArgument2 = TextureArgument.Current;


                                        }

                                        break;
                                    }
                                    if (u == bsp.Shaders.Shader[tempshade].BitmapTextures.Length - 1)
                                    {
                                        device.SetTexture(0, bsp.Shaders.Shader[tempshade].MainTexture);
                                        device.TextureState[0].ColorOperation = TextureOperation.Modulate;
                                        device.TextureState[0].ColorArgument1 = TextureArgument.TextureColor;
                                        device.TextureState[0].ColorArgument2 = TextureArgument.Current;
                                    }
                                }
                            }
                            else if (bsp.Shaders.Shader[tempshade].MainName.IndexOf("ground!") != -1)
                            {



                                if (bsp.Shaders.Shader[tempshade].MainName.IndexOf("unwrapped") != -1)
                                {
                                    // device.SamplerState[0].AddressU = TextureAddress.Clamp;
                                    //  device.SamplerState[0].AddressV = TextureAddress.Clamp;
                                }


                                device.SetTexture(0, bsp.Shaders.Shader[tempshade].MainTexture);
                                device.TextureState[0].ColorOperation = TextureOperation.Modulate;
                                device.TextureState[0].ColorArgument1 = TextureArgument.TextureColor;
                                device.TextureState[0].ColorArgument2 = TextureArgument.Current;
                                device.TextureState[0].AlphaOperation = TextureOperation.SelectArg1;
                                device.TextureState[0].AlphaArgument1 = TextureArgument.TextureColor;
                                //  device.TextureState[0].ResultArgument = TextureArgument.Current;
                                device.TextureState[1].TextureTransform = TextureTransform.Count2;
                                device.Transform.Texture1 = Matrix.Scaling(bsp.Shaders.Shader[tempshade].primarydetailuscale, bsp.Shaders.Shader[tempshade].primarydetailvscale, bsp.Shaders.Shader[tempshade].primarydetailwscale);
                                // device.Transform.Texture1 = Matrix.Scaling(bsp.Shaders.Shader[tempshade].secondarydetailuscale, bsp.Shaders.Shader[tempshade].secondarydetailvscale, bsp.Shaders.Shader[tempshade].secondarydetailwscale);
                                device.SetTexture(1, bsp.Shaders.Shader[tempshade].BitmapTextures[2]);
                                device.TextureState[1].ColorOperation = TextureOperation.Modulate2X;
                                device.TextureState[1].ColorArgument1 = TextureArgument.TextureColor;
                                device.TextureState[1].ColorArgument2 = TextureArgument.Current;
                                device.TextureState[1].AlphaOperation = TextureOperation.Add;
                                device.TextureState[1].AlphaArgument1 = TextureArgument.Current;


                                device.TextureState[2].TextureTransform = TextureTransform.Count2;
                                device.Transform.Texture2 = Matrix.Scaling(bsp.Shaders.Shader[tempshade].secondarydetailuscale, bsp.Shaders.Shader[tempshade].secondarydetailvscale, bsp.Shaders.Shader[tempshade].secondarydetailwscale);
                                device.SetTexture(2, bsp.Shaders.Shader[tempshade].BitmapTextures[3]);
                                device.TextureState[2].ColorOperation = TextureOperation.Modulate2X;
                                device.TextureState[2].ColorArgument1 = TextureArgument.TextureColor;
                                device.TextureState[2].ColorArgument2 = TextureArgument.Current;
                                device.TextureState[2].TextureCoordinateIndex = 3;




                                if (bsp.LightMapTexture[x] != null && bsp.BSPRawDataMetaChunks[x].LightMapUVs.Count != 0)
                                {
                                    device.SetTexture(3, bsp.LightMapTexture[x]);
                                    device.TextureState[3].ColorOperation = TextureOperation.Modulate2X;
                                    device.TextureState[3].ColorArgument1 = TextureArgument.TextureColor;
                                    device.TextureState[3].ColorArgument2 = TextureArgument.Current;
                                    device.TextureState[3].TextureCoordinateIndex = 2;

                                }
                              
                            }






                            else
                            {


                                device.SetTexture(0, bsp.Shaders.Shader[tempshade].MainTexture);
                                device.TextureState[0].ColorOperation = TextureOperation.Modulate;
                                device.TextureState[0].ColorArgument1 = TextureArgument.TextureColor;
                                device.TextureState[0].ColorArgument2 = TextureArgument.Current;
                                device.TextureState[0].AlphaOperation = TextureOperation.BlendTextureAlpha;
                                device.TextureState[0].AlphaArgument1 = TextureArgument.TextureColor;
                                device.TextureState[0].AlphaArgument2 = TextureArgument.Diffuse;
                                /*
                                            if (bsp.Shaders.Shader[tempshade].microdetailName != null)
                                            {

                                                device.TextureState[1].TextureTransform = TextureTransform.Count2;
                                                device.Transform.Texture1 = Matrix.Scaling(bsp.Shaders.Shader[tempshade].microdetailuscale, bsp.Shaders.Shader[tempshade].microdetailvscale, bsp.Shaders.Shader[tempshade].microdetailwscale);
                                                device.SetTexture(1, bsp.Shaders.Shader[tempshade].microdetailTexture);
                                                device.TextureState[1].ColorOperation = TextureOperation.Modulate2X;
                                                device.TextureState[1].ColorArgument1 = TextureArgument.TextureColor;
                                                device.TextureState[1].ColorArgument2 = TextureArgument.Current;
                                                device.TextureState[1].AlphaOperation = TextureOperation.BlendCurrentAlpha;
                                                device.TextureState[1].AlphaArgument1 = TextureArgument.Current;
                                            }

                                          
                                if (bsp.Shaders.Shader[tempshade].primarydetailTexture != null)
                              {

                                  device.TextureState[1].TextureTransform = TextureTransform.Count2;
                                  device.Transform.Texture1 = Matrix.Scaling(bsp.Shaders.Shader[tempshade].primarydetailuscale, bsp.Shaders.Shader[tempshade].primarydetailvscale, bsp.Shaders.Shader[tempshade].primarydetailwscale);
                               device.SetTexture(1, bsp.Shaders.Shader[tempshade].primarydetailTexture);
                               device.TextureState[1].ColorOperation = TextureOperation.Lerp;
                               device.TextureState[1].ColorArgument1 = TextureArgument.TextureColor;
                                  device.TextureState[1].ColorArgument2 = TextureArgument.Current;
                                 device.TextureState[1].AlphaOperation = TextureOperation.Modulate;
                                 device.TextureState[1].AlphaArgument1 = TextureArgument.TextureColor;
                              }
                           
                                if (bsp.Shaders.Shader[tempshade].secondarydetailTexture != null)
                                {
                      
                                    device.TextureState[2].TextureTransform = TextureTransform.Count2;
                                    device.Transform.Texture2 = Matrix.Scaling(bsp.Shaders.Shader[tempshade].secondarydetailuscale, bsp.Shaders.Shader[tempshade].secondarydetailvscale, bsp.Shaders.Shader[tempshade].secondarydetailwscale);
                                    device.SetTexture(2, bsp.Shaders.Shader[tempshade].secondarydetailTexture);
                                    device.TextureState[2].ColorOperation = TextureOperation.Lerp;
                                    device.TextureState[2].ColorArgument1 = TextureArgument.TextureColor;
                                    device.TextureState[2].ColorArgument2 = TextureArgument.Current;
                                    device.TextureState[2].TextureCoordinateIndex = 3;
                                    //device.TextureState[2].AlphaOperation = TextureOperation.BlendCurrentAlpha;
                                    //device.TextureState[1].AlphaArgument1 = TextureArgument.Current;
                   
                                }




                 */


                                if (bsp.LightMapTexture[x] != null && bsp.BSPRawDataMetaChunks[x].LightMapUVs.Count != 0)
                                                                {
                                                                    device.SetTexture(1, bsp.LightMapTexture[x]);
                                                                    device.TextureState[1].ColorOperation = TextureOperation.Modulate2X;
                                                                    device.TextureState[1].ColorArgument1 = TextureArgument.TextureColor;
                                                                    device.TextureState[1].ColorArgument2 = TextureArgument.Current;
                                                                    device.TextureState[1].TextureCoordinateIndex = 2;

                                                                }
                 
                                                      
                            }







                        }


                        else
                        {
                            device.SetTexture(0, null);
                            device.SetTexture(1, null);
                        }
                        #endregion
                        #region TextureArguments
                        #endregion
                        PrimitiveType pt;
                        // if (bsp.BSPRawDataMetaChunks[rawindex].SubMeshInfo[xx].IndiceCount != (bsp.BSPRawDataMetaChunks[rawindex].SubMeshInfo[xx].IndiceCount/3)*3)
                        if (bsp.BSPRawDataMetaChunks[rawindex].FaceCount * 3 != bsp.BSPRawDataMetaChunks[rawindex].IndiceCount)
                        {
                            pt = PrimitiveType.TriangleStrip;
                            device.RenderState.FillMode = FillMode.Solid;
                            Material mm = new Material();
                            mm.Diffuse = System.Drawing.Color.White;
                            mm.Ambient = System.Drawing.Color.White;
                            //     if (x == 1) { mm.Diffuse = System.Drawing.Color.Red; mm.Ambient = System.Drawing.Color.Red; }

                            device.Material = mm;
                            device.DrawIndexedPrimitives(
                                pt,
                                0,
                                0,
                                bsp.BSPRawDataMetaChunks[rawindex].VerticeCount,
                                bsp.BSPRawDataMetaChunks[rawindex].SubMeshInfo[xx].IndiceStart,
                                bsp.BSPRawDataMetaChunks[rawindex].SubMeshInfo[xx].IndiceCount
                             );
                        }
                        else
                        {
                            count++;
                            pt = PrimitiveType.TriangleList;

                            device.RenderState.FillMode = FillMode.Solid;
                            Material mm = new Material();
                            mm.Diffuse = System.Drawing.Color.White;
                            mm.Ambient = System.Drawing.Color.White;
                            //   if (x == 1) { mm.Diffuse = System.Drawing.Color.Red; mm.Ambient = System.Drawing.Color.Red; }
                            device.Material = mm;
                            device.DrawIndexedPrimitives(
                                pt,
                                0,
                                0,
                                bsp.BSPRawDataMetaChunks[rawindex].VerticeCount,
                                bsp.BSPRawDataMetaChunks[rawindex].SubMeshInfo[xx].IndiceStart,
                                bsp.BSPRawDataMetaChunks[rawindex].SubMeshInfo[xx].IndiceCount / 3
                             );
                        }














                        #region RenderToScene


                        #region WireFrame
                        if (Textured == false)
                        {
                            Material m = new Material();
                            m.Diffuse = System.Drawing.Color.White;
                            m.Ambient = System.Drawing.Color.White;
                            device.Material = m;
                            device.RenderState.FillMode = FillMode.WireFrame;
                            device.DrawIndexedPrimitives(
                                pt,
                                0,
                                0,
                                bsp.BSPRawDataMetaChunks[rawindex].VerticeCount,
                                bsp.BSPRawDataMetaChunks[rawindex].SubMeshInfo[xx].IndiceStart,
                                bsp.BSPRawDataMetaChunks[rawindex].SubMeshInfo[xx].IndiceCount / 3);

                        }
                        #endregion
                        #endregion

                    }
                }
                #endregion
                #region BSPPermutationMesh
                for (int x = 0; x < bsp.PermutationInfo.Length; x++)
                {

                    int rawindex = bsp.PermutationInfo[x].sceneryIndex;
                    if (bsp.BSPPermutationRawDataMetaChunks[rawindex].RawDataChunkInfo.Length == 0)
                    {
                        continue;
                    }

                    device.Transform.World = bsp.PermutationInfo[x].mat;
                    device.SetStreamSource(0, bsp.Display.permvertexBuffer[rawindex], 0);
                    device.VertexFormat = Raw.HaloVertex.FVF;
                    device.Indices = bsp.Display.permindexBuffer[rawindex];
                    for (int xx = 0; xx < bsp.BSPPermutationRawDataMetaChunks[rawindex].SubMeshInfo.Length; xx++)
                    {
                        ResetTextureStates(ref device);
                        int tempshade = bsp.BSPPermutationRawDataMetaChunks[rawindex].SubMeshInfo[xx].ShaderNumber;
                        #region AlphaBlending
                        if (bsp.Shaders.Shader[tempshade].Alpha == MetaContainers.ShaderInfo.AlphaType.AlphaBlend)
                        {
                            device.RenderState.AlphaBlendEnable = true;
                            device.RenderState.AlphaTestEnable = false;
                        }
                        else if (bsp.Shaders.Shader[tempshade].Alpha == MetaContainers.ShaderInfo.AlphaType.AlphaTest)
                        {
                            device.RenderState.AlphaBlendEnable = false;
                            device.RenderState.AlphaTestEnable = true;
                        }
                        else
                        {
                            device.RenderState.AlphaBlendEnable = false;
                            device.RenderState.AlphaTestEnable = false;
                        }
                        #endregion
                        #region ChooseTexture
                     





                            device.SetTexture(0, bsp.Shaders.Shader[tempshade].MainTexture);


                            device.TextureState[0].ColorOperation = TextureOperation.Modulate;
                            device.TextureState[0].ColorArgument1 = TextureArgument.TextureColor;
                            device.TextureState[0].ColorArgument2 = TextureArgument.Current;



                            if (bsp.SceneryLightMapTexture[x] != null && bsp.BSPPermutationRawDataMetaChunks[rawindex].LightMapUVs.Count != 0)
                            {
                                device.SetTexture(1, bsp.SceneryLightMapTexture[x]);
                                device.TextureState[1].ColorOperation = TextureOperation.Modulate;
                                device.TextureState[1].ColorArgument1 = TextureArgument.TextureColor;
                                device.TextureState[1].ColorArgument2 = TextureArgument.Current;
                                device.TextureState[1].TextureCoordinateIndex = 2;

                            }
                            else
                            {
                                device.TextureState[0].ColorOperation = TextureOperation.Disable;
                                // device.TextureState[0].ColorOperation = TextureOperation.Modulate2X;
                                // device.TextureState[0].ColorArgument1 = TextureArgument.TextureColor;
                                // device.TextureState[0].ColorArgument2 = TextureArgument.Current;
                            }


                  
                        #endregion
                        #region TextureArguments

                        #endregion
                        #region RenderToScene
                        PrimitiveType pt;
                        pt = PrimitiveType.TriangleList;
                        device.RenderState.FillMode = FillMode.Solid;
                        Material mm = new Material();
                        mm.Diffuse = System.Drawing.Color.White;
                        mm.Ambient = System.Drawing.Color.White;
                        device.Material = mm;
                        device.DrawIndexedPrimitives(
                             pt,
                             0,
                             0,
                             bsp.BSPPermutationRawDataMetaChunks[rawindex].VerticeCount,
                             bsp.BSPPermutationRawDataMetaChunks[rawindex].SubMeshInfo[xx].IndiceStart,
                             bsp.BSPPermutationRawDataMetaChunks[rawindex].SubMeshInfo[xx].IndiceCount / 3
                          );
                        #region WireFrame
                        if (Textured == false)
                        {
                            Material m = new Material();
                            m.Diffuse = System.Drawing.Color.Red;
                            m.Ambient = System.Drawing.Color.Red;
                            device.Material = m;
                            device.RenderState.FillMode = FillMode.WireFrame;
                            device.DrawIndexedPrimitives(
                                pt,
                                0,
                                0,
                                bsp.BSPPermutationRawDataMetaChunks[rawindex].VerticeCount,
                                bsp.BSPPermutationRawDataMetaChunks[rawindex].SubMeshInfo[xx].IndiceStart,
                                bsp.BSPPermutationRawDataMetaChunks[rawindex].SubMeshInfo[xx].IndiceCount / 3
                             );
                        }
                        #endregion
                        #endregion
                    }
                }
                #endregion
                ResetTextureStates(ref device);
            }

            public static void DrawCluster(int rawindex, System.Drawing.Color outlinecolor, ref Device device, ref BSPModel bsp, bool Textured, bool WireFrame, ref entity.Renderer.Camera2 cam)
            {
                #region BSPMainMesh



                if (bsp.BSPRawDataMetaChunks[rawindex].RawDataChunkInfo.Length == 0)
                {
                    return;
                }

                //   Vector3 tempv = new Vector3();
                //   tempv.X = bsp.Spawns.Spawn[x].X;
                //   tempv.Y = bsp.Spawns.Spawn[x].Y;
                //  tempv.Z = bsp.Spawns.Spawn[x].Z;
                //   if (!cam.SphereInFrustum(tempv, 10f)) { continue; }

                device.SetStreamSource(0, bsp.Display.vertexBuffer[rawindex], 0);
                device.VertexFormat = CustomVertex.PositionNormalTextured.Format;
                device.Indices = bsp.Display.indexBuffer[rawindex];
                for (int xx = 0; xx < bsp.BSPRawDataMetaChunks[rawindex].SubMeshInfo.Length; xx++)
                {
                    int tempshade = bsp.BSPRawDataMetaChunks[rawindex].SubMeshInfo[xx].ShaderNumber;
                    #region AlphaBlending
                    if (bsp.Shaders.Shader[tempshade].Alpha == MetaContainers.ShaderInfo.AlphaType.AlphaBlend)
                    {
                        //  MessageBox.Show("test");
                        device.RenderState.AlphaBlendEnable = true;
                        device.RenderState.AlphaTestEnable = false;
                    }
                    else if (bsp.Shaders.Shader[tempshade].Alpha == MetaContainers.ShaderInfo.AlphaType.AlphaTest)
                    {
                        // MessageBox.Show("test2");
                        device.RenderState.AlphaBlendEnable = false;
                        device.RenderState.AlphaTestEnable = true;
                    }
                    else
                    {
                        device.RenderState.AlphaBlendEnable = false;
                        device.RenderState.AlphaTestEnable = false;
                    }
                    #endregion
                    #region ChooseTexture
                    if (Textured == true)
                    {
                        // device.TextureState[0] = device.TextureState[2];
                        device.SetTexture(0, bsp.Shaders.Shader[tempshade].MainTexture);

                    }
                    else
                    {
                        device.SetTexture(0, null);
                    }
                    #endregion
                    #region TextureArguments

                    #endregion

                    #region RenderToScene
                    PrimitiveType pt;
                    pt = PrimitiveType.TriangleList;
                    device.RenderState.FillMode = FillMode.Solid;
                    Material mm = new Material();
                    mm.Diffuse = System.Drawing.Color.White;
                    mm.Ambient = System.Drawing.Color.White;
                    device.Material = mm;
                    device.DrawIndexedPrimitives(
                        pt,
                        0,
                        0,
                        bsp.BSPRawDataMetaChunks[rawindex].VerticeCount,
                        bsp.BSPRawDataMetaChunks[rawindex].SubMeshInfo[xx].IndiceStart,
                        bsp.BSPRawDataMetaChunks[rawindex].SubMeshInfo[xx].IndiceCount / 3
                     );

                    #region WireFrame
                    if (WireFrame == true)
                    {
                        Material m = new Material();
                        m.Diffuse = outlinecolor;
                        m.Ambient = outlinecolor;
                        device.Material = m;
                        device.RenderState.FillMode = FillMode.WireFrame;
                        device.DrawIndexedPrimitives(
                            pt,
                            0,
                            0,
                            bsp.BSPRawDataMetaChunks[rawindex].VerticeCount,
                            bsp.BSPRawDataMetaChunks[rawindex].SubMeshInfo[xx].IndiceStart,
                            bsp.BSPRawDataMetaChunks[rawindex].SubMeshInfo[xx].IndiceCount / 3);

                    }
                    #endregion
                    #endregion
                }

                #endregion

            }

        }
        
    }

}
