using System;
using System.Drawing;
using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.DirectInput;
using Microsoft.DirectX.Direct3D;
using Direct3D = Microsoft.DirectX.Direct3D;

namespace entity
{

    public class PortalViewer : Form
    {
        // Our global variables for this project
        Renderer.Renderer render = new entity.Renderer.Renderer();


        int portal = 0;

       
        entity.Renderer.Camera2 cam;
        BSP.BSPModel bsp;
        BSP.PortalContainer pc;
        BSP.Portal[] portals;

        public PortalViewer(ref Meta meta,int map)
        {
            // Set the initial size of our form
            this.ClientSize = new System.Drawing.Size(400, 300);
            // And its caption
            this.Text = "Model Viewer";
     
            portals = BSP.Portals.GetPortals(ref meta, map);

            this.MouseDown += new MouseEventHandler(ModelViewer_MouseDown);
            this.MouseMove += new System.Windows.Forms.MouseEventHandler(this.ModelViewer_MouseDownx);
            this.MouseUp += new System.Windows.Forms.MouseEventHandler(this.ModelViewer_MouseUp);

            portals = BSP.Portals.GetPortals(ref meta, map);

            bsp = new entity.BSP.BSPModel(ref meta, map);


            
            Main();
        }

        void ModelViewer_MouseDown(object sender, MouseEventArgs e)
        {
            cam.oldx = e.X;
            cam.oldy = e.Y;
        }

        public bool InitializeGraphics()
        {
            //try
            //{

            render.CreateDevice(this);

            pc = new entity.BSP.PortalContainer(portals, ref render.device);
            BSP.BSPModel.BSPDisplayedInfo.CreateVertexBuffers(ref render.device, ref bsp);
            BSP.BSPModel.BSPDisplayedInfo.CreateIndexBuffers(ref render.device, ref bsp);
            BSP.BSPModel.BSPDisplayedInfo.LoadShaderTextures(ref render.device, ref bsp);
            cam = new entity.Renderer.Camera2(this);
            cam.speed = 0.005f;
            // cam.fixedrotation = true;
            return true;
            //}
            //catch (DirectXException)
            //{
            // Catch any errors and return a failure
            //	return false;
            //	}
        }






        private void SetupMatrices()
        {
            render.device.Transform.World = Matrix.Identity;
            render.device.Transform.View = Matrix.LookAtRH(cam.Position, cam.LookAt, cam.UpVector);
            render.device.Transform.Projection = Matrix.PerspectiveFovRH(1.33f, 1.0f, 0.1f, 10000.0f);
        }

        private void Render()
        {
            if (render.pause)
                return;


            render.BeginScene(Color.Blue);
          
            cam.move();
            CheckKeys();
            SetupMatrices();

            render.device.Transform.World = Matrix.Identity;
            pc.DrawIndex(portal,ref render.device);
            render.device.Transform.World = Matrix.Identity;
            BSP.BSPModel.BSPDisplayedInfo.Draw(ref render.device, ref bsp, true, ref cam,null);



            for (int x = 0; x < bsp.ClusterInfo[portals[portal].BackCluster].FrontVisibleClusters.Count; x++)
            {
               // BSP.BSPModel.BSPDisplayedInfo.DrawCluster(bsp.ClusterInfo[portals[portal].BackCluster].FrontVisibleClusters[x], Color.Red, ref render.device, ref bsp, true,true, ref cam);
            }
            for (int x = 0; x < bsp.ClusterInfo[portals[portal].BackCluster].BackVisibleClusters.Count; x++)
            {
               // BSP.BSPModel.BSPDisplayedInfo.DrawCluster(bsp.ClusterInfo[portals[portal].BackCluster].BackVisibleClusters[x], Color.Green, ref render.device, ref bsp, true,true, ref cam);
            }

            for (int x = 0; x < bsp.ClusterInfo[portals[portal].FrontCluster].FrontVisibleClusters.Count; x++)
            {
             //   BSP.BSPModel.BSPDisplayedInfo.DrawCluster(bsp.ClusterInfo[portals[portal].FrontCluster].FrontVisibleClusters[x], Color.Green, ref render.device, ref bsp, true, true, ref cam);
            }
            for (int x = 0; x < bsp.ClusterInfo[portals[portal].FrontCluster].BackVisibleClusters.Count; x++)
            {
              //  BSP.BSPModel.BSPDisplayedInfo.DrawCluster(bsp.ClusterInfo[portals[portal].FrontCluster].BackVisibleClusters[x], Color.Purple, ref render.device, ref bsp, false, ref cam);
            }

            render.EndScene();
        }

        protected override void OnPaint(System.Windows.Forms.PaintEventArgs e)
        {
            this.Render(); // Render on painting
        }
        protected override void OnKeyPress(System.Windows.Forms.KeyPressEventArgs e)
        {
            if ((int)(byte)e.KeyChar == (int)System.Windows.Forms.Keys.Escape)
            {
                this.Dispose(); // Esc was pressed
            }
      
        }
        protected override void OnResize(System.EventArgs e)
        {
            //this.WindowState = (FormWindowState.Minimized) || !this.Visible;
            //   Application.DoEvents();
            //   pause = false;
            // this.Render(
        }



        /// <summary>
        /// The main entry point for the application.
        /// </summary>
        public void Main()
        {

            using (PortalViewer frm = (PortalViewer)this)
            {
                if (!frm.InitializeGraphics()) // Initialize Direct3D
                {
                    MessageBox.Show("Could not initialize Direct3D.  This tutorial will exit.");
                    return;
                }
                frm.Show();

                // While the form is still valid, render and process messages
                while (frm.Created)
                {
                    frm.Render();
                    Application.DoEvents();
                }
            }
        }

        private void InitializeComponent()
        {
            this.SuspendLayout();
            // 
            // ModelViewer
            // 
            this.ClientSize = new System.Drawing.Size(292, 266);
            this.Name = "ModelViewer";

            this.ResumeLayout(false);

        }

        private void ModelViewer_MouseDownx(object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Left)
            {
                cam.change(e.X, e.Y);
            }

        }

        private void ModelViewer_MouseUp(object sender, MouseEventArgs e)
        {



        }

        public void CheckKeys()
        {
            Key[] keys=new Key[0];
            try
            {
                keys = cam.device.GetPressedKeys();
            }
            
                catch
            {
            }
            foreach (Key kk in keys)
                {
                   // tslabel.Text = kk.ToString();
                    switch (kk.ToString())
                    {
                        case "Up":
                            portal += 1;
                            if (portal == portals.Length) { portal = 0; }
                            return;
                        
                        case "DownArrow":
                            portal -= 1;
                            if (portal == -1) { portal = portals.Length - 1; }
                            return;
                            break;
                      }
                }
        }

    }
}
