using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

namespace entity.BSP
{
    public sealed class Portals
    {
        public static Portal[] GetPortals(ref Meta meta,int mapnumber)
        {
            Maps.map[mapnumber].OpenMap(MapTypes.Internal);
            Maps.map[mapnumber].BR.BaseStream.Position=meta.offset+108;
            int tempc=Maps.map[mapnumber].BR.ReadInt32();
            int tempr=Maps.map[mapnumber].BR.ReadInt32()-meta.magic;
            Portal[] temp = new Portal[tempc];
            for (int x = 0; x < tempc; x++)
            {
                Maps.map[mapnumber].BR.BaseStream.Position = tempr + (x * 36);
                temp[x] = new Portal(ref Maps.map[mapnumber].BR, meta.magic);
            }
            Maps.map[mapnumber].CloseMap();
            return temp;
        }
    }




    public class Portal
    {
        public short FrontCluster;
        public short BackCluster;
        public int Plane;
        public float X;
        public float Y;
        public float Z;
        public float BoundingRadius;

        public List<Vector3> Vertices = new List<Vector3>();

        public Portal(ref BinaryReader BR, int magic)
        {
            FrontCluster = BR.ReadInt16();
            BackCluster = BR.ReadInt16();
            Plane = BR.ReadInt32();
            X = BR.ReadSingle();
            Y = BR.ReadSingle();
            Z = BR.ReadSingle();
            BoundingRadius = BR.ReadSingle();
            BR.ReadInt32();
            int tempc = BR.ReadInt32();
            int tempr = BR.ReadInt32()-magic;
            BR.BaseStream.Position = tempr;
            for (int x = 0; x < tempc; x++)
            {
                Vector3 v = new Vector3(BR.ReadSingle(), BR.ReadSingle(), BR.ReadSingle());
                Vertices.Add(v);
            }
        }
    }

    public class PortalContainer
    {
        VertexBuffer[] vb;
        Mesh[] spheres;
        int[] verticeCount;
        Matrix[] mat;
        Material Default;
        Material Red;
           Material RedTransparent;
        Material Black;
        public PortalContainer(Portal[] portals,ref Device device)
        {
            Default = new Material();
            Default.Diffuse = System.Drawing.Color.White;
            Default.Ambient = System.Drawing.Color.White;
            RedTransparent = new Material();
            RedTransparent.Diffuse = System.Drawing.Color.FromArgb(50, 255,0,0);
            RedTransparent.Ambient = System.Drawing.Color.FromArgb(50, 255, 0, 0);
       
            Red = new Material();
            Red.Diffuse = System.Drawing.Color.Red;
            Red.Ambient = System.Drawing.Color.Red;
            Black = new Material();
            Black.Diffuse = System.Drawing.Color.Black;
            Black.Ambient = System.Drawing.Color.Black;
              vb = new VertexBuffer[portals.Length];
              verticeCount = new int[portals.Length];
              spheres = new Mesh[portals.Length];
              mat = new Matrix[portals.Length];
            for (int x=0;x<portals.Length;x++)
            {


                spheres[x] = Mesh.Sphere(device, portals[x].BoundingRadius, 10, 10);
                mat[x] = Matrix.Translation(portals[x].X, portals[x].Y, portals[x].Z);
          
                vb[x] = new VertexBuffer(typeof(CustomVertex.PositionColored), portals[x].Vertices.Count, device, Usage.WriteOnly, CustomVertex.PositionColored.Format, Pool.Managed);
                CustomVertex.PositionColored[] verts = (CustomVertex.PositionColored[])vb[x].Lock(0, 0); // Lock the buffer (which will return our structs)
                verticeCount[x] = portals[x].Vertices.Count;
                for (int i = 0; i < portals[x].Vertices.Count; i++)
                {
                    verts[i].Position = new Vector3(portals[x].Vertices[i].X, portals[x].Vertices[i].Y, portals[x].Vertices[i].Z);
               
                 
                }
                vb[x].Unlock();
            }
        }

        public void Draw(ref Device device)
        {
             for (int i = 0; i < this.vb.Length; i++)
            {

            device.SetStreamSource(0, vb[i], 0);
            device.VertexFormat = CustomVertex.PositionColored.Format;
    

            device.RenderState.AlphaBlendEnable = false;
            device.RenderState.AlphaTestEnable = false;



            device.Transform.World = Matrix.Identity;
            device.SetTexture(0, null);
            device.RenderState.FillMode = FillMode.Solid;
            device.RenderState.AlphaBlendEnable = true;
            device.RenderState.AlphaTestEnable = false;
            device.Material = Red;
            PrimitiveType pt;

            pt = PrimitiveType.TriangleFan;
            device.DrawPrimitives(pt,
           0,
           this.verticeCount[i]-2);


            device.RenderState.AlphaBlendEnable = false;
            device.RenderState.AlphaTestEnable = false;
            device.RenderState.FillMode = FillMode.WireFrame;
            device.Material = Red;

    

            pt = PrimitiveType.TriangleFan;
            device.DrawPrimitives(    pt,
           0,
           this.verticeCount[i]-2);

            device.Material = Default;
            device.Transform.World = mat[i];
          //  spheres[i].DrawSubset(0);
        }
      }


        public void DrawIndex(int portal,ref Device device)
        {
           int i=portal;

                device.SetStreamSource(0, vb[i], 0);
                device.VertexFormat = CustomVertex.PositionColored.Format;


                device.RenderState.AlphaBlendEnable = false;
                device.RenderState.AlphaTestEnable = false;



                device.Transform.World = Matrix.Identity;
                device.SetTexture(0, null);
                device.RenderState.FillMode = FillMode.Solid;
                device.Material = RedTransparent;
                device.RenderState.AlphaBlendEnable = true;
                //device.RenderState.AlphaTestEnable = true;
                PrimitiveType pt;

                pt = PrimitiveType.TriangleFan;
                device.DrawPrimitives(pt,
               0,
               this.verticeCount[i] - 2);




                device.RenderState.AlphaBlendEnable = false;
                device.RenderState.AlphaTestEnable = false;
                device.RenderState.FillMode = FillMode.WireFrame;
                device.Material = Red;



                pt = PrimitiveType.TriangleFan;
                device.DrawPrimitives(pt,
               0,
               this.verticeCount[i] - 2);

                device.Material = Default;
                device.Transform.World = mat[i];
                  spheres[i].DrawSubset(0);
            
        }
        
    }
}
