using System;
using System.Runtime.InteropServices;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using System.Windows.Forms;
namespace entity.Raw
{
	/// <summary>
	/// Summary description for Bitmap.
	/// </summary>
	public class ParsedBitmap
	{
		[DllImport("kernel32.dll")]
		static extern void RtlMoveMemory(IntPtr src,byte[] crap, int cb);
        public HaloBitmapHeader BitmapHeader;
        public int headersize = 0;
		public BitmapInfo[] Properties;  

		public ParsedBitmap(ref Meta meta,int mapnumber)
		{
            BitmapHeader = new HaloBitmapHeader(ref meta);
            switch (Maps.map[mapnumber].HaloVersion)
            {

                case Map.HaloVersionEnum.Halo2:
                    ReadH2ParsedBitmap(ref meta, mapnumber);
                    break;
                case Map.HaloVersionEnum.HaloCE:
                case Map.HaloVersionEnum.Halo1:
                    ReadCEParsedBitmap(ref meta, mapnumber);
                    break;
            }
        }
                 public void ReadCEParsedBitmap(ref Meta meta, int mapnumber)
                {
                    BinaryReader BR = new BinaryReader(meta.MS);
                    int tempr;
                    if (Maps.map[mapnumber].MetaInfo.external[meta.tagnumber] == true)
                    {
                        BR.BaseStream.Position = 96;
                        int tempc = BR.ReadInt32();
                        tempr = BR.ReadInt32();
                        headersize = tempr;
                        Properties = new BitmapInfo[tempc];
                        for (int x = 0; x < tempc; x++)
                        {
                            Properties[x] = new BitmapInfo(tempr + (48 * x), ref meta);

                        }
                    }
                    else
                    {
                       BR.BaseStream.Position =  96;
                        int tempc = BR.ReadInt32();
                        tempr = BR.ReadInt32() - Maps.map[mapnumber].PrimaryMagic - Maps.map[mapnumber].MetaInfo.Offset[meta.tagnumber];
                        headersize = tempr;
                        Properties = new BitmapInfo[tempc];
                        for (int x = 0; x < tempc; x++)
                        {
                            Properties[x] = new BitmapInfo(tempr + (48 * x), ref meta);

                        }
                    }
                 }
      
        public void ReadH2ParsedBitmap(ref Meta meta, int mapnumber)
        {
			BinaryReader BR=new BinaryReader(meta.MS);
			BR.BaseStream.Position=68;
			int tempc=BR.ReadInt32();
            int tempr = BR.ReadInt32() - Maps.map[mapnumber].SecondaryMagic - meta.offset;
            headersize = tempr;
			Properties=new BitmapInfo[tempc];
			for (int x=0;x<tempc;x++)
			{
				Properties[x]=new BitmapInfo(tempr+(116*x) ,ref meta);

			}

		}

        public void ConvertCEtoH2ParsedBitmap(ref Meta meta, int mapnumber)
        {

       
            meta.MS = new MemoryStream();
            meta.items.Clear();
            BinaryWriter BW = new BinaryWriter(meta.MS);
            BitmapHeader.WriteHeaderHaloBitmapHeader(ref BW);
            BW.BaseStream.Position = 32;
            BW.Write(new char[44]);
            int size = 76;
          
            int tempc = 1;
            int tempr = meta.offset + size + meta.magic;
            BW.BaseStream.Position = 60;
            BW.Write(tempc);
            BW.Write(tempr);
            Meta.Reflexive reflex = new Meta.Reflexive();
            reflex.offset = 60;
            reflex.translation = size;
        
            reflex.intag = meta.tagnumber;
            reflex.intagname = meta.name;
            reflex.intagtype = meta.type;

            reflex.chunkcount = tempc;
            reflex.pointstotagnumber = meta.tagnumber;
            reflex.pointstotagname = meta.name;
            reflex.pointstotagtype = meta.type;
            reflex.pointstotagname = meta.name;
            reflex.pointstotagtype = meta.type;
            meta.items.Add(reflex);

            tempc = 1;
            tempr = meta.offset + size +60+ meta.magic;
            BW.BaseStream.Position = 128;
            BW.Write(tempc);
            BW.Write(tempr);
            reflex = new Meta.Reflexive();
            reflex.offset = 128;
            reflex.translation = size + 60;

            reflex.intag = meta.tagnumber;
            reflex.intagname = meta.name;
            reflex.intagtype = meta.type;

            reflex.chunkcount = tempc;
            reflex.pointstotagnumber = meta.tagnumber;
            reflex.pointstotagname = meta.name;
            reflex.pointstotagtype = meta.type;
            reflex.pointstotagname = meta.name;
            reflex.pointstotagtype = meta.type;
            meta.items.Add(reflex);
            BW.BaseStream.Position = size;
            BW.Write(new char[92]);
            size += 92;


            tempc = Properties.Length;
            tempr = meta.offset +   size + meta.magic;
            BW.BaseStream.Position = 68;
            BW.Write(tempc);
            BW.Write(tempr);
            reflex = new Meta.Reflexive();
            reflex.offset = 68;
            reflex.translation = size;

            reflex.intag = meta.tagnumber;
            reflex.intagname = meta.name;
            reflex.intagtype = meta.type;

            reflex.chunkcount = tempc;
            reflex.pointstotagnumber = meta.tagnumber;
            reflex.pointstotagname = meta.name;
            reflex.pointstotagtype = meta.type;
            reflex.pointstotagname = meta.name;
            reflex.pointstotagtype = meta.type;
            meta.items.Add(reflex);



            for (int x = 0; x < tempc; x++)
            {
                BW.BaseStream.Position = size + (x * 116);
                BW.Write(new char[116]);


                BW.BaseStream.Position = size + (x * 116);
                Properties[x].Write(ref BW);
                int neg = -1;
                meta.raw.rawChunks[x].pointerMetaOffset = size + (x * 116) + 28;
                BW.BaseStream.Position = meta.raw.rawChunks[x].pointerMetaOffset + 4;
                BW.Write(neg);
                BW.Write(neg);
                BW.Write(neg);
                BW.Write(neg);
                BW.Write(neg);
                BW.Write(meta.raw.rawChunks[x].size);
                Meta.Ident id=new Meta.Ident();
                id.offset = size + (x * 116) + 76;
                BW.BaseStream.Position=id.offset;
                BW.Write(meta.ident);
                id.pointstotagnumber = meta.tagnumber;
                id.pointstotagname = meta.name;
                id.pointstotagtype = meta.type;
                id.intag = meta.tagnumber;
                id.intagname = meta.name;
                id.intagtype = meta.type;
                id.ident = meta.ident;
                meta.items.Add(id);
            }
            meta.size = size + (Properties.Length * 116);
        }

        public class HaloBitmapHeader
        {
            #region Fields
            public short type;
            public short format;
            public short usage;
            public short usageflags;
            public float detailfadefactor;
            public float sharpenamount;
            public float bumpheight;
            public short spritebudgesize;
            public short spritebudgetcount;
            public short colorplatewidth;
            public short colorplateheight;
            public short compressedcolorplatedata;

            #endregion
            #region Constructor
            public HaloBitmapHeader(ref Meta meta)
            {
                BinaryReader BR = new BinaryReader(meta.MS);
                BR.BaseStream.Position = 0;
                        type=BR.ReadInt16();
          format=BR.ReadInt16();
          usage=BR.ReadInt16();
          usageflags=BR.ReadInt16();
          detailfadefactor = BR.ReadSingle();
          sharpenamount = BR.ReadSingle();
          bumpheight = BR.ReadSingle();
          spritebudgesize=BR.ReadInt16();
          spritebudgetcount=BR.ReadInt16();
          colorplatewidth=BR.ReadInt16();
          colorplateheight=BR.ReadInt16();
          compressedcolorplatedata=BR.ReadInt16();

            }
            public void WriteHeaderHaloBitmapHeader(ref BinaryWriter BW)
            {
               
                BW.BaseStream.Position = 0;
                BW.Write(type);
                 BW.Write(format);
                 BW.Write(usage);
                 BW.Write(usageflags);
                 BW.Write(detailfadefactor);
                 BW.Write(sharpenamount);
                 BW.Write(bumpheight);
                 BW.Write(spritebudgesize);
                 BW.Write(spritebudgetcount);
                 BW.Write(colorplatewidth);
                BW.Write(colorplateheight);
                 BW.Write(compressedcolorplatedata);

            }
            #endregion
        }
    







		public class BitmapInfo
		{
			#region Fields
            public char[] tagtype;
			public ushort width;
			public ushort height;
			public ushort depth;
			public ushort type;
			public ushort format;
			public BitmapType formatname;
            public bool swizzle;
			public ushort flags;
			public ushort regPointX;
			public ushort regPointY;
			public ushort mipMapCount;
			public ushort pixelOffset; 
			#endregion
			#region Constructor
			public BitmapInfo(int offset,ref Meta meta)
			{
				BinaryReader BR=new BinaryReader(meta.MS);
                BR.BaseStream.Position = offset;
                tagtype = BR.ReadChars(4);
				width=BR.ReadUInt16();
				height=BR.ReadUInt16();
				depth=BR.ReadUInt16();
				type=BR.ReadUInt16();
				formatname=(BitmapType)BR.ReadUInt16();
				flags=BR.ReadUInt16();
              // if ((flags & 0x1000) == 0x1000) { swizzle = true; }
                if ((flags & 0x8) == 0x8) { swizzle = true; }
                //MessageBox.Show("test");
				regPointX=BR.ReadUInt16();
				regPointY=BR.ReadUInt16();
				mipMapCount=BR.ReadUInt16();
				pixelOffset=BR.ReadUInt16();
			}	
			#endregion
          
            public void Write(ref BinaryWriter BW)
            {

                BW.Write(tagtype);
                BW.Write(width);
                BW.Write(height);
                BW.Write(depth);
                BW.Write(type );
                BW.Write((short)formatname);
                BW.Write(flags);
                
                BW.Write(regPointX);
                BW.Write(regPointY);
               BW.Write(mipMapCount); 
                BW.Write(pixelOffset);
            }

            public void Generate(DDS.DDSURFACEDESC2 ddsd)
            {
             
                string bitmstring="mtib";
                tagtype = bitmstring.ToCharArray();
               // width = ddsd.width; ;
             //   height = ddsd.height; ;
			//depth=ddsd.depth;
            /*
     type=ddsd.;
     format;
     formatname;
     swizzle;
     flags;
     regPointX;
     regPointY;
     mipMapCount=ddsd.MipMapCount;
     pixelOffset=d; 
          * */
            }
		}
		public enum BitmapType:int
		{
			BITM_FORMAT_A8 = 0,
			BITM_FORMAT_Y8 = 1,
			BITM_FORMAT_AY8 = 2,
			BITM_FORMAT_A8Y8 = 3,
			BITM_FORMAT_R5G6B5 = 6,
			BITM_FORMAT_A1R5G5B5 = 8,
			BITM_FORMAT_A4R4G4B4 = 9,
			BITM_FORMAT_X8R8G8B8 = 10,
			BITM_FORMAT_A8R8G8B8 = 11,
			BITM_FORMAT_DXT1 = 14,
			BITM_FORMAT_DXT2AND3 = 15,
			BITM_FORMAT_DXT4AND5 = 16,
			BITM_FORMAT_P8 = 17,
			BITM_TYPE_2D = 0,
			BITM_TYPE_3D = 1,
			BITM_TYPE_CUBEMAP = 2,
            BITM_TYPE_LIGHTMAP =18,
            BITM_TYPE_UNK =22
		}

        //visualchunkindex is for lightmaps only
		public Bitmap FindChunkAndDecode(int chunknumber,int numberinchunk,ref Meta meta,int mapnumber,int visualchunkindex,int bspnumber)

		{
			for (int x=0;x<meta.raw.rawChunks.Count;x++)
			{
				BitmapRawDataChunk tempb=(BitmapRawDataChunk) meta.raw.rawChunks[x];
				if (tempb.inchunk==chunknumber&&tempb.num==numberinchunk)
				{
					int height=this.Properties[chunknumber].height;
					int depth=this.Properties[chunknumber].depth;
					int width=this.Properties[chunknumber].width;
					int pixeloff=this.Properties[chunknumber].pixelOffset;
                   int tempsize = tempb.size;//width * height * 4;
                    byte[] guh = new byte[tempsize];//];
                   
                    Array.Copy(tempb.MS.ToArray(), 0, guh, 0, tempsize);

                    Bitmap b = DecodeBitm(guh, height, width, depth, pixeloff, this.Properties[chunknumber].formatname, this.Properties[chunknumber].swizzle, mapnumber, visualchunkindex, bspnumber);
					return b;
				}
			}
			return null;
		}

        

		public Bitmap DecodeBitm(byte[] fart,int height,int width,int depth,int pixeloffset,BitmapType format,bool swizzle,int mapnumber,int visualchunkindex,int ident)
		{
        
			byte[] poo=new byte[0];
			IntPtr ptr=new IntPtr();
			
			PixelFormat o=new PixelFormat();
			int poolength;
			byte[] fu;
			Misc.DecodeDXT decode=new entity.Misc.DecodeDXT();

           
            int oldwidth = width;
            if (width % 8 != 0) width = (ushort)(width + (8 - (width % 8)));
           int stride=width;
			switch (format)
			{



				case (BitmapType)14:
					fart=decode.DecodeDXT1(height,width,fart);
					stride*=4;
					o=PixelFormat.Format32bppArgb;
					break;

				case (BitmapType)15:
					fart=decode.DecodeDXT23(height,width,fart);
					stride*=4;
					o=PixelFormat.Format32bppArgb;
					break;

				case (BitmapType)16:
					fart=decode.DecodeDXT45(height,width,fart);
					stride*=4;
					o=PixelFormat.Format32bppArgb;
					break;
 
				case BitmapType.BITM_FORMAT_A8R8G8B8:

                   // fart = Swizzler.Swizzle(fart, width, height, -1, 32, true);
                    if (swizzle == true)
                    {
                        fart = Swizzler.Swizzle(fart, width, height, depth, 32, true);
                    }
                    
       
     
					stride*=4;
					o=PixelFormat.Format32bppArgb;
					break;

				case BitmapType.BITM_FORMAT_X8R8G8B8:
                    if (swizzle == true)
                    {
                       fart=Swizzler.Swizzle(fart, width, height, -1, 32, true);
                    }
               
					stride*=4;
					o=PixelFormat.Format32bppArgb;
					break;



                    ///16 bit



                case BitmapType.BITM_FORMAT_A4R4G4B4:
                    if (swizzle == true)
                    {
                        fart = Swizzler.Swizzle(fart, pixeloffset, width, height, -1, 16, true);
                    }
					stride*=2;
					o=PixelFormat.Format16bppRgb565;
					break;



                case (BitmapType)22:
                    if (swizzle == true)
                    {
                        fart = Swizzler.Swizzle(fart, 0, width, height, -1, 16, true);
                    }
                    stride *= 2;
                    o = PixelFormat.Format32bppArgb;
                    poolength = fart.Length;
                    fu = new byte[poolength / 2 * 4];
                    for (int e = 0; e < poolength / 2; e++)
                    {
                        int r = e * 2;
                        fu[r * 2 + 0] = fart[r + 1];
                        fu[r * 2 + 1] = fart[r + 1];
                        fu[r * 2 + 2] = fart[r + 1];
                        fu[r * 2 + 3] = fart[r + 0];
                    }
                    fart = fu;
                    break;


                case BitmapType.BITM_FORMAT_A1R5G5B5:
                    if (swizzle == true)
                    {
                        fart = Swizzler.Swizzle(fart, pixeloffset, width, height, -1, 16, true);
                    }
						stride*=2;
					o=PixelFormat.Format16bppRgb555;
					break;
                case BitmapType.BITM_FORMAT_R5G6B5:
                    if (swizzle == true)
                    {
                        fart = Swizzler.Swizzle(fart, pixeloffset, width, height, -1, 16, true);
                    }
						stride*=2;
					o=PixelFormat.Format16bppRgb565;
					break;
                case BitmapType.BITM_FORMAT_A8Y8:
                    if (swizzle == true)
                    {
                        fart = Swizzler.Swizzle(fart, pixeloffset, width, height, depth, 16, true);
                    }
                    o = PixelFormat.Format32bppArgb;
					poolength=fart.Length;
					fu=new byte[poolength/2*4];
					for (int e=0;e<poolength/2;e++)
					{
						int r=e*2;
                        fu[r + 0] = fart[r + 1];
                        fu[r + 1] = fart[r + 1];
                        fu[r + 2] = fart[r + 1];
                        fu[r + 3] = fart[r + 0];
					}
                    fart = fu;
					stride*=4;
					break;
                //8 bit
                case BitmapType.BITM_FORMAT_P8:
                    if (swizzle == true)
                    {
                        fart = Swizzler.Swizzle(fart, pixeloffset, width, height, depth, 8, true);
                    }
                    o = PixelFormat.Format32bppArgb;
                    poolength = fart.Length;
                    fu = new byte[poolength * 4];
                    for (int e = 0; e < poolength; e++)
                    {
                        int r = e * 4;
                        fu[r + 0] = fart[e];
                        fu[r + 1] = fart[e];
                        fu[r + 2] = fart[e];
                        fu[r + 3] = 255;// 255;
                        if (fart[e] == 0) { fu[r + 3] = 0; }
                    }
                    fart = fu;
                    stride *= 4;
                    break;
                case BitmapType.BITM_TYPE_LIGHTMAP:
                    if (swizzle == true)
                    {
fart = Swizzler.Swizzle(fart, pixeloffset, width, height, depth, 8, true);
                    }
                    o = PixelFormat.Format32bppArgb;// Format32bppArgb;
                    poolength = fart.Length;
                    fu = new byte[poolength * 4];
                    int bspnumber = ident;
                    int paletteindex;
                    if (visualchunkindex < 0)
                    {
                        int wtf=0 - (visualchunkindex+1);
                        paletteindex = Maps.map[mapnumber].BSP.sbsp[bspnumber].SceneryChunk_LightMap_Index[wtf];
                    }
                    else
                    {

                        paletteindex = Maps.map[mapnumber].BSP.sbsp[bspnumber].VisualChunk_LightMap_Index[visualchunkindex];
                    }
                    if (paletteindex==255) return null;
                    for (int e = 0; e < poolength; e++)
                    {
                        int r = e * 4;
                        fu[r + 0] = (byte)Maps.map[mapnumber].BSP.sbsp[bspnumber].LightMap_Palettes[paletteindex][fart[e]].r;
                        fu[r + 1] = (byte)Maps.map[mapnumber].BSP.sbsp[bspnumber].LightMap_Palettes[paletteindex][fart[e]].g;
                        fu[r + 2] = (byte)Maps.map[mapnumber].BSP.sbsp[bspnumber].LightMap_Palettes[paletteindex][fart[e]].b;
                        fu[r + 3] = (byte)Maps.map[mapnumber].BSP.sbsp[bspnumber].LightMap_Palettes[paletteindex][fart[e]].a;
                       
                    }
                    fart = fu;
                   stride *= 4;
                    break;

                case (BitmapType)1:
                    if (swizzle == true)
                    {
                        fart = Swizzler.Swizzle(fart, pixeloffset, width, height, depth, 8, true);
                    }
                    o = PixelFormat.Format32bppArgb;
                    poolength = fart.Length;
                    fu = new byte[poolength * 4];
                    for (int e = 0; e < poolength; e++)
                    {
                        int r = e * 4;
                        fu[r + 0] = fart[e];
                        fu[r + 1] = fart[e];
                        fu[r + 2] = fart[e];
                        fu[r + 3] = fart[255];
                    }
                    fart = fu;
                    stride *= 4;
                    break;

                
				default:
					return null;
			}

			Marshal.FreeHGlobal(ptr);
			ptr = Marshal.AllocHGlobal(fart.Length);

			RtlMoveMemory(ptr,fart,fart.Length);

			Bitmap temp=new System.Drawing.Bitmap(width, height,stride,o, ptr);
			return temp;

		}



	}

}
