using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

namespace entity
{
    public class DDS
    {
        #region DDS
         //DDS
         //The dwFlags member of the modified DDSURFACEDESC2 structure can be set to one or more of the following values.
        enum DDSEnum:int
  {
    DDSD_CAPS = 0x1,
      DDSD_HEIGHT = 0x2,
      DDSD_WIDTH = 0x4,
      DDSD_PITCH = 0x8,
      DDSD_PIXELFORMAT = 0x1000,
      DDSD_MIPMAPCOUNT = 0x20000,
      DDSD_LINEARSIZE = 0x80000,
      DDSD_DEPTH = 0x800000,

    DDPF_ALPHAPIXELS = 0x1,
    DDPF_FOURCC = 0x4,
    DDPF_RGB = 0x40,

    //'The dwCaps1 member of the DDSCAPS2 structure can be set to one or more of the following values.
    DDSCAPS_COMPLEX = 0x8,
    DDSCAPS_TEXTURE = 0x1000,
    DDSCAPS_MIPMAP = 0x400000,

    //The dwCaps2 member of the DDSCAPS2 structure can be set to one or more of the following values.
      DDSCAPS2_CUBEMAP = 0x200,
      DDSCAPS2_CUBEMAP_POSITIVEX = 0x400,
      DDSCAPS2_CUBEMAP_NEGATIVEX = 0x800,
      DDSCAPS2_CUBEMAP_POSITIVEY = 0x1000,
    DDSCAPS2_CUBEMAP_NEGATIVEY = 0x2000,
    DDSCAPS2_CUBEMAP_POSITIVEZ = 0x4000,
    DDSCAPS2_CUBEMAP_NEGATIVEZ = 0x8000,
    DDSCAPS2_VOLUME = 0x200000
    }
        #endregion
        #region DDS Structures

        public struct DDS_HEADER_STRUCTURE
        {
            public string magic;
            public DDSURFACEDESC2 ddsd;

    public void ReadStruct(ref BinaryReader BR)
    {
      magic = new string(BR.ReadChars(4));
      ddsd.readStruct(ref BR);
    }
      public void WriteStruct(ref BinaryWriter BW)
        {
            BW.Write(magic.ToCharArray());
            ddsd.writeStruct(ref BW);
        }
        // Generate a DDS Header
        //////////////////////////////////////////////////////
         public void generate(ref Raw.ParsedBitmap.BitmapInfo  b2)
         {
           magic = "DDS ";
          ddsd.generate(ref b2);
        }


        }
   

    
    //////////////////////////////////////////////////////

  public struct DDSURFACEDESC2
  {
    public int size_of_structure ;
    public int flags ;
    public int height;
    public int width;
    public int PitchOrLinearSize;
    public int depth; //Only used for volume textures.
    public int  MipMapCount; //Total Number of MipMaps
      public int[] Reserved1;
    public  DDPIXELFORMAT ddfPixelFormat;
    public DDSCAPS2 ddsCaps;
    public int Reserved2;
    public void readStruct(ref BinaryReader br )
    {

      size_of_structure = br.ReadInt32();
      flags = br.ReadInt32();
      height = br.ReadInt32();
      width = br.ReadInt32();
      PitchOrLinearSize = br.ReadInt32();
      depth = br.ReadInt32();
      MipMapCount = br.ReadInt32();

      Reserved1 = new int[11]; //11 ints long
      for (int x = 0 ;x<11;x++)
          Reserved1[x] = br.ReadInt32();
  
      ddfPixelFormat.readStruct(ref br);
      ddsCaps.readStruct(ref br);
      Reserved2 = br.ReadInt32();
    }
    public void writeStruct(ref BinaryWriter bw)
    {
      bw.Write(size_of_structure);
      bw.Write(flags);
      bw.Write(height);
      bw.Write(width);
      bw.Write(PitchOrLinearSize);
      bw.Write(depth);
      bw.Write(MipMapCount);
  
      for (int x = 0 ;x<11;x++)
        bw.Write(Reserved1[x]);

     ddfPixelFormat.writeStruct(ref bw);
      ddsCaps.writeStruct(ref bw);
      bw.Write(Reserved2);
    }
    //////////////////////////////////////////////////////
    // Generate the DDSURFACEDESC2 chunk of the header
    //////////////////////////////////////////////////////
    public void generate(ref entity.Raw.ParsedBitmap.BitmapInfo b2)
    {
        Reserved1 = new int[11];
     //'Size of structure. This member must be set to 124.
      size_of_structure = 124;

      //Flags to indicate valid fields. Always include DDSD_CAPS, DDSD_PIXELFORMAT,
      //DDSD_WIDTH, DDSD_HEIGHT and either DDSD_PITCH or DDSD_LINEARSIZE.
      flags = flags + (int)DDSEnum.DDSD_CAPS;
      flags = flags + (int)DDSEnum.DDSD_PIXELFORMAT;
      flags = flags + (int)DDSEnum.DDSD_WIDTH;
      flags = flags + (int)DDSEnum.DDSD_HEIGHT;
      switch (b2.formatname )
      {
          case entity.Raw.ParsedBitmap.BitmapType.BITM_FORMAT_DXT1:
              case entity.Raw.ParsedBitmap.BitmapType.BITM_FORMAT_DXT2AND3:
              case entity.Raw.ParsedBitmap.BitmapType.BITM_FORMAT_DXT4AND5:
                  flags = flags + (int)DDSEnum.DDSD_LINEARSIZE;
              break;

      default:
          flags = flags + (int)DDSEnum.DDSD_PITCH;
                    break;
    }

      //Height of the main image in pixels
      height = b2.height;

      //Width of the main image in pixels
      width = b2.width;

      //For uncompressed formats, this is the number of bytes per scan line (DWORD aligned) for the main
      //image. dwFlags should include DDSD_PITCH in this case. For compressed formats, this is the (total)
     //number of bytes for the main image. dwFlags should be include DDSD_LINEARSIZE in this case.
      int RGBBitCount  = 0;
    switch (b2.formatname )
      {
          case entity.Raw.ParsedBitmap.BitmapType.BITM_FORMAT_R5G6B5:
          RGBBitCount = 16;
             break;
         case entity.Raw.ParsedBitmap.BitmapType.BITM_FORMAT_A1R5G5B5:
          RGBBitCount = 16;
             break;
         case entity.Raw.ParsedBitmap.BitmapType.BITM_FORMAT_A4R4G4B4:
          RGBBitCount = 16;
             break;
         case entity.Raw.ParsedBitmap.BitmapType.BITM_FORMAT_X8R8G8B8:
          RGBBitCount = 32;
             break;
         case entity.Raw.ParsedBitmap.BitmapType.BITM_FORMAT_A8R8G8B8:
          RGBBitCount = 32;
             break;
    }

      PitchOrLinearSize = b2.width * (RGBBitCount / 8);
      //For volume textures, this is the depth of the volume.
      //dwFlags should include DDSD_DEPTH in this case.
      depth = 0;// 'There are no volume textures (that I know of) in Halo.

     //For items with mipmap levels, this is the total number of levels in the mipmap
      //chain of the main image. dwFlags should include DDSD_MIPMAPCOUNT in this case
      MipMapCount = b2.mipMapCount;
     if (MipMapCount > 0) flags = flags + (int)DDSEnum.DDSD_MIPMAPCOUNT;



      //A 32-byte value that specifies the pixel format structure.
      ddfPixelFormat.generate(b2);

      //A 16-byte value that specifies the capabilities structure.
      ddsCaps.generate(ref b2);//
  //  End Sub
    }
  }

  public struct DDPIXELFORMAT
  {
    public int size;//   '32
    public int  Flags;//  '4
    public string  FourCC;// 'DXT1, DXT2, etc..
    public int  RGBBitCount;
    public int  RBitMask;
    public int  GBitMask;
    public int  BBitMask;
     public int RGBAlphaBitMask;
    public  void  readStruct(ref BinaryReader br)
    {
      size = br.ReadInt32();// '32
      Flags = br.ReadInt32();// '4
      FourCC = new string(br.ReadChars(4));// 'DXT1, DXT2, etc..
      RGBBitCount = br.ReadInt32();
      RBitMask = br.ReadInt32();
      GBitMask = br.ReadInt32();
      BBitMask = br.ReadInt32();
      RGBAlphaBitMask = br.ReadInt32();
    }
    public void writeStruct(ref BinaryWriter bw )
    {
      bw.Write(size);
      bw.Write(Flags);
      bw.Write(FourCC.ToCharArray());
    
      bw.Write(RGBBitCount);
      bw.Write(RBitMask);
      bw.Write(GBitMask);
      bw.Write(BBitMask);
      bw.Write(RGBAlphaBitMask);
    }
     //////////////////////////////////////////////////////
    // Generate the DDPIXELFORMAT chunk of the header
    //////////////////////////////////////////////////////
    public void generate(Raw.ParsedBitmap.BitmapInfo b2)
    {
      //Size of structure. This member must be set to 32. 
        size = 32;

      //Flags to indicate valid fields. Uncompressed formats will usually use DDPF_RGB to indicate
      //an RGB format, while compressed formats will use DDPF_FOURCC with a four-character code. 
      //This is accomplished in the below structure.

      //This is the four-character code for compressed formats. dwFlags should include DDPF_FOURCC in
      //this case. For DXTn compression, this is set to "DXT1", "DXT2", "DXT3", "DXT4", or "DXT5". 
      Flags = 0;
     switch ((int)b2.formatname )
      {
         case 0xE:
          FourCC = "DXT1";
          Flags = Flags + (int)DDSEnum.DDPF_FOURCC;
             break;
        case 0xF:
          FourCC = "DXT3";
          Flags = Flags + (int)DDSEnum.DDPF_FOURCC;
             break;
        case 0x10:
          FourCC = "DXT4";
          Flags = Flags + (int)DDSEnum.DDPF_FOURCC;
             break;
         default:
          FourCC = "0000";
          Flags = Flags + (int)DDSEnum.DDPF_RGB;
          break;
     }

      //For RGB formats, this is the total number of bits in the format. dwFlags should include DDPF_RGB
      //in this case. This value is usually 16, 24, or 32. For A8R8G8B8, this value would be 32. 
      switch (b2.formatname )
      {


          case Raw.ParsedBitmap.BitmapType.BITM_FORMAT_P8:
              RGBBitCount = 8;
              break;
          case Raw.ParsedBitmap.BitmapType.BITM_FORMAT_R5G6B5:
          RGBBitCount = 16;
          break;
          case Raw.ParsedBitmap.BitmapType.BITM_FORMAT_A1R5G5B5:
          RGBBitCount = 16;
     
          Flags = Flags + (int)DDSEnum.DDPF_ALPHAPIXELS;
          break;
          case Raw.ParsedBitmap.BitmapType.BITM_FORMAT_A4R4G4B4:
          RGBBitCount = 16;
          Flags = Flags + (int)DDSEnum.DDPF_ALPHAPIXELS;
          break;
          case Raw.ParsedBitmap.BitmapType.BITM_FORMAT_X8R8G8B8:
              Flags = Flags + (int)DDSEnum.DDPF_ALPHAPIXELS;
          RGBBitCount = 32;
          break;
          case Raw.ParsedBitmap.BitmapType.BITM_FORMAT_A8R8G8B8:
              Flags = Flags + (int)DDSEnum.DDPF_ALPHAPIXELS;
          RGBBitCount = 32;
          break;
          default:
          RGBBitCount = 0;
          break;

      }

      //For RGB formats, this contains the masks for the red, green, and blue channels. For A8R8G8B8, these
      //values would be 0x00ff0000, 0x0000ff00, and 0x000000ff respectively. 
      switch (b2.formatname )
      {
          case Raw.ParsedBitmap.BitmapType.BITM_FORMAT_R5G6B5:
          RBitMask = 0xF800;
          GBitMask = 0x7E0;
          BBitMask = 0x1F;
          RGBAlphaBitMask = 0x0;
          break;
          case Raw.ParsedBitmap.BitmapType.BITM_FORMAT_A1R5G5B5:
          RBitMask = 0x7C00;
          GBitMask = 0x3E0;
          BBitMask = 0x1F;
          RGBAlphaBitMask = 0x8000;
          break;
          case Raw.ParsedBitmap.BitmapType.BITM_FORMAT_A4R4G4B4:
          RBitMask = 0xF000;
          GBitMask = 0xF0;
          BBitMask = 0xF;
          RGBAlphaBitMask = 0xF00000;
          break;
          case Raw.ParsedBitmap.BitmapType.BITM_FORMAT_X8R8G8B8:
          RBitMask = 0xFF000;
          GBitMask = 0xFF00;
          BBitMask = 0xFF;
          RGBAlphaBitMask = 0xF000000;
          break;
          case Raw.ParsedBitmap.BitmapType.BITM_FORMAT_A8R8G8B8:
          RBitMask = 0xFF0000;
          GBitMask = 0xFF00;
          BBitMask = 0xFF;
          RGBAlphaBitMask = int.Parse ("FF000000", System.Globalization.NumberStyles.HexNumber );
          break;
       default:
          RBitMask = 0;
          GBitMask = 0;
          BBitMask = 0;
          RGBAlphaBitMask = 0x0;
          break;
      }

      //For RGB formats, this contains the mask for the alpha channel, if any. dwFlags should include 
      //DDPF_ALPHAPIXELS in this case. For A8R8G8B8, this value would be 0xff000000. 
      //^^ This was set in the previous block of code.
    }
  }

  public struct DDSCAPS2
  {
    public int caps1;
    public int  caps2;
    public int[] Reserved;
   public void readStruct(ref BinaryReader br)
   {
      caps1 = br.ReadInt32();
      caps2 = br.ReadInt32();
      Reserved= new int[2];
      Reserved[0] = br.ReadInt32();
      Reserved[1] = br.ReadInt32();
   }
   public void  writeStruct(ref BinaryWriter bw)
  {
      bw.Write(caps1);
      bw.Write(caps2);
      bw.Write(0);//Reserved[0]);
      bw.Write(0); //Reserved[1]);
  }
    //////////////////////////////////////////////////////
    // Generate the DDSCAPS2 chunk of the header
   //////////////////////////////////////////////////////
    public void generate(ref Raw.ParsedBitmap.BitmapInfo b2)
    {
      //DDS files should always include DDSCAPS_TEXTURE. If the file contains mipmaps, DDSCAPS_MIPMAP
      //should be set. For any DDS file with more than one main surface,such as a mipmaps, cubic
      //environment map, or volume texture, DDSCAPS_COMPLEX should also be set. 
      caps1 = caps1 + (int)DDSEnum.DDSCAPS_TEXTURE;
      if (b2.mipMapCount > 0) caps1 = caps1 + (int)DDSEnum.DDSCAPS_MIPMAP;
      caps1 = caps1 + (int)DDSEnum.DDSCAPS_COMPLEX;

      //For cubic environment maps, DDSCAPS2_CUBEMAP should be included as well as one or more faces of
      //the map (DDSCAPS2_CUBEMAP_POSITIVEX, DDSCAPS2_CUBEMAP_NEGATIVEX, DDSCAPS2_CUBEMAP_POSITIVEY,
      // DDSCAPS2_CUBEMAP_NEGATIVEY, DDSCAPS2_CUBEMAP_POSITIVEZ, DDSCAPS2_CUBEMAP_NEGATIVEZ). For volume
      //textures, DDSCAPS2_VOLUME should be included. 
      //******************************************************************************
      //This is where I stopped - this code needs to be added for cubemaps.
      //******************************************************************************
      caps2 = 0;

      
    }
  }

#endregion

    public static void ExtractDDS(Meta m,Raw.ParsedBitmap pm, ref BinaryWriter bw,int index)
    {
    DDS_HEADER_STRUCTURE dds=new DDS_HEADER_STRUCTURE();
    
    byte[] bChunk;
    byte[] tmp;
    
  bChunk=m.raw.rawChunks[index].MS.ToArray();

   int width, height;

      width =pm.Properties[index].width;
      height = pm.Properties[index].height;

    
        if (pm.Properties[index].formatname == Raw.ParsedBitmap.BitmapType.BITM_FORMAT_A8R8G8B8 ||
            pm.Properties[index].formatname == Raw.ParsedBitmap.BitmapType.BITM_FORMAT_X8R8G8B8)
        {tmp = Swizzler.Swizzle(bChunk, width, height, -1, 32, false);  bChunk = tmp;}
        if (pm.Properties[index].formatname == Raw.ParsedBitmap.BitmapType.BITM_FORMAT_A1R5G5B5 || 
            pm.Properties[index].formatname == Raw.ParsedBitmap.BitmapType.BITM_FORMAT_R5G6B5 ||
            pm.Properties[index].formatname == Raw.ParsedBitmap.BitmapType.BITM_FORMAT_A4R4G4B4 )
        { tmp = Swizzler.Swizzle(bChunk, width, height, -1, 16, false); bChunk = tmp; }
      


      dds.generate(ref pm.Properties[index]);
      bw.BaseStream.SetLength(128);
      bw.BaseStream.Position = 0;
      dds.WriteStruct(ref bw);
      bw.BaseStream.Position = 128;
      bw.Write(bChunk);
      bw.Close();


      }

    public static void InjectDDS(Meta m, Raw.ParsedBitmap pm ,ref BinaryReader br, int index )
    {
    Maps.map[m.mapnumber].OpenMap(m.raw.rawChunks[index].rawLocation);
    DDS_HEADER_STRUCTURE dds = new DDS_HEADER_STRUCTURE() ;
   

    //' Read dds header
    dds.ReadStruct(ref br);


    



 
    switch (dds.ddsd.ddfPixelFormat.FourCC)
    {
        case "DXT1":
       // If Not bitmheader._items(index).Format = bitmEnum.BITM_FORMAT_DXT1 Then
       //   bCompatible = False
        //  strErrorMsg &= "Image being injected must be DXT1." & vbCrLf
        //End If
        case "DXT2":
         case    "DXT3":
       // If Not bitmheader._items(index).Format = bitmEnum.BITM_FORMAT_DXT2AND3 Then
       //   bCompatible = False
        //  strErrorMsg &= "Image being injected must be DXT2 or DXT3." & vbCrLf
        //End If
        case "DXT4":
        case "DXT5":
       // If Not bitmheader._items(index).Format = bitmEnum.BITM_FORMAT_DXT4AND5 Then
       //   bCompatible = False
       //   strErrorMsg &= "Image being injected must be DXT4 or DXT5." & vbCrLf
      //  End If
            break;
    }

    byte[] bChunk;
    int rawsize=(int)br.BaseStream.Length-128;
        br.BaseStream.Position=128;
        bChunk = br.ReadBytes(rawsize);


        
        if (pm.Properties[index].formatname == Raw.ParsedBitmap.BitmapType.BITM_FORMAT_A8R8G8B8 ||
            pm.Properties[index].formatname == Raw.ParsedBitmap.BitmapType.BITM_FORMAT_X8R8G8B8 )
                       bChunk = Swizzler.Swizzle(bChunk, pm.Properties[index].width, pm.Properties[index].height, -1, 32,false);
        if (pm.Properties[index].formatname == Raw.ParsedBitmap.BitmapType.BITM_FORMAT_A1R5G5B5 ||
          pm.Properties[index].formatname == Raw.ParsedBitmap.BitmapType.BITM_FORMAT_R5G6B5 ||
            pm.Properties[index].formatname == Raw.ParsedBitmap.BitmapType.BITM_FORMAT_A4R4G4B4 )
            bChunk = Swizzler.Swizzle(bChunk, pm.Properties[index].width, pm.Properties[index].height, -1, 16, false);

        Maps.map[m.mapnumber].BW.BaseStream.Position=m.raw.rawChunks[index].offset;
         Maps.map[m.mapnumber].BW.Write(bChunk);




         br.Close();

         Maps.map[m.mapnumber].CloseMap();
         Maps.map[m.mapnumber].OpenMap(MapTypes.Internal);
         Maps.map[m.mapnumber].BW.BaseStream.Position = m.offset + pm.headersize;
         pm.Properties[index].Write(ref Maps.map[m.mapnumber].BW);
         Maps.map[m.mapnumber].BW.BaseStream.Position = m.offset + m.raw.rawChunks[index].pointerMetaOffset+24;
         Maps.map[m.mapnumber].BW.Write(rawsize);
         Maps.map[m.mapnumber].CloseMap();
    }
    }
    
}
