using System;
using System.Windows.Forms.Layout;
using System.Windows.Forms;
using System.Collections;
using System.Collections.Generic;
using System.Drawing;

namespace entity
{
	/// <summary>
	/// Summary description for FormFunctions.
	/// </summary>
	public class FormFunctions
	{
		
		public List<ListViewItem> ListItems=new List<ListViewItem>();

        public enum MetaView
        {
            FolderView,
            TagTypeView
        }

		public void AddMetasToTreeView(Map map,System.Windows.Forms.TreeView tv,MetaView mv,bool checkbox)
		{
            if (mv == MetaView.TagTypeView)
            {
                AddMetaByTagTypeToTreeView(ref map, ref tv, false);
            }
            else
            {
                AddMetaByFolderToTreeView(ref map, ref tv,checkbox);
            }
        }


        /// <summary>
        /// /thanks to ch2r for this code for the folderview
        /// </summary>
        #region File List Code
        Hashtable root = new Hashtable();
        Hashtable curDirectory;
        private void AddMetaByFolderToTreeView(ref Map map, ref System.Windows.Forms.TreeView tv,bool checkbox)
        {
            string[] steps;
            string step;
            Hashtable currentDir;
            TreeNode currentNode;
            string path;

            
            // Clear current treeview
            tv.Nodes.Clear();
            tv.Sorted = false;
            tv.Nodes.Add(FolderNode(map.filePath, ""));
            tv.Nodes[0].Checked = true;
            if (checkbox == true) { tv.CheckBoxes = true; }
            for (int x = 0; x < map.IndexHeader.metaCount;x++ )
            {
                string temp = map.FileNames.Name[x];
                steps = temp.Split('\\');
                currentDir = root;
                currentNode = tv.Nodes[0];
                path = "";

                for (int i = 0; i < (steps.Length - 1); i++)
                {
                    step = steps[i];
                    path = path + step + "\\";
                    if (!currentDir.ContainsKey(step))
                    {
                        currentDir.Add(step, new Hashtable());
                        TreeNode tempx = FolderNode(step, path);
                        tempx.Checked = checkbox;
                        currentNode.Nodes.Add(tempx);
                    }
                    currentDir = (Hashtable)currentDir[step];
                    currentNode = FindNode(currentNode, path);
                }

               currentDir.Add(steps[steps.GetUpperBound(0)] + "." + map.MetaInfo.TagType[x], x);
               TreeNode tn = FolderNode(steps[steps.GetUpperBound(0)] + "." + map.MetaInfo.TagType[x], map.FileNames.Name[x] + "." + map.MetaInfo.TagType[x]);
               tn.Checked = checkbox;
               currentNode.Nodes.Add(tn);
                 // MessageBox.Show(path);
            }

            tv.Nodes[0].Expand();
            tv.Sorted = true;

        }

        private TreeNode FolderNode(string text, string path)
        {
            TreeNode node = new TreeNode(text);
            node.Tag = path;
            return node;
        }

        private TreeNode FindNode(TreeNode root, string path)
        {
            for (int i = 0; i < root.Nodes.Count; i++)
            {
                if ((string)root.Nodes[i].Tag == path)
                {
                    return root.Nodes[i];
                }
                if (path.StartsWith((string)root.Nodes[i].Tag))
                {
                    return FindNode(root.Nodes[i], path);
                }
            }
            return root;
        }
        #endregion


        private void AddMetaByTagTypeToTreeView(ref Map map,ref System.Windows.Forms.TreeView tv,bool checkbox)
        {
			tv.Nodes.Clear();
			tv.Sorted=false;
            if (checkbox == true) { tv.CheckBoxes = true; }	
            Hashtable temptable = new Hashtable();
            IEnumerator i=map.MetaInfo.TagTypes.Keys.GetEnumerator();
            while (i.MoveNext())
            {
            temptable.Add((string)i.Current, tv.Nodes.Count);
            tv.Nodes.Add((string)i.Current);
            
			}
            
			for (int x=0;x<map.IndexHeader.metaCount;x++)
			{
                object tempobj = temptable[map.MetaInfo.TagType[x]];
                TreeNode tn = new TreeNode(map.FileNames.Name[x]);
                tn.Checked = checkbox;
                tv.Nodes[(int)tempobj].Nodes.Add(tn);
				
			}
			tv.Sorted=true;
			
		}

        public void AddColumnsToListView( System.Windows.Forms.ListView templv, Meta.ItemType type)
        {
            templv.Columns.Clear();
            switch (type)
            {
                case Meta.ItemType.Reflexive:
                    templv.Columns.Add("Description", 120, HorizontalAlignment.Center);

                    templv.Columns.Add("Offset", 80, HorizontalAlignment.Center);

                    templv.Columns.Add("Translation", 80, HorizontalAlignment.Center);
                    templv.Columns.Add("ChunkCount", 80, HorizontalAlignment.Center);
                    templv.Columns.Add("ChunkSize", 80, HorizontalAlignment.Center);
                    templv.Columns.Add("Points To Tag Type", 80, HorizontalAlignment.Center);
                    templv.Columns.Add("Points To Tag Name", 180, HorizontalAlignment.Center);
                    templv.Columns.Add("In Tag Type", 80, HorizontalAlignment.Center);
                    templv.Columns.Add("In Tag Name", 180, HorizontalAlignment.Center);
                    break;
                case Meta.ItemType.Ident:
                    templv.Columns.Add("Description", 120, HorizontalAlignment.Center);
                    templv.Columns.Add("Offset", 80, HorizontalAlignment.Center);
                    templv.Columns.Add("Points To Tag Type", 80, HorizontalAlignment.Center);
                    templv.Columns.Add("Points To Tag Name", 180, HorizontalAlignment.Center);
                    templv.Columns.Add("In Tag Type", 80, HorizontalAlignment.Center);
                    templv.Columns.Add("In Tag Name", 1800, HorizontalAlignment.Center);
                    break;
                case Meta.ItemType.String:
                    templv.Columns.Add("Description", 120, HorizontalAlignment.Center);
                    templv.Columns.Add("Offset", 80, HorizontalAlignment.Center);
                    templv.Columns.Add("Name", 120, HorizontalAlignment.Center);
                    templv.Columns.Add("In Tag Type", 80, HorizontalAlignment.Center);
                    templv.Columns.Add("In Tag Name", 180, HorizontalAlignment.Center);
                    break;
            }

        }

		public void AddReferencesToListView(Meta meta,System.Windows.Forms.ListView templv,Meta.ItemType type)
		{
            if (meta == null) { return; }
            
            templv.VirtualMode = true;
            templv.RetrieveVirtualItem -= new RetrieveVirtualItemEventHandler(templv_RetrieveVirtualItem);
            templv.BeginUpdate();
            templv.AutoScrollOffset = new System.Drawing.Point(0,0);
            ListItems.Clear();
         
            templv.View = View.Details;
            
			
            ListViewItem l;
			for (int x=0;x<meta.items.Count;x++)
			{
				Meta.Item i=(Meta.Item) meta.items[x];
			
				if (i.type!=type){continue;}
				switch (i.type)
				{
					case Meta.ItemType.Reflexive:
					Meta.Reflexive r=(Meta.Reflexive)i;
                    l = new ListViewItem(r.description);
                    
                    l.SubItems.Add(r.offset.ToString());
					l.SubItems.Add(r.translation.ToString());
					l.SubItems.Add(r.chunkcount.ToString());
					l.SubItems.Add(r.chunksize.ToString());
					l.SubItems.Add(r.pointstotagtype);
					l.SubItems.Add(r.pointstotagname);
					l.SubItems.Add(r.intagtype);
					l.SubItems.Add(r.intagname);
                    ListItems.Add(l);
					break;
					case Meta.ItemType.Ident:
					Meta.Ident id=(Meta.Ident)meta.items[x];
                    l = new ListViewItem(id.description);
              		l.SubItems.Add(id.offset.ToString());
					l.SubItems.Add(id.pointstotagtype);
					l.SubItems.Add(id.pointstotagname);
					l.SubItems.Add(id.intagtype);
					l.SubItems.Add(id.intagname);
                    ListItems.Add(l);
					break;
					case Meta.ItemType.String:
					Meta.String s=(Meta.String)meta.items[x];
                    l = new ListViewItem(s.description);
        			l.SubItems.Add(s.offset.ToString());
					l.SubItems.Add(s.name);
					l.SubItems.Add(s.intagtype);
					l.SubItems.Add(s.intagname);
                    ListItems.Add(l);
					break;
             
				}
                //System.Windows.Forms.Layout.LayoutEngine
                //Application.DoEvents();
                RichTextBox rr=new RichTextBox();
               

			}


            try
            {
                templv.VirtualListSize = ListItems.Count;
            }
            catch
            {
            }
            templv.RetrieveVirtualItem += new RetrieveVirtualItemEventHandler(templv_RetrieveVirtualItem);
            templv.EndUpdate();
            Application.DoEvents();
		}

        void templv_RetrieveVirtualItem(object sender, RetrieveVirtualItemEventArgs e)
        {
            e.Item = ListItems[e.ItemIndex];
        }




        public void AddMetasToListView(ref Map map, System.Windows.Forms.ListView templv)
        {
           

            templv.VirtualMode = true;
            templv.RetrieveVirtualItem -= new RetrieveVirtualItemEventHandler(metalist_RetrieveVirtualItem);
            templv.BeginUpdate();
            templv.AutoScrollOffset = new System.Drawing.Point(0, 0);
            ListItems.Clear();
            templv.Columns.Clear();
            templv.View = View.Details;
            templv.CheckBoxes = true;
            templv.Columns.Add("Size", 80, HorizontalAlignment.Center);
            templv.Columns.Add("Tag Type", 80, HorizontalAlignment.Center);
            templv.Columns.Add("Tag Name", 180, HorizontalAlignment.Center);
            //templv.Sort = false;
           
            ListViewItem l;
            for (int x = 0; x < map.IndexHeader.metaCount; x++)
            {
                        l = new ListViewItem(map.MetaInfo.Size[x].ToString());

                        l.SubItems.Add(map.MetaInfo.TagType[x]);
                        l.SubItems.Add(map.FileNames.Name[x]);
                        l.Checked=true;
                
                ListItems.Add(l);
                        
                }
                

            


            try
            {
                templv.VirtualListSize = ListItems.Count;
            }
            catch
            {
            }
            //templv.
            templv.RetrieveVirtualItem += new RetrieveVirtualItemEventHandler(templv_RetrieveVirtualItem);
            templv.EndUpdate();
         
            Application.DoEvents();
        }

        void metalist_RetrieveVirtualItem(object sender, RetrieveVirtualItemEventArgs e)
        {
            e.Item = ListItems[e.ItemIndex];
        }


  

	}
}
