using System;
using System.IO;
using System.Collections;
using System.Windows.Forms;
using System.Xml;
namespace entity
{
	public class LayOutSorter:IComparer
	{
		public int Compare(object x,object y)
		{
			return ((LayOutChunk)x).startoffset.CompareTo(((LayOutChunk)y).startoffset);
		}
	}
	public class MapLayout
	{
		#region	Fields
		public ArrayList chunks=new ArrayList(0);
		public ArrayList metas=new ArrayList(0);
		#endregion
		#region	Constructor
		public MapLayout()
		{
		}
		#endregion
		#region	Method
		public void ReadChunks(int mapnumber)
		{
			Maps.map[mapnumber].OpenMap(MapTypes.Internal);
			for (int x=0;x<chunks.Count;x++)
			{
				LayOutChunk l=(LayOutChunk) chunks[x];
				l.Read(mapnumber);
                GC.Collect(0);
                GC.WaitForPendingFinalizers();
			}
			Maps.map[mapnumber].CloseMap();
		}
		public void SaveToXml(string path,int mapnumber)
		{
			XmlTextWriter xtw=new XmlTextWriter(path,System.Text.Encoding.Default);
			xtw.Formatting=Formatting.Indented;
			xtw.WriteStartElement("MapLayOut");
			xtw.WriteAttributeString("Map",Maps.map[mapnumber].filePath);
            int tempindex=path.LastIndexOf("\\");
            string tempfilepath = path.Substring(0,tempindex) + "\\" + path.Substring(tempindex+1, path.Length - tempindex-1) + " - Meta Chunks\\";
            Directory.CreateDirectory(tempfilepath);
            Maps.map[mapnumber].OpenMap(MapTypes.Internal);
			for	(int x=0;x<chunks.Count;x++)
			{
				LayOutChunk	c=(LayOutChunk)	chunks[x];
             //   c.Read(mapnumber);
              
                FileStream FS = new FileStream(tempfilepath + "MapMetaChunk[" + x.ToString() + "] - "+c.rawType.ToString()+".meta", FileMode.Create);
                BinaryWriter BW = new BinaryWriter(FS);
         //     BW.Write(c.MS.ToArray());
           //     c.Write(BW, 0);
  
                Maps.map[mapnumber].BR.BaseStream.Position=c.startoffset;
                Maps.map[mapnumber].BufferReadWrite(ref Maps.map[mapnumber].BR, ref BW,c.size );
                BW.Close();
                FS.Close();
               
               
				xtw.WriteStartElement("LayOutChunk");
				xtw.WriteAttributeString("Type",c.rawType.ToString());	
				xtw.WriteAttributeString("StartOffset",c.startoffset.ToString("X"));	
				xtw.WriteAttributeString("EndOffset",c.endoffset.ToString("X"));
				xtw.WriteAttributeString("Size",c.size.ToString("X"));
				
				for	(int xx=0;xx<c.rawPieces.Count;xx++)
				{
					RawInfoChunk cc=(RawInfoChunk) c.rawPieces[xx];
					xtw.WriteStartElement("RawDataPiece");
					xtw.WriteAttributeString("Type",cc.rawType.ToString());	
					xtw.WriteAttributeString("Offset",cc.offset.ToString("X"));	
					xtw.WriteAttributeString("Size",cc.size.ToString("X"));
					xtw.WriteAttributeString("PointerOffset",cc.offsetOfPointer.ToString("X"));
					xtw.WriteEndElement();
				}


				xtw.WriteEndElement();
			}
			
			xtw.WriteEndElement();
			xtw.Close();
            Maps.map[mapnumber].CloseMap();
		}
		public void	SortChunksByOffset()
		{
 		 chunks.Sort(new LayOutSorter());
		}
		public void SortRawByOffset()
		{
			for (int x=0;x<chunks.Count;x++)
			{
				LayOutChunk l=(LayOutChunk) chunks[x];
				l.rawPieces.Sort(new RawInfoSorter());
			}
		}

		public int FindByType(RawDataContainerType type)
		{
			for	(int x=0;x<chunks.Count;x++)
			{
				LayOutChunk	c=(LayOutChunk)	chunks[x];
				if (c.rawType==type)
				{
					return x;
				}
			}
			return -1;
		}
        public int FindByType(RawDataContainerType type,int index)
        {
            int tempc = 0;
            for (int x = 0; x < chunks.Count; x++)
            {
                LayOutChunk c = (LayOutChunk)chunks[x];
                if (c.rawType == type)
                {
                    tempc += 1;
                    if (tempc == index) { return x; }
                }
            }
            return -1;
        }
		#endregion
	}


	public class RawInfoSorter:IComparer
	{
		public int Compare(object x,object y)
		{
			return ((RawInfoChunk)x).offset.CompareTo(((RawInfoChunk)y).offset);
		}
	}

	public class RawInfoChunk
	{
		#region	Fields
		public RawDataType rawType;
		public uint offset;
		public int offsetOfPointer;
		public int size;
		public MapTypes location;
		#endregion
	}

	public class LayOutChunk
	{
		#region	Fields
		public RawDataContainerType	rawType;
		public ArrayList rawPieces=new ArrayList(0);
		public int startoffset;
		public int endoffset;
		public int size;
		public MemoryStream MS;
		public int shift;
		#endregion
		#region	Constructor
		public LayOutChunk(int mapsize)
		{
			startoffset=mapsize;
			endoffset=0;
		}
		#endregion
		#region	Constructor
		public void Read(int mapnumber)
		{
            Maps.map[mapnumber].BR.BaseStream.Position = startoffset;
            MS = new MemoryStream(size);
            BinaryWriter BW = new BinaryWriter(MS);
            //MS.Write(Maps.map[mapnumber].BR.ReadBytes(size), 0, size);


            int buffersize = (int)0x1900000;
            int currread = 0;
            do
            {
                int tempsize=(size-currread);
                int buffsize=buffersize;
                if (tempsize < buffersize )
                {
                    buffsize=size-currread;
                    BW.Write(Maps.map[mapnumber].BR.ReadBytes(buffsize));
                    break;
                }
                BW.Write(Maps.map[mapnumber].BR.ReadBytes(buffsize));
                currread += buffersize;
                GC.Collect(0);
                GC.WaitForPendingFinalizers();
            }
            while (currread != -1);

  
		}
        public void Write(BinaryWriter BW,int offset)
        {
            BW.BaseStream.Position = offset;
            BinaryReader BR = new BinaryReader(MS);



            int size = (int)MS.Length;
            int buffersize = (int)0x1400000;
            int currread = 0;
            do
            {
                int tempsize = (size - currread);
                int buffsize = buffersize;
                if (tempsize < buffersize)
                {
                    buffsize = size - currread;
                    BW.Write(BR.ReadBytes(buffsize));
                    break;
                }
                BW.Write(BR.ReadBytes(buffsize));
                currread += buffersize;
                GC.Collect(0);
                GC.WaitForPendingFinalizers();
            }
            while (currread != -1);


        }
		#endregion
	}









	public class MapAnalyzer
	{
		#region	Fields
		#endregion
		#region	Constructor
		public MapAnalyzer()
		{
		
		}
		#endregion
		#region	Methods
		
	
		public MapLayout ScanMapForLayOut(int mapnumber, bool addexternalchunks)
		{
			MapLayout layout=new MapLayout();
			MetaScanner metaScanner=new MetaScanner();
		


			for (int x=0;x<Maps.map[mapnumber].BSP.sbsp.Length;x++)
			{
				LayOutChunk l=new LayOutChunk(0);
				l.rawType=RawDataContainerType.BSPMeta;
                l.startoffset = Maps.map[mapnumber].BSP.sbsp[x].offset;
                l.endoffset = Maps.map[mapnumber].BSP.sbsp[x].offset + Maps.map[mapnumber].BSP.sbsp[x].size;
                l.size = Maps.map[mapnumber].BSP.sbsp[x].size;
				layout.chunks.Add(l);
			}

            LayOutChunk lo = new LayOutChunk(0);
            lo.rawType = RawDataContainerType.Header;
            lo.startoffset = 0;
            lo.size = 2048;
            lo.endoffset = lo.startoffset + lo.size;
            layout.chunks.Add(lo);

            if (Maps.map[mapnumber].HaloVersion == Map.HaloVersionEnum.Halo2)
            {
                lo = new LayOutChunk(0);
                lo.rawType = RawDataContainerType.StringsIndex;
                lo.startoffset = Maps.map[mapnumber].MapHeader.offsetToStringIndex;
                lo.size = Maps.map[mapnumber].MapHeader.scriptReferenceCount * 4;
                lo.size += Maps.map[mapnumber].Functions.Padding(lo.size, 512);
                lo.endoffset = lo.startoffset + lo.size;
                layout.chunks.Add(lo);


                lo = new LayOutChunk(0);
                lo.rawType = RawDataContainerType.Strings1;
                lo.startoffset = Maps.map[mapnumber].MapHeader.offsetToStringNames1;
                lo.size = Maps.map[mapnumber].MapHeader.scriptReferenceCount * 128;
                lo.size += Maps.map[mapnumber].Functions.Padding(lo.size, 512);
                lo.endoffset = lo.startoffset + lo.size;
                layout.chunks.Add(lo);

                lo = new LayOutChunk(0);
                lo.rawType = RawDataContainerType.Strings2;
                lo.startoffset = Maps.map[mapnumber].MapHeader.offsetToStringNames2;
                lo.size = Maps.map[mapnumber].MapHeader.sizeOfScriptReference;
                lo.size += Maps.map[mapnumber].Functions.Padding(lo.size, 512);
                lo.endoffset = lo.startoffset + lo.size;
                layout.chunks.Add(lo);


                lo = new LayOutChunk(0);
                lo.rawType = RawDataContainerType.Crazy;
                lo.startoffset = Maps.map[mapnumber].MapHeader.offsetToCrazy;
                lo.size = Maps.map[mapnumber].MapHeader.sizeOfCrazy;
                lo.size += Maps.map[mapnumber].Functions.Padding(lo.size, 512);
                lo.endoffset = lo.startoffset + lo.size;
                layout.chunks.Add(lo);


                lo = new LayOutChunk(0);
                lo.rawType = RawDataContainerType.FileNamesIndex;
                lo.startoffset = Maps.map[mapnumber].MapHeader.offsetTofileIndex;
                lo.size = Maps.map[mapnumber].MapHeader.fileCount * 4;
                lo.size += Maps.map[mapnumber].Functions.Padding(lo.size, 512);
                lo.endoffset = lo.startoffset + lo.size;
                layout.chunks.Add(lo);
            }

			lo=new LayOutChunk(0);
			lo.rawType=RawDataContainerType.FileNames;
            lo.startoffset = Maps.map[mapnumber].MapHeader.offsetTofileNames;
            lo.size = Maps.map[mapnumber].MapHeader.fileNamesSize;
            if (Maps.map[mapnumber].HaloVersion == Map.HaloVersionEnum.Halo2)
            {
                lo.size += Maps.map[mapnumber].Functions.Padding(lo.size, 512);
            }
			lo.endoffset=lo.startoffset+lo.size;
			layout.chunks.Add(lo);
			
			lo=new LayOutChunk(0);
			lo.rawType=RawDataContainerType.MetaIndex;
            lo.startoffset = Maps.map[mapnumber].MapHeader.indexOffset;
            if (Maps.map[mapnumber].HaloVersion == Map.HaloVersionEnum.Halo2)
            {
                lo.size = Maps.map[mapnumber].MapHeader.metaStart;
                lo.size += Maps.map[mapnumber].Functions.Padding(lo.size, 512);
              //Maps.map[mapnumber].MapHeader.fileSize - Maps.map[mapnumber].MapHeader.indexOffset;
            }
            else
            {
                lo.size = Maps.map[mapnumber].MapHeader.offsetTofileNames - Maps.map[mapnumber].MapHeader.indexOffset;
              //    Maps.map[mapnumber].MetaInfo.Offset[Maps.map[mapnumber].IndexHeader.metaCount - 1] + Maps.map[mapnumber].MetaInfo.Size[Maps.map[mapnumber].IndexHeader.metaCount-1];
            }
            
			lo.endoffset=lo.startoffset+lo.size;
			layout.chunks.Add(lo);

         
            lo=new LayOutChunk(0);
			lo.rawType=RawDataContainerType.MetaData;
            lo.startoffset = Maps.map[mapnumber].MetaInfo.Offset[0];

                lo.size = Maps.map[mapnumber].MapHeader.fileSize-lo.startoffset;
                if (Maps.map[mapnumber].HaloVersion != Map.HaloVersionEnum.Halo2)
                {
                    int padding=Maps.map[mapnumber].Functions.Padding(lo.startoffset+lo.size, 4096);
                    lo.size += padding;
                }
			lo.endoffset=lo.startoffset+lo.size;
			layout.chunks.Add(lo);
        

            if (Maps.map[mapnumber].HaloVersion == Map.HaloVersionEnum.Halo2)
            {
                for (int x = 0; x < Maps.map[mapnumber].Unicode.ut.Length; x++)
                {
                    lo = new LayOutChunk(0);
                    lo.rawType = RawDataContainerType.UnicodeNamesIndex;
                    lo.startoffset = Maps.map[mapnumber].Unicode.ut[x].indexOffset;
                    lo.size = Maps.map[mapnumber].Unicode.ut[x].count * 8;
                    lo.size += Maps.map[mapnumber].Functions.Padding(lo.size, 512);
                    lo.endoffset = lo.startoffset + lo.size;
                    layout.chunks.Add(lo);

                    lo = new LayOutChunk(0);
                    lo.rawType = RawDataContainerType.UnicodeNames;
                    lo.startoffset = Maps.map[mapnumber].Unicode.ut[x].tableOffset;
                    lo.size = Maps.map[mapnumber].Unicode.ut[x].tableSize;
                    lo.size += Maps.map[mapnumber].Functions.Padding(lo.size, 512);
                    lo.endoffset = lo.startoffset + lo.size;
                    layout.chunks.Add(lo);
                }
            }
            Maps.map[mapnumber].OpenMap(MapTypes.Internal);
            for (int x = 0; x < Maps.map[mapnumber].IndexHeader.metaCount; x++)
            {

                if (Maps.map[mapnumber].HaloVersion == Map.HaloVersionEnum.Halo2)
                {
                    if (Maps.map[mapnumber].MetaInfo.TagType[x] == "snd!" | Maps.map[mapnumber].MetaInfo.TagType[x] == "ltmp") { continue; }
                    Application.DoEvents();
                }
				Meta m=new Meta();
                m.offset = Maps.map[mapnumber].MetaInfo.Offset[x];
			//checks if type has raw data
            m.rawType = Maps.map[mapnumber].Functions.Meta.CheckForRaw(Maps.map[mapnumber].MetaInfo.TagType[x]);
         

				
	
				if (m.rawType!=RawDataContainerType.Empty)
				{
                    m.raw = Maps.map[mapnumber].Functions.Meta.ReadRaw(x, mapnumber, true);
                    LayOutChunk l = new LayOutChunk(Maps.map[mapnumber].MapHeader.fileSize);

					int tempint=layout.FindByType(m.rawType);
					if (tempint==-1) 
					{
                        l = new LayOutChunk(Maps.map[mapnumber].MapHeader.fileSize);
						l.rawType=m.rawType;
						layout.chunks.Add(l);
						tempint=layout.FindByType(m.rawType);
					}
					else
					{
						l=(LayOutChunk) layout.chunks[tempint];
					}
					for (int y=0;y<m.raw.rawChunks.Count;y++)
					{
						RawDataChunk r=(RawDataChunk) m.raw.rawChunks[y];
                        if (r.offset == -1) { continue; } 
						if (r.rawLocation==MapTypes.Internal)
						{
							RawInfoChunk tempr=new RawInfoChunk();
							tempr.offset=(uint)r.offset;
							tempr.size=r.size;
							tempr.rawType=r.rawDataType;
							tempr.location=r.rawLocation;
							tempr.offsetOfPointer=m.offset+r.pointerMetaOffset;
							l.rawPieces.Add(tempr);
						}
                        else if (addexternalchunks)
                        {
                            RawInfoChunk tempr = new RawInfoChunk();
                            tempr.offset = (uint)r.offset;
                            tempr.size = r.size;
                            tempr.rawType = r.rawDataType;
                            tempr.location = r.rawLocation;
                            tempr.offsetOfPointer = m.offset + r.pointerMetaOffset;
                            l.rawPieces.Add(tempr);
                        }
						
						if (r.offset<l.startoffset && r.rawLocation==MapTypes.Internal)
						{
							l.startoffset=r.offset;
							l.size=l.endoffset-l.startoffset;
							layout.chunks[tempint]=l;
							
						}
						if (r.offset+r.size>l.endoffset && r.rawLocation==MapTypes.Internal)
						{
							l.endoffset=r.offset+r.size;
                            l.endoffset += Maps.map[mapnumber].Functions.Padding(l.endoffset, 512);
							l.size=l.endoffset-l.startoffset;
							layout.chunks[tempint]=l;
							
                        }
				
						
					}
				
				}
                m = null;
                GC.WaitForPendingFinalizers();
                //GC.Collect();
                
			}
            Maps.map[mapnumber].CloseMap() ;

            if (Maps.map[mapnumber].HaloVersion == Map.HaloVersionEnum.Halo2)
            {
                lo = new LayOutChunk(0);
                LayOutChunk templo = (LayOutChunk)layout.chunks[layout.FindByType(RawDataContainerType.Model)];
                lo.rawType = RawDataContainerType.Sound;
                lo.startoffset = 2048;
                lo.size = templo.startoffset - 2048;
                lo.size += Maps.map[mapnumber].Functions.Padding(lo.size, 512);
                lo.endoffset = lo.startoffset + lo.size;
                layout.chunks.Add(lo);
            }

			for (int y=0;y<layout.chunks.Count;y++)
			{
				LayOutChunk l=(LayOutChunk) layout.chunks[y];
				if (l.size==0) {layout.chunks.RemoveAt(y);y--;}
				
			}
			layout.SortChunksByOffset();
			layout.SortRawByOffset();
     
          
			return layout;
		}
		#endregion	

	}
}
