using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Text;
using System.Windows.Forms;

namespace entity.Forms.Meta_Editor
{
    public partial class Bitmask : UserControl
    {
        #region fields
        public int LineNumber;
        private int mapIndex;
        public int chunkOffset;
        public int offsetInMap;
        private object value;
        private int optionNamePadding=50;
        public string EntName = "Error in getting plugin element name";
        private bool isNulledOutReflexive = true;
        private bool[] Bits;
        private int bitCount;
        public object[] Options;
        private bool[] visibleBits;
        #endregion
        #region Constructor
        public Bitmask(string iEntName, int map, int iOffsetInChunk, int iBitCount, object[] ioptions, int iLineNumber)
        {
            this.LineNumber = iLineNumber;
            this.Options = ioptions;
            this.Bits = new bool[iBitCount];
            this.visibleBits = new bool[iBitCount];
            this.bitCount = iBitCount;
            this.chunkOffset = iOffsetInChunk;
            mapIndex = map;
            this.EntName = iEntName;
            InitializeComponent();
            this.getLongestName();
            this.MakeControls();
            this.Size = this.PreferredSize;
            this.Dock = DockStyle.Top;
            this.Controls[0].Text = EntName;
            this.AutoSize = true;
        }
        #endregion
        #region Methods/Functions
        #region Setting up the bitmask controls
        //Cookie for who ever figures out what this method does
        private void getLongestName()
        {
            if (this.Options == null)
                return;
            for (int counter = 0; counter < this.Options.Length; counter++)
            {
                this.checkBox1.Text = ((IFPIO.Option)this.Options[counter]).name;
                if (this.checkBox1.PreferredSize.Width>this.optionNamePadding)
                    this.optionNamePadding = this.checkBox1.PreferredSize.Width+5;
            }
        }
        //Load the the int into bits whose values (1 or 0) are stored in a bool array
        private void BitsToBool()
        {
            uint tempANDInt = 1;
            uint tempvalue = Convert.ToUInt32(this.value);
            for (int counter = 0; counter < this.bitCount; counter++)
            {
                this.Bits[counter] = (tempANDInt & tempvalue) == tempANDInt ? true : false;
                tempANDInt <<= 1;
            }
        }
        //Make the checkboxes, give them names and line them up
        private void MakeControls()
        {
            this.Controls[0].Controls.Clear();

            for (int counter = 0; counter < this.bitCount; counter++)
            {
                //Making the checkbox
                CheckBox tempcheckBox1 = new CheckBox();
                tempcheckBox1.AutoSize = true;
                tempcheckBox1.Size = new System.Drawing.Size(80, 17);
                tempcheckBox1.Text = "bit " + counter.ToString();
                tempcheckBox1.Text.PadLeft(this.optionNamePadding);
                tempcheckBox1.UseVisualStyleBackColor = true;
                tempcheckBox1.CheckedChanged += new EventHandler(checkBox1_CheckedChanged);
                tempcheckBox1.Tag = counter.ToString();
                this.visibleBits[counter] = false;
                //Fill in names
                if (this.Options != null)
                {
                    for (int counter2 = 0; counter2 < this.Options.Length; counter2++)
                    {
                        if (((IFPIO.Option)this.Options[counter2]).value == counter)
                        {
                            this.visibleBits[counter] = true;
                            tempcheckBox1.Text = ((IFPIO.Option)this.Options[counter2]).name;
                               
                            break;
                        }
                    }
                }
                tempcheckBox1.Visible = false;
                this.Controls[0].Controls.Add(tempcheckBox1);
            }
            //Place the check box in the right spot
            int down = 19;
            int over = 6;
            int CounterForPlacementKeeping = 0;
            for (int counter = 0; counter < this.bitCount; counter++)
            {
                if (this.visibleBits[counter] ==false)
                    if (Meta_Editor.MetaEditor.ShowInvisibles == false)
                        continue;                    
                this.Controls[0].Controls[counter].Location = new Point(over, down);
                over += this.optionNamePadding;
                this.Controls[0].Controls[counter].Visible = true;                
                if ((CounterForPlacementKeeping+1) % 4 == 0)
                {
                    down += 23;
                    over = 6;
                }
                CounterForPlacementKeeping++;
            }
        }
        private void BoolsToControls()
        {
            for (int counter = 0; counter < this.bitCount; counter++)
            {
                ((CheckBox)this.Controls[0].Controls[counter]).Checked = this.Bits[counter];
            }
        }
        #endregion
        #region Load and Save
        public void Populate(int iOffset)
        {
            this.isNulledOutReflexive = false;
            //Open up map
            bool openedMap = false;
            if (Maps.map[mapIndex].isOpen == false)
            {
                Maps.map[mapIndex].OpenMap(MapTypes.Internal);
                openedMap = true;
            }
            //set offsets
            Maps.map[mapIndex].BR.BaseStream.Position = iOffset + this.chunkOffset;
            this.offsetInMap = iOffset + this.chunkOffset;
            //Decide how big the bitmask is
            switch (this.bitCount)
            {
                case 8:
                    {
                        this.value = Maps.map[mapIndex].BR.ReadByte();
                        break;
                    }
                case 16:
                    {
                        this.value = Maps.map[mapIndex].BR.ReadUInt16();
                        break;
                    }
                case 32:
                    {
                        this.value = Maps.map[mapIndex].BR.ReadUInt32();
                        break;
                    }
            }
            //convert this.value (an object) into a bool array, then update the checkboxes with that bool array
            BitsToBool();
            BoolsToControls();
            if (openedMap == true)
                Maps.map[mapIndex].CloseMap();            
        }
        public void Save()
        {
            if (this.isNulledOutReflexive == true)
                return;
            bool openedMap = false;
            if (Maps.map[mapIndex].isOpen == false)
            {
                Maps.map[mapIndex].OpenMap(MapTypes.Internal);
                openedMap = true;
            }
            Maps.map[mapIndex].BW.BaseStream.Position = this.offsetInMap;
            try
            {
                uint intToShift = 1;
                uint tempIntToWrite = 0;
                for (int counter = 0; counter < this.bitCount; counter++)
                {
                    if (this.Bits[counter] == true)
                    {
                        tempIntToWrite += intToShift;
                    }
                    intToShift <<= 1;
                }
                this.value = tempIntToWrite;
                switch (this.bitCount)
                {
                    case 8:
                        {
                            Maps.map[mapIndex].BW.Write(Convert.ToByte(this.value));
                            break;
                        }
                    case 16:
                        {
                            Maps.map[mapIndex].BW.Write(Convert.ToUInt16(this.value));
                            break;
                        }
                    case 32:
                        {
                            Maps.map[mapIndex].BW.Write(Convert.ToUInt32(this.value));
                            break;
                        }
                }
            }
            catch
            {
                MessageBox.Show("Man the hatches! Error in writing bitmask info to map!");
            }
            if (openedMap == true)
                Maps.map[mapIndex].CloseMap();
        }
        public void Poke()
        {
            uint Address = (uint)(this.offsetInMap + Maps.map[mapIndex].SelectedMeta.magic);
            switch (this.bitCount)
            {
                case 8:
                    {
                        Feline_rawr.RTH_Imports.Poke(Address, (uint)Convert.ToByte(this.value), 8);
                        break;
                    }
                case 16:
                    {
                        Feline_rawr.RTH_Imports.Poke(Address, (uint)Convert.ToInt16(this.value), 16);
                        break;
                    }
                case 32:
                    {
                        Feline_rawr.RTH_Imports.Poke(Address, (uint)Convert.ToInt32(this.value), 32);
                        break;
                    }
            }
        }
        #endregion
        #region Events
        private void checkBox1_CheckedChanged(object sender, EventArgs e)
        {
            //make the bool array match the check boxes
            this.Bits[Convert.ToInt32(((CheckBox)sender).Tag)] = ((CheckBox)sender).Checked;
        }
        #endregion
        public void SetFocus(int LineToCheck)
        {
            if (this.LineNumber == LineToCheck)
                this.Focus();
        }
        #endregion
    }
}
