using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Text;
using System.Windows.Forms;

namespace entity.Forms.Meta_Editor
{
    public partial class DataValues : UserControl
    {        
        public enum ENTType
        {            
            Int32,
            UInt32,
            Int16,
            Uint16,
            Float,
            b,
            ub,
            undefined,
            string32,
            unicodestring64,
            string256,
            unicodestring256,
            sid,
            ident,
            nothing,
            bitmask32,
            bitmask16
        }
        public int LineNumber;
        private int mapIndex;
        public int chunkOffset;
        public int offsetInMap;
        public ENTType ValueType;
        public object Value;
        public string EntName = "Error in getting plugin element name";
        private bool isNulledOutReflexive = true;
        public DataValues(string iEntName, int map, int iOffsetInChunk, IFPIO.ObjectEnum type, int iLineNumber)
        {
            this.LineNumber = iLineNumber;            
            this.chunkOffset = iOffsetInChunk;
            mapIndex = map;
            this.EntName = iEntName;
            InitializeComponent();
            this.Dock = DockStyle.Top;
            this.Controls[0].Text = EntName;
            switch (type)
            {
                case IFPIO.ObjectEnum.Short:
                    {
                        this.Controls[2].Text = "Int16";
                        this.ValueType = ENTType.Int16;
                        break;
                    }
                case IFPIO.ObjectEnum.Int:
                    {
                        this.Controls[2].Text = "Int32";
                        this.ValueType = ENTType.Int32;
                        break;
                    }
                case IFPIO.ObjectEnum.UShort:
                    {
                        this.Controls[2].Text = "UInt16";
                        this.ValueType = ENTType.Uint16;
                        break;
                    }
                case IFPIO.ObjectEnum.UInt:
                    {
                        this.Controls[2].Text = "UInt32";
                        this.ValueType = ENTType.UInt32;
                        break;
                    }
                case IFPIO.ObjectEnum.Float:
                    {
                        this.Controls[2].Text = "Float";
                        this.ValueType = ENTType.Float;
                        break;
                    }
                case IFPIO.ObjectEnum.Unknown:
                    {
                        this.Controls[1].Location = new Point(65,6);
                        this.Controls[0].Location = new Point(171, 9);
                        this.Controls[2].Text = "Undefined";
                        this.ValueType = ENTType.undefined;
                        break;
                    }
                case IFPIO.ObjectEnum.Byte:
                    {
                        this.Controls[2].Text = "Byte";
                        this.ValueType = ENTType.b;
                        break;
                    }
            }
            this.AutoSize = false;
        }
        public void Populate(int iOffset)
        {
            this.isNulledOutReflexive = false;
            bool openedMap = false;
            if (Maps.map[mapIndex].isOpen == false)
            {
                Maps.map[mapIndex].OpenMap(MapTypes.Internal);
                openedMap = true;
            }
            Maps.map[mapIndex].BR.BaseStream.Position = iOffset + this.chunkOffset;
            this.offsetInMap = iOffset + this.chunkOffset;
            switch (ValueType)
            {
                case ENTType.Int16:
                    {
                        this.Value = Maps.map[mapIndex].BR.ReadInt16();
                        break;
                    }
                case ENTType.Int32:
                    {
                        this.Value = Maps.map[mapIndex].BR.ReadInt32();
                        break;
                    }
                case ENTType.Uint16:
                    {
                        this.Value = Maps.map[mapIndex].BR.ReadUInt16();
                        break;
                    }
                case ENTType.UInt32:
                    {
                        this.Value = Maps.map[mapIndex].BR.ReadUInt32();
                        break;
                    }
                case ENTType.Float:
                    {
                        this.Value = Maps.map[mapIndex].BR.ReadSingle();
                        break;
                    }
                case ENTType.undefined:
                    {
                        this.Value = Maps.map[mapIndex].BR.ReadSingle();
                        break;
                    }
                case ENTType.b:
                    {
                        this.Value = Maps.map[mapIndex].BR.ReadByte();
                        break;
                    }
            }
            this.Controls[1].Text = this.Value.ToString();
            if (openedMap == true)
                Maps.map[mapIndex].CloseMap();
        }
        public void Save()
        {
            if (this.isNulledOutReflexive == true)
                return;
            bool openedMap = false;
            if (Maps.map[mapIndex].isOpen == false)
            {
                Maps.map[mapIndex].OpenMap(MapTypes.Internal);
                openedMap = true;
            }
            try
            {
                Maps.map[mapIndex].BW.BaseStream.Position = this.offsetInMap;
                switch (ValueType)
                {
                    case ENTType.Int16:
                        {
                            Maps.map[mapIndex].BW.Write(Convert.ToInt16(this.Value));
                            break;
                        }
                    case ENTType.Int32:
                        {
                            Maps.map[mapIndex].BW.Write(Convert.ToInt32(this.Value));
                            break;
                        }
                    case ENTType.Uint16:
                        {
                            Maps.map[mapIndex].BW.Write(Convert.ToUInt16(this.Value));
                            break;
                        }
                    case ENTType.UInt32:
                        {
                            Maps.map[mapIndex].BW.Write(Convert.ToUInt32(this.Value));
                            break;
                        }
                    case ENTType.Float:
                        {
                            Maps.map[mapIndex].BW.Write(Convert.ToSingle(this.Value));
                            break;
                        }
                    case ENTType.undefined:
                        {
                            Maps.map[mapIndex].BW.Write(Convert.ToSingle(this.Value));
                            break;
                        }
                    case ENTType.b:
                        {
                            Maps.map[mapIndex].BW.Write(Convert.ToByte(this.Value));
                            break;
                        }
                }
            }
            catch
            {
                MessageBox.Show("Something is wrong with this " +this.ValueType.ToString()+ this.EntName + " Offset " + this.chunkOffset.ToString());
            }
            if (openedMap == true)
                Maps.map[mapIndex].CloseMap();
        }
        public void Poke()
        {
            uint Address = (uint)(this.offsetInMap+Maps.map[mapIndex].SelectedMeta.magic);
            switch (ValueType)
            {
                case ENTType.Int16:
                    {
                        Feline_rawr.RTH_Imports.Poke(Address, (uint)Convert.ToInt16(this.Value), 16);
                        break;
                    }
                case ENTType.Int32:
                    {
                        Feline_rawr.RTH_Imports.Poke(Address, (uint)Convert.ToInt32(this.Value),32);
                        break;
                    }
                case ENTType.Uint16:
                    {
                        Feline_rawr.RTH_Imports.Poke(Address, (uint)Convert.ToUInt16(this.Value),16);
                        break;
                    }
                case ENTType.UInt32:
                    {
                        Feline_rawr.RTH_Imports.Poke(Address, (uint)Convert.ToUInt32(this.Value),32);
                        break;
                    }
                case ENTType.Float:
                    {
                        Feline_rawr.RTH_Imports.Poke(Address, (uint)Feline_rawr.RTH_Imports.ConvertFloat(Convert.ToSingle(this.Value)), 32);
                        break;
                    }
                case ENTType.undefined:
                    {
                        Feline_rawr.RTH_Imports.Poke(Address, (uint)Feline_rawr.RTH_Imports.ConvertFloat(Convert.ToSingle(this.Value)), 32);
                        break;
                    }
                case ENTType.b:
                    {
                        Feline_rawr.RTH_Imports.Poke(Address, (uint)Convert.ToByte(this.Value),8);
                        break;
                    }
            }
        }
        private void textBox1_TextChanged(object sender, EventArgs e)
        {
            this.Value = ((TextBox)sender).Text;
        }
        public void SetFocus(int LineToCheck)
        {
            if (this.LineNumber == LineToCheck)
                this.Focus();
        }
    }
}
