using System;
using System.IO;
using System.Collections;
using System.Collections.Generic;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using D3D=Microsoft.DirectX.Direct3D;
using System.Windows.Forms;
using System.Drawing;

namespace entity.Renderer
{
    public class Renderer : IDisposable
    {
        public Device device = null; // Our rendering device
        public bool pause = false;

        public bool lighting = false;
        public void Dispose()
        {
        }
        public static int VectorToRgba(Vector3 v, float height)
        {
            int r = (int)(127.0f * v.X + 128.0f);
            int g = (int)(127.0f * v.Y + 128.0f);
            int b = (int)(127.0f * v.Z + 128.0f);
            int a = (int)(255.0f * height);

            return ((a << 24) + (r << 16) + (g << 8) + (b << 0));
        }
        public void CreateDevice(Control renderpanel)
        {


            PresentParameters presentParams = new PresentParameters();

            presentParams.Windowed = true; // We don't want to run fullscreen
            /*
            presentParams.BackBufferCount = 1; //Number of backbuffers to create
            presentParams.BackBufferFormat = Manager.Adapters.Default.CurrentDisplayMode.Format; //The current format of the display device
            presentParams.BackBufferWidth = renderpanel.Width;
            presentParams.BackBufferHeight = renderpanel.Height;
             */
            presentParams.SwapEffect = SwapEffect.Discard; // Discard the frames 
            presentParams.AutoDepthStencilFormat = DepthFormat.D16; //24 bits for the depth and 8 bits for the stencil
            presentParams.EnableAutoDepthStencil = true; //Let direct3d handle the depth buffers for the application
            presentParams.PresentationInterval = PresentInterval.One;

            IEnumerator i = Manager.Adapters.GetEnumerator();

            while (i.MoveNext())
            {
                AdapterInformation ai = i.Current as AdapterInformation;

                int adapterOrdinal = ai.Adapter;



                Caps hardware = Manager.GetDeviceCaps(adapterOrdinal, DeviceType.Hardware);
         
                CreateFlags flags = CreateFlags.SoftwareVertexProcessing;

                if (hardware.DeviceCaps.SupportsHardwareTransformAndLight)
                {
                    flags = CreateFlags.HardwareVertexProcessing;

                    //   if (hardware.MaxActiveLights > 0)
                    //  this.lighting = true;
                    // flags |= CreateFlags.PureDevice;
                    //  if (hardware.DeviceCaps.m
                }


                device = new Device(adapterOrdinal, hardware.DeviceType, renderpanel, flags, presentParams); //Create a device
                if (device != null)
                {
                    break;
                }
            }









            device.DeviceReset += new System.EventHandler(OnResetDevice);
            this.pause = false;
            OnResetDevice(device, null);
        }


        public void OnResetDevice(object sender, EventArgs e)
        {
            Device dev = (Device)sender;
            // Turn off culling, so we see the front and back of the triangle
            dev.RenderState.CullMode = Cull.None;
            // Turn off D3D lighting
           
            dev.RenderState.Lighting = true;

            // Turn on the ZBuffer
            dev.RenderState.ZBufferEnable = true;
            dev.RenderState.FillMode = FillMode.Solid;

            if (dev.DeviceCaps.SourceBlendCaps.SupportsInverseSourceAlpha && dev.DeviceCaps.SourceBlendCaps.SupportsSourceAlpha)
            {
                dev.RenderState.SourceBlend = Blend.SourceAlpha;
                dev.RenderState.DestinationBlend = Blend.InvSourceAlpha;
            }

            if (dev.DeviceCaps.TextureFilterCaps.SupportsMinifyLinear)
            {
                device.SamplerState[0].MinFilter = TextureFilter.Linear;
                device.SamplerState[1].MinFilter = TextureFilter.Linear;
                device.SamplerState[2].MinFilter = TextureFilter.Linear;
                device.SamplerState[3].MinFilter = TextureFilter.Linear;
            }
            if (dev.DeviceCaps.TextureFilterCaps.SupportsMagnifyLinear)
            {
                device.SamplerState[0].MagFilter = TextureFilter.Linear;
                device.SamplerState[1].MagFilter = TextureFilter.Linear;
                device.SamplerState[2].MagFilter = TextureFilter.Linear;
                device.SamplerState[3].MagFilter = TextureFilter.Linear;
            }
            if (dev.DeviceCaps.TextureFilterCaps.SupportsMipMapLinear)
            {
                device.SamplerState[0].MipFilter = TextureFilter.Linear;
                device.SamplerState[1].MipFilter = TextureFilter.Linear;
                device.SamplerState[2].MipFilter = TextureFilter.Linear;
                device.SamplerState[3].MipFilter = TextureFilter.Linear;
            }






      /*
 
     device.Lights[0].Diffuse = Color.White;
     device.Lights[0].Type = LightType.Point; ;
     device.Lights[0].Position = new Vector3(0, -1, -3);
     device.Lights[0].Direction = new Vector3(0, 0, -3);
     device.Lights[0].Range = 1000f;
     device.Lights[0].Update(); // Or Update() with the 2004-Oct SDK
     device.Lights[0].Enabled = true;
    */
             
            // Add a little ambient light to the scene
            device.RenderState.Ambient = Color.FromArgb(0xAF, 0xAF, 0xAF);

        }

        public void BeginScene(Color backgroundcolor)
        {
            //Clear the backbuffer to a blue color 
            device.Clear(ClearFlags.Target | ClearFlags.ZBuffer, backgroundcolor, 1.0f, 0);
            //Begin the scene
            device.BeginScene();
        }

        public void EndScene()
        {
            //End the scene
            device.EndScene();
            // Update the screen
            device.Present();
        }

        public Mesh MakeMesh(List<Vector3> vertices, List<short> Indices, List<Vector2> uvs)
        {
            Mesh m = new Mesh(Indices.Count / 3, vertices.Count, MeshFlags.Managed, Raw.HaloVertex.FVF, device);
            Raw.HaloVertex[] tempv = new entity.Raw.HaloVertex[vertices.Count];
            for (int x = 0; x < vertices.Count; x++)
            {
                tempv[x].Position = vertices[x];
                tempv[x].Tu0 = uvs[x].X;
                tempv[x].Tv0 = uvs[x].Y;
            }
            VertexBuffer vb = m.VertexBuffer;
            vb.SetData(tempv, 0, LockFlags.None);
            vb.Unlock();

            IndexBuffer ibx = m.IndexBuffer;
            ibx.SetData(Indices.ToArray(), 0, LockFlags.None);
         
            return m;
        }

        public static Mesh MakeMesh(ref Raw.ParsedModel pm, int chunknumber,ref Device device)
        {
            short[] indices;
            int[] facecount = new int[pm.RawDataMetaChunks[chunknumber].SubMeshInfo.Length];
            int[] facestart = new int[pm.RawDataMetaChunks[chunknumber].SubMeshInfo.Length];
            if (pm.RawDataMetaChunks[chunknumber].FaceCount * 3 != pm.RawDataMetaChunks[chunknumber].Indices.Length)
            {
                short[] tempindicesx = new short[100000]; 
                int tempc=0;
                for (int x = 0; x < pm.RawDataMetaChunks[chunknumber].SubMeshInfo.Length; x++)
                {
                    short[] tempindices = DecompressIndices(pm.RawDataMetaChunks[chunknumber].Indices, pm.RawDataMetaChunks[chunknumber].SubMeshInfo[x].IndiceStart, pm.RawDataMetaChunks[chunknumber].SubMeshInfo[x].IndiceCount);
                    Array.ConstrainedCopy(tempindices, 0, tempindicesx, tempc, tempindices.Length);
                    
                    facecount[x] = tempindices.Length/3;
                    facestart[x] = tempc;
                    tempc += tempindices.Length;
                }
                indices = new short[tempindicesx.Length];
                Array.Copy(tempindicesx, indices, tempindicesx.Length);
            
            }
            else
            {
                indices = pm.RawDataMetaChunks[chunknumber].Indices;
                for (int x = 0; x < pm.RawDataMetaChunks[chunknumber].SubMeshInfo.Length; x++)
                {
                    facecount[x] = pm.RawDataMetaChunks[chunknumber].SubMeshInfo[x].IndiceCount/3;
                    facestart[x] = pm.RawDataMetaChunks[chunknumber].SubMeshInfo[x].IndiceStart/3;
                }
            }

            Mesh m = new Mesh(indices.Length/3, pm.RawDataMetaChunks[chunknumber].Vertices.Count, MeshFlags.Managed, Raw.HaloVertex.FVF, device);
            VertexBuffer vb = m.VertexBuffer;
            Raw.HaloVertex[] tempv = new entity.Raw.HaloVertex[pm.RawDataMetaChunks[chunknumber].Vertices.Count];
            for (int x = 0; x < pm.RawDataMetaChunks[chunknumber].Vertices.Count; x++)
            {
                tempv[x].Position = pm.RawDataMetaChunks[chunknumber].Vertices[x];
                tempv[x].Normal = pm.RawDataMetaChunks[chunknumber].Normals[x];
                tempv[x].Tu0 = pm.RawDataMetaChunks[chunknumber].UVs[x].X;
                tempv[x].Tv0 = pm.RawDataMetaChunks[chunknumber].UVs[x].Y;
               
            }
            vb.SetData(tempv, 0, LockFlags.None);
            vb.Unlock();
           

            IndexBuffer ibx = m.IndexBuffer;
            ibx.SetData(indices, 0, LockFlags.None);

            return m;
        }

        #region MeshIntersection
        public bool MeshPick(float x, float y, Mesh mesh, Matrix mat)
        {

            Vector3 s = Vector3.Unproject(new Vector3(x, y, 0),
                device.Viewport,
                 device.Transform.Projection,
                 device.Transform.View,
                mat);

            Vector3 d = Vector3.Unproject(new Vector3(x, y, 1),
                 device.Viewport,
                 device.Transform.Projection,
                device.Transform.View,
                mat);

            Vector3 rPosition = s;
            Vector3 rDirection = Vector3.Normalize(d - s);


            if (mesh.Intersect(rPosition, rDirection))
            {

                return true;
            }
            else
            {
                return false;
            }


        }

        public Vector3 Mark3DCursorPosition(float x, float y, Matrix m)
        {


            Vector3 tempV3 = new Vector3();
            tempV3.Project(device.Viewport, device.Transform.Projection, device.Transform.View, m);

            tempV3.X = x;
            tempV3.Y = y;
            tempV3.Unproject(device.Viewport, device.Transform.Projection,
                device.Transform.View, m);

            return tempV3;

        }

        #endregion

        public static void SetAlphaBlending(MetaContainers.ShaderInfo.AlphaType alphatype, ref Device device)
        {
            if (alphatype == MetaContainers.ShaderInfo.AlphaType.AlphaBlend)
            {

                device.RenderState.AlphaBlendEnable = true;
                device.RenderState.AlphaTestEnable = false;
            }
            else if (alphatype == MetaContainers.ShaderInfo.AlphaType.AlphaTest)
            {

                device.RenderState.AlphaBlendEnable = false;
                device.RenderState.AlphaTestEnable = true;
            }
            else
            {
                device.RenderState.AlphaBlendEnable = false;
                device.RenderState.AlphaTestEnable = false;
            }

        }

        public static short[] DecompressIndices(short[] indices, int start, int count)
        {
            bool dir = false;
            short tempx;
            short tempy;
            short tempz;
            short[] shite=new short[50000];
            int m = start;
            int s = 0;
            do
            {
               

                tempx = indices[m];
                tempy = indices[m + 1];
                tempz = indices[m + 2];


                if (tempx != tempy && tempx != tempz && tempy != tempz)
                {
                    if (dir == false)
                    {
                        shite[s] = tempx;
                        shite[s + 1] = tempy;
                        shite[s + 2] = tempz;
                        s += 3;

                        dir = true;
                    }
                    else
                    {
                        shite[s] = tempx;
                        shite[s + 1] = tempz;
                        shite[s + 2] = tempy;
                        s += 3;
                        dir = false;
                    }
                    m += 1;
                }
                else
                {
                    if (dir == true) { dir = false; }
                    else { dir = true; }

                    m += 1;
                }
            }
            while (m < start+count - 2);
            short[] uncompressedindices = new short[s];
            Array.Copy(shite, uncompressedindices, s);
            return uncompressedindices;
        }
        public static float DegreeToRadian(float degree)
        {
            return (float)((Math.PI * degree) / 180.0);
        }
           public static float RadianToDegree(float radian)
        {
            return (float)((radian * 180.0) / Math.PI);
        }


    }

}
